#ifndef _ENGINE_CORE_CONCURRENCY_DYNAMICBITSET_HPP
#define _ENGINE_CORE_CONCURRENCY_DYNAMICBITSET_HPP

#include<vector>

/* Baustelle der Idee einen effizienten Iterator ber die positiven Elemente eines dichten Bitsets zu finden */
class SetElementIterator;

template<size_t STRIDE, size_t BITS_PER_STRIDE>
class DynamicBitset
{
public:
	DynamicBitset();
	DynamicBitset(size_t initial_size);
	
	DynamicBitset<STRIDE, BITS_PER_STRIDE>& set();
	DynamicBitset<STRIDE, BITS_PER_STRIDE>& set(size_t pos);
	
	DynamicBitset<STRIDE, BITS_PER_STRIDE>& reset();
	DynamicBitset<STRIDE, BITS_PER_STRIDE>& reset(size_t pos);
	
	bool test(size_t pos) const;
	bool any() const;
	bool any(size_t stride) const;
	

	class SetElementIterator
	{
	public:
		bool operator!=(const SetElementIterator& other) const;
		size_t operator*() const;
		const SetElementIterator& operator++();
	};
	
	SetElementIterator begin() const;
	SetElementIterator end() const;
	
};


template<>
class DynamicBitset<4, 32>
{
	static_assert(sizeof(unsigned int) == 4, "datentyp muss ein Stride von 4 besitzen");
	std::vector<unsigned int> m_data;
	size_t m_size; /* Anzahl an verwendeten Bits */

	size_t requiredStrides(size_t bits) const
	{
		return (bits + 31u) / 32u;
	}

	size_t fullStrides() const
	{
		return (m_size % 32u == 0) ? m_data.size() : m_data.size() - 1;
	}

	size_t bitsInIncompleteStride() const
	{
		return m_size % 32u;
	}

public:
	DynamicBitset() : m_size(0), m_data()
	{}

	DynamicBitset(size_t initial_bits) : m_size(initial_bits), m_data(requiredStrides(initial_bits))
	{ /* default initialization ist bereits 0 */ }

	DynamicBitset<4, 32>& set()
	{
		size_t fullstrides = fullStrides();
		for(size_t stride = 0; stride < fullstrides; stride++)
			m_data[stride] = 0xffffffff;

		size_t extrabits = bitsInIncompleteStride();
		if(extrabits > 0)
			m_data[fullstrides] = (1<<extrabits) - 1;

		return *this;
	}

	DynamicBitset<4, 32>& set(size_t pos)
	{
		if(pos >= m_size) throw std::out_of_range("Not that many bits in the bitset");
		
		size_t stridepos = pos/32u;
		m_data[stridepos] |= (1u<<(pos % 32u)); /* egal ob fullstride oder ein extrabit */
		
		return *this;
	}

	DynamicBitset<4, 32>& reset()
	{
		for(size_t i = 0; i < m_data.size(); i++)
			m_data[i] = 0u;

		return *this;
	}

	DynamicBitset<4, 32>& reset(size_t pos)
	{
		if(pos >= m_size) throw std::out_of_range("Not that many bits in the bitset");

		size_t stridepos = pos/32u;
		m_data[stridepos] &= ~(1u<<(pos % 32u));

		return *this;
	}

	bool test(size_t pos) const
	{
		if(pos >= m_size) throw std::out_of_range("Not that many bits in the bitset");

		size_t stridepos = pos/32u;
		
		return (m_data[stridepos] & (1u<<(pos % 32u))) != 0u;
	}

	bool any() const
	{
		for(size_t i = 0; i < m_data.size(); i++)
			if(m_data[i] != 0) return true;
	}

	bool any(size_t stride) const
	{
		return m_data[stride] != 0;
	}

	class SetElementIterator
	{
		size_t m_bitpos;
		const std::vector<unsigned int>& m_data;

		/* http://rosettacode.org/wiki/Find_first_and_last_set_bit_of_a_long_integer */
		static_assert(sizeof(unsigned int) == 4 && sizeof(int) == 4, "datentyp muss ein Stride von 4 besitzen");
		unsigned int lsb32(unsigned int n) {return (unsigned int)(n) & -(int)(n); }
		int lsb32_idx(unsigned int n) { return msb32_idx(lsb32(n)); }
		int msb32_idx(unsigned int n)
		{
			int b = 0;
			if (!n) return -1;
 
			#define step(x) if (n >= ((unsigned int)1) << x) b += x, n >>= x
				step(16); step(8); step(4); step(2); step(1);
			#undef step
			return b;
		}

	public:
		SetElementIterator(const std::vector<unsigned int>& data) : m_data(data)
		{
			/* wir mssen Startpunkt finden, weil begin() nicht zwangsweise bei 0 liegt */
			m_bitpos = 0;
			for(size_t i = 0; i < m_data.size(); i++)
				if(m_data[i] != 0)
				{
					m_bitpos = 32*i + lsb32_idx(m_data[i]);
					break;
				}
		}
		bool operator!=(const SetElementIterator& other) const { return m_bitpos == other.m_bitpos; }
		bool operator*() const { return false; }
		const SetElementIterator& operator++() { m_bitpos++; }
	};
	
	SetElementIterator begin() const 
	{
		return SetElementIterator(m_data);
	}
	SetElementIterator end() const
	{
		return SetElementIterator(m_data);
	}

};


#endif
