#ifndef _ENGINE_CORE_CONCURRENCY_ATOMIC_HPP
#define _ENGINE_CORE_CONCURRENCY_ATOMIC_HPP

/*
Zur Zeit werden atomics von kaum einem Compiler untersttzt.
(Clang 3.1 setzt jedoch N2427 um)
N2427: http://www.open-std.org/jtc1/sc22/wg21/docs/papers/2007/n2427.html
N2664: http://www.open-std.org/jtc1/sc22/wg21/docs/papers/2008/n2664.htm

Selbiges gilt fr alias templates, notwendig fr sinnvolles conditional compiling
N2258: http://www.open-std.org/jtc1/sc22/wg21/docs/papers/2007/n2258.pdf

Sobald diese Standards von einem Compiler untersttzt werden,
kann folgender Pseudocode geschrieben werden:

#if N2427 implementiert && N2664 implementiert && N2258 implementiert
	#include<atomic>
	template<typename T>
	using Atomic = std::atomic<T>;
#else
	Eigene Implementierung fr Atomic hier
#endif

Bis dahin inlined wrapper fr std::atomic.
*/


#ifdef _EDITOR_BUILD
/* hide atomic implementation behind void pointer of doom */

#include "../../dllconfig.hpp"
#include <Windows.h>

namespace Engine
{
	namespace Core
	{
		namespace Concurrency
		{
			template <typename T, unsigned Size = sizeof(T)>
			class Atomic
			{
			public:
				Atomic();
				Atomic(const T& _value);
				Atomic(const Atomic<T, Size>& _value);
				template <typename U> Atomic(const Atomic<U, Size>& _value);
				~Atomic();

				inline T CompareExchange(const T& _exchange, const T& _compare) const;
				inline T Exchange(const T& _exchange) const;

				inline T& operator ++();
				inline T& operator --();

				inline bool operator == (const T& _compare);
				inline bool operator == (const Atomic<T, Size> _compare);
				inline bool operator != (const T& _compare);
				inline bool operator != (const Atomic<T, Size> _compare);

				T& operator = (const T& _exchange);
				T& operator = (const Atomic<T, Size>& _other);
				template<typename U> T& operator = (const Atomic<U, Size>& _other);

				T* Get();
			};
		}
	}
}

#include "Atomic32.hpp"
#include "Atomic64.hpp"

#else

#include<atomic>

namespace Engine
{
	namespace Core
	{
		namespace Concurrency
		{
			template <typename T>
			class Atomic
			{
				std::atomic<T> m_data;

			public:
				Atomic() {}
				Atomic(const T& _value) : m_data(_value) {}
				Atomic(const Atomic<T>& _value) : m_data(_value) {}
				template <typename U> Atomic(const Atomic<U>& _value) { m_data = _value.m_data }
				~Atomic() {};

				inline T CompareExchange(const T& _exchange, const T& _compare)
				{
					T expected = _compare;
					while (!m_data.compare_exchange_weak(expected, _exchange));
					return expected;
				}

				inline T Exchange(const T& _exchange)
				{
					T expected = m_data.load();
					while (!m_data.compare_exchange_weak(expected, _exchange));
					return expected;

				}

				/* Referenzen ergeben nur bei nicht-integralen Typen Sinn.
				   Werden diese fr Atomics gebraucht, besser std::atomic direkt verwenden */
				inline T operator ++() { return ++m_data; }
				inline T operator --() { return --m_data; }

				inline bool operator == (const T& _compare) { return m_data.load() == _compare; }
				inline bool operator == (const Atomic<T> _compare) { return m_data.load() == _compare.m_data.load(); }
				inline bool operator != (const T& _compare) { return m_data.load() != _compare; }
				inline bool operator != (const Atomic<T> _compare) { return m_data.load() != _compare.m_data.load(); }

				T operator = (const T& _exchange) { return Exchange(_exchange); }
				T operator = (const Atomic<T>& _other) { return Exchange(_other.m_data.load()); }
				template<typename U> T& operator = (const Atomic<U>& _other) { return Exchange(_other.m_data.load()); }

				T* Get() { return &(m_data.load()); }
			};
		}
	}
}

#endif /* #ifdef _MANAGED */

#endif