#define DLL_EXPORT

#include "CPUInformation.hpp"

#include <stdio.h>
#include <string.h>
#include <intrin.h>
#include <iomanip>

#include "../../Util/Logger.hpp"
#include "../Control/SystemConsole.hpp"

using namespace Engine;
using namespace Engine::Core;
using namespace Engine::Core::Hardware;
using namespace Engine::Util;
using namespace Engine::Core::Control;

char CPUInformation::cpuString[32] = "";
char CPUInformation::cpuBrandString[64] = "";

int CPUInformation::nLogicalProcessors = 0;
int CPUInformation::nCacheLineSize = 0;
int CPUInformation::nL2Associativity = 0;
int CPUInformation::nCacheSizeK = 0;
int CPUInformation::nFeatureInfo = 0;

bool    CPUInformation::sse1Support = false;
bool    CPUInformation::sse2Support = false;
bool    CPUInformation::sse3Support = false;
bool    CPUInformation::supplementalSSE3Support = false;
bool    CPUInformation::sse41Support = false;
bool    CPUInformation::sse42Support = false;
bool    CPUInformation::misalignedSSESupport = false;
bool    CPUInformation::mmxSupport = false;
bool    CPUInformation::multithreading = false;

unsigned int CPUInformation::nIds = 0;
unsigned int CPUInformation::nExIds = 0;
unsigned int CPUInformation::i = 0;

CPUInformation::Initializer CPUInformation::init;

void CPUInformation::Init()
{
	int CPUInfo[4] = {-1};

	__cpuid(CPUInfo, 0);
    nIds = CPUInfo[0];
    memset(cpuString, 0, sizeof(cpuString));
    *((int*)cpuString) = CPUInfo[1];
    *((int*)(cpuString+4)) = CPUInfo[3];
    *((int*)(cpuString+8)) = CPUInfo[2];

    // Get the information associated with each valid Id
    for (i=0; i<=nIds; ++i)
    {
        __cpuid(CPUInfo, i);
        // Interpret CPU feature information.
        if  (i == 1)
        {
			
			nLogicalProcessors = ((CPUInfo[1] >> 16) & 0xff);
			sse3Support = (CPUInfo[2] & 0x1) || false;
			supplementalSSE3Support = (CPUInfo[2] & 0x200) || false;
			sse41Support = (CPUInfo[2] & 0x80000) || false;
			sse42Support = (CPUInfo[2] & 0x100000) || false;
			nFeatureInfo = CPUInfo[3];
			multithreading = (nFeatureInfo & (1 << 28)) || false;
			sse1Support = (nFeatureInfo & (1 << 25)) || false;
			sse2Support = (nFeatureInfo & (1 << 26)) || false;
        }
    }

    // Calling __cpuid with 0x80000000 as the InfoType argument
    // gets the number of valid extended IDs.
    __cpuid(CPUInfo, 0x80000000);
    nExIds = CPUInfo[0];
    memset(cpuBrandString, 0, sizeof(cpuBrandString));

    // Get the information associated with each extended ID.
    for (i=0x80000000; i<=nExIds; ++i)
    {
        __cpuid(CPUInfo, i);

        if  (i == 0x80000001)
        {
            misalignedSSESupport = (CPUInfo[2] & 0x80) || false;
            mmxSupport = (CPUInfo[3] & 0x40000) || false;
        }

        // Interpret CPU brand string and cache information.
        if  (i == 0x80000002)
            memcpy(cpuBrandString, CPUInfo, sizeof(CPUInfo));
        else if  (i == 0x80000003)
            memcpy(cpuBrandString + 16, CPUInfo, sizeof(CPUInfo));
        else if  (i == 0x80000004)
            memcpy(cpuBrandString + 32, CPUInfo, sizeof(CPUInfo));
        else if  (i == 0x80000006)
        {
            nCacheLineSize = CPUInfo[2] & 0xff;
            nL2Associativity = (CPUInfo[2] >> 12) & 0xf;
            nCacheSizeK = (CPUInfo[2] >> 16) & 0xffff;
        }
        else if  (i == 0x80000008)
        {
          
        }
        else if  (i == 0x8000000A)
        {
            
        }
        else if  (i == 0x8000001A)
        {
        }
    }
}

void CPUInformation::Print()
{
	int CPUInfo[4] = {-1};

	SystemConsole::SetColor(ConsoleColor::WhiteConsoleColor);
	std::cout << "Hardware Information" << std::endl;

	SystemConsole::SetColor(ConsoleColor::GrayConsoleColor);
	if  (nExIds >= 0x80000004)
        printf_s("%s\n", cpuBrandString);

    if  (nExIds >= 0x80000006)
    {
        printf_s("Cache Line Size = %d\n", nCacheLineSize);
        printf_s("L2 Associativity = %d\n", nL2Associativity);
        printf_s("Cache Size = %dK\n", nCacheSizeK);
    }

	SystemConsole::SetColor(ConsoleColor::WhiteConsoleColor);
	std::cout << std::endl << "Checking feature set" << std::endl;

	SystemConsole::SetColor(ConsoleColor::GrayConsoleColor);
	std::cout << "SSE Instructions Set 1               "; PrintBool(sse1Support, "ok", "failed"); std::cout << std::endl;
	std::cout << "SSE Instructions Set 2               "; PrintBool(sse2Support, "ok", "failed"); std::cout << std::endl;
	std::cout << "SSE Instructions Set 3               "; PrintBool(sse3Support, "ok", "failed"); std::cout << std::endl;
	std::cout << "SSE 3 Supplementals                  "; PrintBool(supplementalSSE3Support, "ok", "failed"); std::cout << std::endl;
	std::cout << "SSE Instructions Set 4.1             "; PrintBool(sse41Support, "ok", "failed"); std::cout << std::endl;
	std::cout << "SSE Instructions Set 4.2             "; PrintBool(sse42Support, "ok", "failed"); std::cout << std::endl;
	std::cout << "SSE Misaligned Support               "; PrintBool(misalignedSSESupport, "ok", "failed"); std::cout << std::endl;
}


inline bool CPUInformation::HasSSE41Extensions()
{
	return sse41Support;
}

void CPUInformation::PrintBool(bool _value, std::string _trueText, std::string _falseText)
{
	if (_value)
	{
		SystemConsole::SetColor(ConsoleColor::GreenConsoleColor);
		std::cout << _trueText;
	}
	else
	{
		SystemConsole::SetColor(ConsoleColor::RedConsoleColor);
		std::cout << _falseText;
	}

	SystemConsole::SetColor(ConsoleColor::GrayConsoleColor);
}

CPUInformation::Initializer::Initializer()
{
	CPUInformation::Init();
}