#define DLL_EXPORT

#include "AssetLoader.hpp"

#include <cassert>
#include <fstream>

#include "PngLoader.hpp"
#include "../../Scene/World.hpp"
#include "../../Scene/SceneObject.hpp"
#include "../../Scene/Lights/SpotLight.hpp"
#include "../../Scene/Lights/PointLight.hpp"
#include "../../Scene/Lights/DirectionalLight.hpp"
#include "../../Scene/Animations/Animation.hpp"
#include "../../Scene/Animations/Curve.hpp"
#include "../../Scene/Animations/CurveData.hpp"
#include "../../Scene/Animations/CurvePoint.hpp"
#include "../../Graphic/Materials/DiffuseColorMaterialPart.hpp"
#include "../../Graphic/Materials/SpecularColorMaterialPart.hpp"
#include "../../Graphic/Materials/DiffuseTextureMaterialPart.hpp"
#include "../../Graphic/Materials/OpacityMaterialPart.hpp"
#include "../../Graphic/Materials/EmittingColorMaterialPart.hpp"
#include "../../Graphic/Materials/NormalTextureMaterialPart.hpp"
#include "../../Scene/Components/PositionComponent.hpp"
#include "../../Scene/Components/MeshRenderComponent.hpp"
#include "../../Scene/Components/LightComponents/DirectionalLightComponent.hpp"
#include "../../Scene/Components/LightComponents/SpotLightComponent.hpp"
#include "../../Scene/Components/LightComponents/PointLightComponent.hpp"
#include "../../Scene/Components/TransparentMeshRenderComponent.hpp"
#include "../../Scene/Components/AnimationComponent.hpp"
#include "../../Core/System/SystemConfiguration.hpp"
#include "../../Util/Convert.hpp"

using namespace Engine;
using namespace Engine::Math;
using namespace Engine::Core;
using namespace Engine::Core::System;
using namespace Engine::Core::IO;
using namespace Engine::Util;
using namespace Engine::Scene;
using namespace Engine::Scene::Animations;
using namespace Engine::Scene::Lights;
using namespace Engine::Scene::Components;
using namespace Engine::Scene::Components::LightComponents;
using namespace Engine::Graphic;
using namespace Engine::Graphic::Meshes;
using namespace Engine::Graphic::Shaders;
using namespace Engine::Graphic::Materials;
using namespace Engine::Graphic::Buffers;


#define VERSIONNUMBER ((unsigned short)8)

#define DICT_FILE "../../../TestData/Output/AssetOutput.blubbdict"
#define SHADER_FILE "../../../TestData/Output/AssetOutput.blubbsh"
#define DATA_FILE "../../../TestData/Output/AssetOutput.blubbbin"
#define DICTIONARY_ENTRY_SIZE 10
#define DICTIONARY_NAME_ENTRY_SIZE 12 //12 Bytes without Name string

AssetLoader::AssetLoader(GraphicContext* _context, EngineManager* _manager) : assetNames(), context(_context), manager(_manager)
{
	LoadDictionaries();

	shaderFile = fopen(SHADER_FILE, "rb");
	if (shaderFile == nullptr)
	{
		Logger::Fatal("Failed to open shader file");
		return;
	}

	unsigned short versionNumber;
	fread(&versionNumber, sizeof(unsigned short), 1, shaderFile);
	if (versionNumber != VERSIONNUMBER)
	{
		Logger::Fatal(std::string("Wrong binary file version (version is ") + Convert::ToString(versionNumber) + ", but should be " + Convert::ToString(VERSIONNUMBER) + ")");
		Logger::Fatal("You may recompile all assets in the editor");
		system("PAUSE");
		exit(-1);
	}

	dataFile = fopen(DATA_FILE, "rb");
	if (dataFile == nullptr)
	{
		Logger::Fatal("Failed to open data file");
		return;
	}

	fread(&versionNumber, sizeof(unsigned short), 1, dataFile);
	if (versionNumber != VERSIONNUMBER)
	{
		Logger::Fatal(std::string("Wrong binary file version (version is ") + Convert::ToString(versionNumber) + ", but should be " + Convert::ToString(VERSIONNUMBER) + ")");
		Logger::Fatal("You may recompile all assets in the editor");
		system("PAUSE");
		exit(-1);
	}

	//Load Shader
	unifiedVertexShaderCode = Shader::LoadShaderFile(SystemConfiguration::Get<std::string>("UnifiedVertexShaderPath"));
	unifiedNormalShaderCode = Shader::LoadShaderFile(SystemConfiguration::Get<std::string>("UnifiedNormalShaderPath"));
	unifiedDeferredShaderCode = Shader::LoadShaderFile(SystemConfiguration::Get<std::string>("UnifiedDeferredShaderPath"));
	unifiedForwardShaderCode = Shader::LoadShaderFile(SystemConfiguration::Get<std::string>("UnifiedForwardShaderPath"));
	unifiedOctreeVertexCode = Shader::LoadShaderFile(SystemConfiguration::Get<std::string>("OctreeUnifiedVertexShaderPath"));
	unifiedOctreeGeometryCode = Shader::LoadShaderFile(SystemConfiguration::Get<std::string>("OctreeUnifiedGeometryShaderPath"));
	unifiedOctreeFragmentCode = Shader::LoadShaderFile(SystemConfiguration::Get<std::string>("OctreeUnifiedFragmentShaderPath"));
}

AssetLoader::~AssetLoader()
{
	if (shaderFile != nullptr)
	{
		fclose(shaderFile);
		shaderFile = nullptr;
	}

	if (dataFile != nullptr)
	{
		fclose(dataFile);
		dataFile = nullptr;
	}

	if (deferredShaderCache != nullptr)
		delete[] deferredShaderCache;
	if (forwardShaderCache != nullptr)
		delete[] forwardShaderCache;
	if (normalShaderCache != nullptr)
		delete[] normalShaderCache;
	if (textureCache != nullptr)
		delete[] textureCache;
	if (materialCache != nullptr)
		delete[] materialCache;
	if (meshCache != nullptr)
		delete[] meshCache;
	if (transparentMeshCache != nullptr)
		delete[] transparentMeshCache;
	if (octreeShaderCache != nullptr)
		delete[] octreeShaderCache;
	
	if (curveCache != nullptr)
	{
		for (unsigned int i = 0; i < curveCacheCount; i++)
		{
			SharedPointer<CurveData>** cdlist = curveCache[i];
			if (cdlist != nullptr)
				delete[] cdlist;
		}
		delete[] curveCache;
	}
}

void AssetLoader::LoadDictionaries()
{
	FILE* dictionaryFile = fopen(DICT_FILE, "rb");
	if (dictionaryFile == nullptr)
	{
		Logger::Fatal("Failed to open dictionary file");
		return;
	}

	unsigned short versionNumber;
	fread(&versionNumber, sizeof(unsigned short), 1, dictionaryFile);
	if (versionNumber != VERSIONNUMBER)
	{
		Logger::Fatal(std::string("Wrong binary file version (version is ") + Convert::ToString(versionNumber) + ", but should be " + Convert::ToString(VERSIONNUMBER) + ")");
		Logger::Fatal("You may recompile all assets in the editor");
		system("PAUSE");
		exit(-1);
	}

	unsigned short dictSize;

	Logger::Debug("AssetDictionary");
	LoadNamedDictionary(dictionaryFile, &assetDictionary, assetNames, true);
	
	//Logger::Debug("MeshDictionary");
	dictSize = LoadDictionary(dictionaryFile, &meshDictionary);
	meshCache = new SharedPointer<Mesh>*[dictSize];
	transparentMeshCache = new SharedPointer<TransparentMesh>*[dictSize];
	ZeroMemory(meshCache, sizeof(SharedPointer<Mesh>*) * dictSize);
	ZeroMemory(transparentMeshCache, sizeof(SharedPointer<TransparentMesh>*) * dictSize);
	
	//Logger::Debug("MaterialDictionary");
	dictSize = LoadNamedDictionary(dictionaryFile, &materialDictionary, materialNames);
	materialCache = new SharedPointer<Material>*[dictSize];
	ZeroMemory(materialCache, sizeof(SharedPointer<Material>*) * dictSize);
	
	//Logger::Debug("TextureDictionary");
	dictSize = LoadDictionary(dictionaryFile, &textureDictionary);
	textureCache = new SharedPointer<Texture2D>*[dictSize];
	ZeroMemory(textureCache, sizeof(SharedPointer<Texture2D>*) * dictSize);

	//Logger::Debug("AnimationDictionary");
	dictSize = LoadDictionary(dictionaryFile, &animationDictionary);
	curveCacheCount = dictSize;
	curveCache = new SharedPointer<CurveData>**[dictSize];
	ZeroMemory(curveCache, sizeof(SharedPointer<CurveData>**) * dictSize);

	//Logger::Debug("ShaderDictionary");
	dictSize = LoadDictionary(dictionaryFile, &shaderDictionary);
	deferredShaderCache = new ShaderPtr*[dictSize];
	forwardShaderCache = new ShaderPtr*[dictSize];
	normalShaderCache = new ShaderPtr*[dictSize];
	octreeShaderCache = new ShaderPtr*[dictSize];
	ZeroMemory(deferredShaderCache, sizeof(ShaderPtr*) * dictSize);
	ZeroMemory(forwardShaderCache, sizeof(ShaderPtr*) * dictSize);
	ZeroMemory(normalShaderCache, sizeof(ShaderPtr*) * dictSize);
	ZeroMemory(octreeShaderCache, sizeof(ShaderPtr*) * dictSize);

	fclose(dictionaryFile);
}

unsigned short AssetLoader::LoadDictionary(FILE *_file, DictionaryEntry** _dictionary)
{
	unsigned short count;
	fread(&count, sizeof(unsigned short), 1, _file);

	//Initialize Dictionary
	*_dictionary = new DictionaryEntry[count];

	for (unsigned short i = 0; i < count; i++)
	{
		ReadEntry(_file, *_dictionary);
	}

	return count;
}

unsigned short AssetLoader::LoadNamedDictionary(FILE *_file, DictionaryEntry** _dictionary, std::unordered_map<std::string, unsigned short>& _names, bool _showLog)
{
	unsigned short count;
	fread(&count, sizeof(unsigned short), 1, _file);

	//Initialize Dictionary
	*_dictionary = new DictionaryEntry[count];
	_names.reserve(count);

	for (unsigned short i = 0; i < count; i++)
	{
		ReadNamedEntry(_file, *_dictionary, _names, _showLog);
	}

	return count;
}

DictionaryEntry AssetLoader::ReadEntry(FILE *_file, DictionaryEntry* _dictionary)
{
	unsigned short id;
	unsigned int start, length;
		
	//Read AssetEntries
	char buffer[10];
	fread(&buffer, DICTIONARY_ENTRY_SIZE, 1, _file);
	id = *(unsigned short*)buffer;
	start = *(unsigned int*)(buffer + 2);
	length = *(unsigned int*)(buffer + 6);

	DictionaryEntry entry(id, start, length);
	_dictionary[id] = entry;

	//Logger::Debug(std::string("Id: ") + Convert::ToString(id) + ", StartAddress: " + Convert::ToString(start) + ", Length: " + Convert::ToString(length));

	return entry;
}

DictionaryEntry AssetLoader::ReadNamedEntry(FILE *_file, DictionaryEntry* _dictionary, std::unordered_map<std::string, unsigned short>& _names, bool _showLog)
{
	unsigned short id, stringLength;
	unsigned int start, length;
		
	//Read AssetEntries
	char buffer[256];
	fread(&buffer, DICTIONARY_NAME_ENTRY_SIZE, 1, _file);
	id = *(unsigned short*)buffer;
	start = *(unsigned int*)(buffer + 2);
	length = *(unsigned int*)(buffer + 6);
	stringLength = *(unsigned short*)(buffer + 10);

	DictionaryEntry entry(id, start, length);
	_dictionary[id] = entry;

	//Load Name
	fread(&buffer, stringLength, 1, _file);
	buffer[stringLength] = '\0';
	std::string name(buffer);
		
	_names.insert(std::pair<std::string, unsigned short>(name, id));

	if (_showLog)
		Logger::Debug(std::string("Id: ") + Convert::ToString(id) + ", StartAddress: " + Convert::ToString(start) + ", Length: " + Convert::ToString(length) + ", Alias: " + name);

	return entry;
}

ShaderPtr AssetLoader::LoadShader(unsigned short _id, unsigned int _mode, ShaderPtr** _dictionary)
{
	//Check if in cache
	if (_dictionary[_id] != nullptr && _dictionary[_id]->GetCounter() > 0)
		return *_dictionary[_id];
	else
	{
		//Binrdaten laden
		DictionaryEntry& entry = shaderDictionary[_id];
		fseek(shaderFile, entry.StartAddress(), SEEK_SET);
		char buffer[100 * 1024];
		assert(entry.Lenght() < 1024 * 1024);
		fread(&buffer, entry.Lenght(), 1, shaderFile);

		//Binrdaten zerlegen
		char* readPtr = buffer;
		unsigned char shaderType =*(unsigned char*)readPtr;
		readPtr += sizeof(unsigned char);

		Shader* shader = new Shader(GL_TRIANGLES, context);

		if (shaderType == 0) //Files
		{
			char* codePointer = readPtr + (sizeof(unsigned short) * 6); //Zeigt auf den Beginn des ersten Code-Blocks
			const GLenum types[2] = { GL_VERTEX_SHADER, GL_FRAGMENT_SHADER };

			for (unsigned int i = 0; i < 6; i++)
			{
				unsigned short length = *(unsigned short*)(readPtr + (sizeof(unsigned short) * i));

				if (_mode <= 2)
				{
					if (i / 2 == _mode)
					{
						if (length != 0) //Falls Shader existieren
							shader->AttachShader(codePointer, types[i % 2]);
					}
				}
				else
					Logger::Fatal("Unable to load FileShader as LinkedListShader");

				codePointer += length;
			}
		}
		else if (shaderType == 1) //Unified
		{
			char* codePointer = readPtr + sizeof(unsigned short);
			std::string code(unifiedVertexShaderCode);

			if (_mode != 3)
			{
				code.append(codePointer);			
				shader->AttachShader(code.c_str(), GL_VERTEX_SHADER);
			}

			switch (_mode)
			{
			case 0: //Normal Output
				code = unifiedNormalShaderCode;
				code.append(codePointer);
				shader->AttachShader(code.c_str(), GL_FRAGMENT_SHADER);
				break;
			case 1: //Deferred
				code = unifiedDeferredShaderCode;
				code.append(codePointer);
				shader->AttachShader(code.c_str(), GL_FRAGMENT_SHADER);
				break;
			case 2: //Forward
				code = unifiedForwardShaderCode;
				code.append(codePointer);
				shader->AttachShader(code.c_str(), GL_FRAGMENT_SHADER);
				break;
			case 3: //Voxelization
				code = unifiedOctreeVertexCode;
				code.append(codePointer);
				shader->AttachShader(code.c_str(), GL_VERTEX_SHADER);
				code = unifiedOctreeGeometryCode;
				code.append(codePointer);
				shader->AttachShader(code.c_str(), GL_GEOMETRY_SHADER);
				code = unifiedOctreeFragmentCode;
				code.append(codePointer);
				shader->AttachShader(code.c_str(), GL_FRAGMENT_SHADER);
				break;
			default:
				Logger::Fatal("Unknown shader mode. Valid range: [0-3] but is " + Convert::ToString(_mode));
				break;
			}
		}
		else
			Logger::Fatal("Unknown shader type: "+ Convert::ToString(shaderType));


		//Bind attributes and uniforms
		shader->BindAttribute("position", SHADER_POSITION);
		shader->BindAttribute("normal", SHADER_MATERIAL_NORMAL);
		shader->BindAttribute("texcoord_diffuse", SHADER_MATERIAL_TEXCOORD_DIFFUSE);

		switch (_mode)
		{
		case 0: //Normal Output
			shader->BindOutput("fragNormal", 0);
			shader->BindOutput("fragDepth", 1);
			break;
		case 1: //Deferred
		case 2: //Forward
		case 3: //Voxelization			
			shader->BindOutput("fragColor", 0);
			break;
		}

		shader->Link();

		ShaderPtr* shaderSP = new ShaderPtr(shader);
		--(*(shaderSP->GetCounter()));

		_dictionary[_id] = shaderSP;
		
		return *shaderSP;
	}
}

ShaderPtr AssetLoader::LoadDeferredShader(unsigned short _id)
{
	ShaderPtr shader = LoadShader(_id,  1, deferredShaderCache);
	return shader;
}

ShaderPtr AssetLoader::LoadForwardShader(unsigned short _id)
{
	ShaderPtr shader = LoadShader(_id,  2, forwardShaderCache);
	return shader;
}

ShaderPtr AssetLoader::LoadNormalShader(unsigned short _id)
{
	ShaderPtr shader = LoadShader(_id,  0, normalShaderCache);
	return shader;
}

ShaderPtr AssetLoader::LoadOctreeShader(unsigned short _id)
{
	ShaderPtr shader = LoadShader(_id,  3, octreeShaderCache);
	return shader;
}

Util::SharedPointer<Texture2D> AssetLoader::LoadTexture(unsigned short _id, bool _isGammaCorrect)
{
	if (textureCache[_id] != nullptr && textureCache[_id]->GetCounter() != 0)
		return *textureCache[_id];
	else
	{
		//Not cached
		DictionaryEntry& entry = textureDictionary[_id];
		fseek(dataFile, entry.StartAddress(), SEEK_SET);
		Texture2D* texture = PngLoader::Load(dataFile, context, _isGammaCorrect);
		SharedPointer<Texture2D>* textureSP = new SharedPointer<Texture2D>(texture);
		--(*textureSP->GetCounter());

		textureCache[_id] = textureSP;

		return *textureSP;
	}
}

Util::SharedPointer<Graphic::Materials::Material> AssetLoader::LoadMaterial(unsigned short _id)
{
	if (materialCache[_id] != nullptr && materialCache[_id]->GetCounter() != 0)
		return *materialCache[_id];
	else
	{
		DictionaryEntry& entry = materialDictionary[_id];
		
		fseek(dataFile, entry.StartAddress(), SEEK_SET);
		char buffer[1024];
		assert(entry.Lenght() < 1024);
		fread(&buffer, entry.Lenght(), 1, dataFile);

		//Parse Material
		char* dataPointer = buffer;
		unsigned short shaderId = *(unsigned short*)dataPointer;
		dataPointer += sizeof(unsigned short);

		ShaderPtr shader = LoadDeferredShader(shaderId);
		ShaderPtr forwardShader = LoadForwardShader(shaderId);
		ShaderPtr normalShader = LoadNormalShader(shaderId);
		ShaderPtr octreeShader = LoadOctreeShader(shaderId);
		Material* material = new Material(shader, forwardShader, normalShader, octreeShader);

		//Shadeless
		material->IsShadeless(*(bool*)dataPointer);
		dataPointer += sizeof(bool);

		//Parts
		char partCount = *(char*)dataPointer;
		dataPointer += sizeof(char);

		for (char i = 0; i < partCount; i++)
		{
			MaterialPartType partType = *(MaterialPartType*)dataPointer;
			dataPointer += sizeof(MaterialPartType);
			IMaterialPart* part = nullptr;

			switch (partType)
			{
			case MaterialPartType::DiffuseColorMaterialPartType:
				{
					float* x = (float*)dataPointer;
					Vec3f color((float*)dataPointer);
					dataPointer += sizeof(float)*3;
					part = new DiffuseColorMaterialPart(color);
				}
				break;
			case MaterialPartType::SpecularColorMaterialPartType:
				{
					Vec3f color((float*)dataPointer);
					dataPointer += sizeof(float)*3;
					float hardness = *(float*)dataPointer;
					dataPointer += sizeof(float);
					part = new SpecularColorMaterialPart(color, hardness);
				}
				break;
			case MaterialPartType::DiffuseTextureMaterialPartType:
				{
					unsigned short textureId = *(unsigned short*)dataPointer;
					dataPointer += sizeof(unsigned short);
					part = new DiffuseTextureMaterialPart(LoadTexture(textureId, true));
				}
				break;
			case MaterialPartType::NormalTextureMaterialPartType:
				{
					unsigned short textureId = *(unsigned short*)dataPointer;
					dataPointer += sizeof(unsigned short);
					part = new NormalTextureMaterialPart(LoadTexture(textureId, false));
				}
				break;
			case MaterialPartType::OpacityMaterialPartType:
				{
					float opacity = *(float*)dataPointer;

					if (opacity < 0.99)
						material->IsTransparent(true);
					else
						opacity = 1.0f;

					dataPointer += sizeof(float);
					part = new OpacityMaterialPart(opacity);
				}
				break;
			case MaterialPartType::EmittingColorMaterialPartType:
				{
					Vec3f color((float*)dataPointer);
					dataPointer += sizeof(float)*3;
					part = new EmittingColorMaterialPart(color);
				}
				break;
			}

			if (part == nullptr)
				Logger::Fatal("Unknown MaterialPartId");
			else
			{
				material->Parts().push_back(part);
			}
		}

		SharedPointer<Material>* sp = new SharedPointer<Material>(material);
		materialCache[_id] = sp;
		--(*sp->GetCounter());

		return *sp;
	}
}

Util::SharedPointer<Material> AssetLoader::LoadMaterial(std::string _name)
{
	unsigned short id = materialNames[_name];
	return LoadMaterial(id);
}

Util::SharedPointer<Mesh> AssetLoader::LoadMesh(unsigned short _id)
{
	if (meshCache[_id] != nullptr && meshCache[_id]->GetCounter() != 0)
		return *meshCache[_id];
	else
	{
		std::vector<VertexStreamElement> mappings;
		float* data;
		unsigned int* indices;
		unsigned short dataCount, indexCount, vertexSize;
		Vec4f boundingLocation;
		Vec3f boundingSize;
		ReadMeshData(_id, &data, dataCount, vertexSize, &indices, indexCount, mappings, boundingLocation, boundingSize);

		Mesh* m = new Mesh(data, dataCount, vertexSize, indices, indexCount, mappings, boundingLocation, boundingSize, context);
		SharedPointer<Mesh>* sp = new SharedPointer<Mesh>(m);
		meshCache[_id] = sp;

		--*sp->GetCounter();
		return *sp;
	}
}

Util::SharedPointer<Graphic::Meshes::TransparentMesh> AssetLoader::LoadTransparentMesh(unsigned short _id)
{
	if (transparentMeshCache[_id] != nullptr && transparentMeshCache[_id]->GetCounter() != 0)
		return *transparentMeshCache[_id];
	else
	{
		std::vector<VertexStreamElement> mappings;
		float* data;
		unsigned int* indices;
		unsigned short dataCount, indexCount, vertexSize;
		Vec4f boundingLocation;
		Vec3f boundingSize;
		ReadMeshData(_id, &data, dataCount, vertexSize, &indices, indexCount, mappings, boundingLocation, boundingSize);

		TransparentMesh* m = new TransparentMesh(data, dataCount, vertexSize, indices, indexCount, mappings, boundingLocation, boundingSize, context);
		SharedPointer<TransparentMesh>* sp = new SharedPointer<TransparentMesh>(m);
		transparentMeshCache[_id] = sp;

		--*sp->GetCounter();
		return *sp;
	}
}

void AssetLoader::ReadMeshData(unsigned int _id, float** _data, unsigned short& _dataCount, unsigned short& _vertexSize, unsigned int **_indices, unsigned short& _indexCount,
				  std::vector<Graphic::Meshes::VertexStreamElement>& _mappings, Math::Vec4f& _boundingLocation, Math::Vec3f& _boundingSize)
{
	DictionaryEntry& entry = meshDictionary[_id];
	fseek(dataFile, entry.StartAddress(), SEEK_SET);

	assert(entry.Lenght() < 1024*100);

	char buffer[100*1024];
	fread(&buffer, entry.Lenght(), 1, dataFile);
	char* dataPointer = buffer;
		
	//Load Mappings
	char mappingCount = *buffer;
	dataPointer += sizeof(char);
	_mappings.reserve(mappingCount);

	for (char i = 0; i < mappingCount; i++)
	{
		unsigned int handle = *(unsigned int*)dataPointer;
		dataPointer += sizeof(unsigned int);
		int size = *(int*)dataPointer;
		dataPointer += sizeof(int);
		unsigned int offset = *(unsigned int*)dataPointer;
		dataPointer += sizeof(unsigned int);

		_mappings.push_back(VertexStreamElement(handle, size, offset));
	}
		
	_dataCount = *(unsigned short*)dataPointer;
	dataPointer += sizeof(unsigned short);
	_vertexSize = *(unsigned short*)dataPointer;
	dataPointer += sizeof(unsigned short);
	*_data = (float*)dataPointer;
	dataPointer += _dataCount * _vertexSize;
	_indexCount = *(unsigned short*)dataPointer;
	dataPointer += sizeof(unsigned short);
	*_indices = (unsigned int*)dataPointer;
	dataPointer += _indexCount * sizeof(unsigned int);

	Vec4f boundingLocation((float*)dataPointer);
	boundingLocation.W(1.0f);
	_boundingLocation = boundingLocation;
	dataPointer += sizeof(float) * 3;
	Vec3f boundingSize((float*)dataPointer);
	_boundingSize = boundingSize;
	dataPointer += sizeof(float) * 3;
}

Scene::SceneObject* AssetLoader::LoadAsset(unsigned short _id, Scene::World* _world)
{
	DictionaryEntry& entry = assetDictionary[_id];
	fseek(dataFile, entry.StartAddress(), SEEK_SET);

	char buffer[1024*100];
	assert(entry.Lenght() <= 1024*100);

	fread(&buffer, entry.Lenght(), 1, dataFile);
	char* bufferPointer = buffer;

	return LoadNode(&bufferPointer, _world)[0];
}

Scene::SceneObject* AssetLoader::LoadAsset(std::string _name, Scene::World* _world)
{
	auto it = assetNames.find(_name);
	
	if (it == assetNames.end())
	{
		Logger::Fatal("Unknown Asset: " + _name);
		system("PAUSE");
		exit(-1);
	}

	return LoadAsset(it->second, _world);
}

std::vector<Scene::SceneObject*> AssetLoader::LoadNode(char** _buffer, Scene::World* _world)
{
	char* dataPointer = *_buffer;
	
	unsigned short nodeCount = *(unsigned short*)dataPointer;
	dataPointer += sizeof(unsigned short);

	std::vector<Scene::SceneObject*> nodes;
	nodes.reserve(nodeCount);

	//Load Nodes
	for (unsigned short i = 0; i < nodeCount; i++)
	{
		Vec3f translate((float*)dataPointer);
		dataPointer += sizeof(float)*3;
		Vec3f rotate((float*)dataPointer);
		dataPointer += sizeof(float)*3;
		Vec3f scale((float*)dataPointer);
		dataPointer += sizeof(float)*3;

		unsigned char nameLength = *(unsigned char*)dataPointer;
		dataPointer += sizeof(unsigned char);
		std::string name(dataPointer, nameLength);
		dataPointer += nameLength;

		SceneObject* so = new SceneObject(_world, name);

		Mat4f matrix = Mat4f::Translate(translate) * Mat4f::RotateZ(rotate.Z()) * Mat4f::RotateY(rotate.Y()) * Mat4f::RotateX(rotate.X()) * Mat4f::Scale(scale);

		PositionComponent* positionComp = new PositionComponent(so);
		so->AddComponent(positionComp);
		so->GetComponent<PositionComponent>()->Matrix(matrix);

		char componentCount = *(char*)dataPointer;
		dataPointer += sizeof(char);

		//Load Components
		for (char j = 0; j < componentCount; j++)
		{
			ComponentType ctype = *(ComponentType*)dataPointer;
			dataPointer += sizeof(ComponentType);

			switch (ctype)
			{
			case ComponentType::MeshRenderComponentType:
				{
					unsigned short meshId = *(unsigned short*)dataPointer;
					dataPointer += sizeof(unsigned short);
					unsigned short materialId = *(unsigned short*)dataPointer;
					dataPointer += sizeof(unsigned short);

					SharedPointer<Material> matSp = LoadMaterial(materialId);

					matSp->IsTransparent(false);

					MeshRenderComponent* mrComp = new MeshRenderComponent(so, LoadMesh(meshId), matSp, context);
					so->AddComponent(mrComp);
				}
				break;
			case ComponentType::TransparentMeshRenderComponentType:
				{
					unsigned short meshId = *(unsigned short*)dataPointer;
					dataPointer += sizeof(unsigned short);
					unsigned short materialId = *(unsigned short*)dataPointer;
					dataPointer += sizeof(unsigned short);

					TransparentMeshRenderComponent* mrComp = new TransparentMeshRenderComponent(so, LoadTransparentMesh(meshId), LoadMaterial(materialId), context);
					so->AddComponent(mrComp);
				}
				break;
			case ComponentType::DirectionalLightComponentType:
				{
					DirectionalLightComponent* lComp = new DirectionalLightComponent(so, new DirectionalLight(Vec3f((float*)dataPointer)));
					dataPointer += sizeof(float) * 3;
					so->AddComponent(lComp);
				}
				break;
			case ComponentType::PointLightComponentType:
				{
					Vec3f intensity((float*)dataPointer);
					dataPointer += sizeof(float) * 3;
					Vec3f att((float*)dataPointer);
					dataPointer += sizeof(float) * 3;
					PointLightComponent* lComp = new PointLightComponent(so, new PointLight(intensity, att), context);
					so->AddComponent(lComp);
				}
				break;
			case ComponentType::SpotLightComponentType:
				{
					Vec3f intensity((float*)dataPointer);
					dataPointer += sizeof(float) * 3;
					Vec3f att((float*)dataPointer);
					dataPointer += sizeof(float) * 3;
					float angle = *(float*)dataPointer;
					dataPointer += sizeof(float);
					float angularAttenuation = *(float*)dataPointer;
					dataPointer += sizeof(float);

					SpotLightComponent* lComp = new SpotLightComponent(so, new SpotLight(intensity, angle, att, angularAttenuation), context);
					so->AddComponent(lComp);
				}
				break;
			case ComponentType::AnimationComponentType:
				{
					unsigned char animCount = *(unsigned char*)dataPointer;
					dataPointer += sizeof(unsigned char);

					AnimationComponent* acomp = new AnimationComponent(so, manager, translate, rotate, scale);

					for (unsigned char i = 0; i < animCount; i++)
					{
						unsigned short animId = *(unsigned short*)dataPointer;
						dataPointer += sizeof(unsigned short);
						Animation* animation = LoadAnimation(animId);

						acomp->RegisterAnimation(Core::Collections::HashString(animation->Name().c_str()), animation);
					}

					so->AddComponent(acomp);
				}
				break;
			default:
				Logger::Fatal("Unsupported ComponentType");
			}
		}

		//Load Children
		std::vector<SceneObject*> children = LoadNode(&dataPointer, _world);

		for (auto it = children.begin(); it != children.end(); ++it)
		{
			so->AddChild(*it);
		}

		nodes.push_back(so);
	}

	*_buffer = dataPointer;
	return nodes;
}

Animation* AssetLoader::LoadAnimation(unsigned int _id)
{
	DictionaryEntry& entry = animationDictionary[_id];
	fseek(dataFile, entry.StartAddress(), SEEK_SET);

	char buffer[1024*100];
	assert(entry.Lenght() <= 1024*100);

	fread(&buffer, entry.Lenght(), 1, dataFile);
	char* bufferPointer = buffer;

	//Parse Animation
	unsigned short nameLength = *(unsigned short*)bufferPointer;
	bufferPointer += sizeof(unsigned short);

	char* name = new char[nameLength+1];
	memcpy(name, bufferPointer, nameLength);
	name[nameLength] = '\0';
	bufferPointer += nameLength;

	bool isLooped = *(bool*)bufferPointer;
	bufferPointer += sizeof(bool);

	unsigned char curveCount = *(unsigned char*)bufferPointer;
	bufferPointer += sizeof(unsigned char);

	//If CurveDatas are not loaded now -> load them
	if (curveCache[_id] == nullptr)
	{
		SharedPointer<CurveData>** thisCache = new SharedPointer<CurveData>*[curveCount];

		for (unsigned char i = 0; i < curveCount; i++)
		{
			CurveTarget target = *(CurveTarget*)bufferPointer;
			bufferPointer += sizeof(CurveTarget);

			unsigned short pointCount = *(unsigned short*)bufferPointer;
			bufferPointer += sizeof(unsigned short);

			CurvePoint* points = new CurvePoint[pointCount];

			for (unsigned short j = 0; j < pointCount; j++)
			{
				/*points[j] = CurvePoint(
					Vec2f(*(float*)bufferPointer, *(float*)(bufferPointer + sizeof(float))),
					Vec2f(*(float*)(bufferPointer + 2 * sizeof(float)), *(float*)(bufferPointer + 3 * sizeof(float))),
					Vec2f(*(float*)(bufferPointer + 4 * sizeof(float)), *(float*)(bufferPointer + 5 * sizeof(float))),
					*(InterpolationMode*)(bufferPointer + 6 * sizeof(float))
					);*/
				points[j] = *(CurvePoint*)bufferPointer;
				bufferPointer += 25;
			}

			CurveData* cdata = new CurveData(points, pointCount, target);
			SharedPointer<CurveData>* sp = new SharedPointer<CurveData>(cdata);
			thisCache[i] = sp;
			--(*sp->GetCounter());
		}

		curveCache[_id] = thisCache;
	}

	//Build animation from cached data
	Curve** curves = new Curve*[curveCount];
	for (unsigned short i = 0; i < curveCount; i++)
	{
		curves[i] = new Curve(*(curveCache[_id][i]));
	}

	Animation* animation = new Animation(std::string(name), curves, curveCount, isLooped);
	return animation;
}