#define DLL_EXPORT

#include "PngLoader.hpp"
#include "../../Graphic/Buffers/Texture2D.hpp"
#include "../../Graphic/GraphicContext.hpp"

#include <libpng\png.h>
#include <libpng\pnginfo.h>
#include <fstream>

#include "../../Util/Logger.hpp"

using namespace Engine;
using namespace Engine::Core::IO;
using namespace Engine::Util;
using namespace Engine::Graphic;
using namespace Engine::Graphic::Buffers;

Texture2D* PngLoader::Load(FILE* _file, Graphic::GraphicContext* _context, bool _isGammaCorrect)
{
	//Check png signature
	unsigned char header[8];
	fread(header, 1, 8, _file);

	if (!png_check_sig(header, 8))
	{
		Logger::Fatal("Png signature check failed.");
		return nullptr;
	}

	png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (!png_ptr)
	{
		Logger::Fatal("PNG out of memory");
		return nullptr;
	}

	png_infop info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr)
	{
		png_destroy_read_struct(&png_ptr, NULL, NULL);
		Logger::Fatal("PNG out of memory");
		return nullptr;
	}

	if (setjmp(png_jmpbuf(png_ptr)))
	{
		png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
		Logger::Fatal("PNG: setjmp failed");
		return nullptr;
	}

	png_init_io(png_ptr, _file);
	png_set_sig_bytes(png_ptr, 8);
	png_read_info(png_ptr, info_ptr);

	png_uint_32 height, width;
	int bit_depth, color_type;

	png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
		NULL, NULL, NULL);

	double gamma = 0.0, screengamma = 2.2;

	char* screengammastr;
	if ((screengammastr = getenv("SCREEN_GAMMA")) != NULL)
		screengamma = atof(screengammastr);

	if (!png_get_gAMA(png_ptr, info_ptr, &gamma))
		gamma = 1.0 / 2.2;

	png_set_gamma(png_ptr, screengamma, gamma);

	if (color_type == PNG_COLOR_TYPE_PALETTE)
		png_set_palette_to_rgb(png_ptr);
	if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
		png_set_expand_gray_1_2_4_to_8(png_ptr);
	if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
		png_set_tRNS_to_alpha(png_ptr);
	if (bit_depth == 16)
		png_set_strip_16(png_ptr);
	if (color_type == PNG_COLOR_TYPE_GRAY ||
		color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
		png_set_gray_to_rgb(png_ptr);

	png_read_update_info(png_ptr, info_ptr);

	//Read
	png_size_t rowbytes = png_get_rowbytes(png_ptr, info_ptr);
	int channels = (int)png_get_channels(png_ptr, info_ptr);

	unsigned char *image_data;
	png_bytepp row_pointers = nullptr;

	if ((image_data = (unsigned char*)malloc(rowbytes*height)) == nullptr) {
		png_destroy_read_struct(&png_ptr, &info_ptr, nullptr);
		return nullptr;
	}
	if ((row_pointers = (png_bytepp)malloc(height*sizeof(png_bytep))) == nullptr) {
		png_destroy_read_struct(&png_ptr, &info_ptr, nullptr);
		free(image_data);
		image_data = nullptr;
		return nullptr;
	}

	for (unsigned int i = 0; i < height; i++)
		row_pointers[i] = image_data + (height - i - 1) * rowbytes;

	png_read_image(png_ptr, row_pointers);

	free(row_pointers);

	png_read_end(png_ptr, NULL);

	//GLTexture erstellen
	Texture2D* texture;

	//Pixel Alignment setzen
	_context->UnpackAlignement().Set(1);

	if (info_ptr->channels == 4)
	{
		if (_isGammaCorrect)
			texture = new Texture2D(GL_SRGB_ALPHA, GL_RGBA, true);
		else
			texture = new Texture2D(GL_RGBA, GL_RGBA, true);
		texture->Bind(0);
		texture->BufferImage(info_ptr->width, info_ptr->height, GL_UNSIGNED_BYTE, image_data);
	}
	else if (info_ptr->channels == 3)
	{
		if (_isGammaCorrect)
			texture = new Texture2D(GL_SRGB, GL_RGB, true);
		else
			texture = new Texture2D(GL_RGB, GL_RGB, true);
		texture->Bind(0);
		texture->BufferImage(info_ptr->width, info_ptr->height, GL_UNSIGNED_BYTE, image_data);
	}
	else
		Logger::Fatal("Only RGB/RGBA format is supported");

	texture->SetFilter(GL_LINEAR_MIPMAP_LINEAR, GL_LINEAR);
	//texture->SetWrap(GL_REPEAT, GL_REPEAT, GL_REPEAT);
	texture->SetWrap(GL_CLAMP, GL_CLAMP, GL_CLAMP);

	free(image_data);
	png_destroy_read_struct(&png_ptr, &info_ptr, NULL);

	return texture;
}

Texture2D* PngLoader::LoadFile(std::string _path, Graphic::GraphicContext* _context, bool _isGammaCorrect)
{
	FILE* file = fopen(_path.c_str(), "rb");
	
	if (file != nullptr)
	{
		Texture2D* texture = Load(file, _context, _isGammaCorrect);
		fclose(file);
		return texture;
	}
	else
	{
		Logger::Fatal("Cannot open file: " + _path);
		return nullptr;
	}
}