#define DLL_EXPORT

#include "PoolAllocator.hpp"

#include <malloc.h>
#include <assert.h>

#include "../../Util/Logger.hpp"

using namespace Engine::Core::Memory;
using namespace Engine::Util;

PoolAllocator::PoolAllocator(unsigned int _maxElements, unsigned int _elementSize, Allocator& _allocator, unsigned int _alignment) :
					elementSize(_elementSize), maxElements(_maxElements), data(nullptr), allocator(_allocator)
{
	//This allocator does not work for elements smaller than sizeof(unsigned int).
	assert(_elementSize >= sizeof(unsigned int));

	//Elements will only be aligned correct if elementSize is a multiple of the alignment
	assert(_elementSize % _alignment == 0);

	//Request memory
	data = (byte*)allocator.Alloc(_elementSize * _maxElements, _alignment);
					
	if (data == nullptr)
	{
		Logger::Fatal("PoolAllocator failed to allocate memory");
	}

	//Set up free indices
	for (unsigned int i = 0; i < _maxElements; i++)
	{
		*(unsigned int*)(data + i * _elementSize) = i+1;
	}

	firstFree = 0;
}

PoolAllocator::~PoolAllocator()
{
	if (data != nullptr)
		allocator.Free(data);;
}

void PoolAllocator::Free(void* _data)
{
	if (_data > LastElement() || _data < data)
		Logger::Fatal("Free: Out of range");

	unsigned int freeId = IdFromPointer((byte*)_data);
	void* lastFree = &firstFree;

	while(NextFreeFromPointer(lastFree) < _data)
	{
		lastFree = NextFreeFromPointer(lastFree);
	}

	unsigned int x = *(unsigned int*)lastFree;

	*(unsigned int*)_data = *(unsigned int*)lastFree;
	*(unsigned int*)lastFree = freeId;
}

void PoolAllocator::FreeArray(void* _data)
{
	if (_data < PointerFromId(1))
		Logger::Fatal("Free: Out of range");

	byte* arrayStart = (byte*)_data - elementSize;
	unsigned int arrayLength = *(unsigned int*)arrayStart;

	if (IdFromPointer((byte*)_data) + arrayLength > maxElements)
		Logger::Fatal("Free: Out of range");

	void* lastFree = &firstFree;

	while(NextFreeFromPointer(lastFree) < arrayStart)
	{
		lastFree = NextFreeFromPointer(lastFree);
	}

	//Set last elements id to the target of the last free pointer before the array
	*(unsigned int*)(arrayStart + arrayLength * elementSize) = *(unsigned int*)lastFree;

	//Set last free before array to arrayStart
	*(unsigned int*)lastFree = IdFromPointer(arrayStart);

	//Set pointers inbetween
	for (unsigned int i = *(unsigned int*)lastFree; i < *(unsigned int*)lastFree + arrayLength; i++)
	{
		*(unsigned int*)(data + i * elementSize) = i+1;
	}
}

inline void* PoolAllocator::PointerFromId(unsigned int _id)
{
	return (void*)(data + _id * elementSize);
}
inline unsigned int PoolAllocator::IdFromPointer(byte* _pointer)
{
	return (unsigned int)(_pointer - data) / elementSize;
}
inline void* PoolAllocator::NextFreeFromPointer(void* _pointer)
{
	return (void*)(data + *(unsigned int*)_pointer * elementSize);
}
inline void* PoolAllocator::LastElement()
{
	return (void*)(data + elementSize * maxElements);
}

void* PoolAllocator::InternalAlloc(unsigned int _size, unsigned int _alignment)
{
	assert(_size == elementSize);

	if(firstFree > maxElements-1)
		return nullptr;
					
	//First free
	void* freeElementPointer = PointerFromId(firstFree);

	//Update pointer
	firstFree = *(unsigned int*)freeElementPointer;

	return freeElementPointer;
										
}
void* PoolAllocator::InternalAllocArray(unsigned int _elementSize, unsigned int _count, unsigned int _alignment) 
{
	assert(_elementSize == elementSize);

	//Search for _count + 1 continouse free memory blocks
	unsigned int freeFound = 1;
	byte* p = (byte*)PointerFromId(firstFree);
	byte* arrayStart = p;
	unsigned int* lastFree = &firstFree;

	while(freeFound < _count + 1 && p < data + maxElements * _elementSize)
	{
		if (NextFreeFromPointer(p) == p + elementSize)
			freeFound++;
		else
		{
			freeFound=1;
			lastFree = (unsigned int*)p;
			arrayStart = (byte*)NextFreeFromPointer(p);
		}

		p = (byte*)NextFreeFromPointer(p);
	}

	if (freeFound == _count + 1)
	{
		//Write length to first element
		*(unsigned int*)arrayStart = _count;

		//Set last pointer to first element to the next free element after the array
		*lastFree = *(unsigned int*)(arrayStart + _count * _elementSize);

		//Return pointer to the first data element in the array
		return (void*)(arrayStart + _elementSize);
	}
	else
		return nullptr;
}