#define DLL_EXPORT

#include "EngineManager.hpp"
#include "Util\Logger.hpp"
#include "Core\System\SystemConfiguration.hpp"
#include "Scene\Cameras\RenderCamera.hpp"
#include "Scene\SceneObject.hpp"
#include "Core\EventHandling\EventHandler.hpp"
#include "Core\Control\RenderControl.hpp"
#include "Core\System\WorkerThread.hpp"
#include "Graphic\FontRenderer.hpp"
#include "Scene\World.hpp"
#include "Scene\Components\CameraComponent.hpp"
#include "Graphic\GraphicContext.hpp"

#include <gl\glew.h>
#include <stdlib.h>
#include <time.h>

using namespace Engine;
using namespace Engine::Util;
using namespace Engine::Graphic;
using namespace Engine::Graphic::Shaders;
using namespace Engine::Scene;
using namespace Engine::Scene::Components;
using namespace Engine::Core::System;
using namespace Engine::Core::EventHandling;
using namespace Engine::Core::System::Tasks;

EngineManager::EngineManager(Scene::World* _world, Core::Control::RenderControl* _renderControl) :world(_world), renderControl(_renderControl), frameId(0),
	nextGuiFrame(0.0), frameTimeSW("FrameTime"), updateSW("Update"),
	frameStartCounter(0), finishedPositionUpdateCounter(0), finishedCullingCounter(0), finishedUpdateCounter(0), lastUpdateThreadCounter(0),
	frameStart(true), positionUpdateFinished(true), cullingFinished(true), updateFinished(true)
{
	srand((unsigned int)time(NULL));

	_renderControl->SizeChanged += new EventHandler<EngineManager, EventArgs>(this, &EngineManager::RenderControl_SizeChanged);

	workers = new WorkerThread*[SystemConfiguration::Get<unsigned int>("WorkerThreadCount")];

	for (unsigned int i = 0; i < SystemConfiguration::Get<unsigned int>("WorkerThreadCount"); ++i)
	{
		workers[i] = new WorkerThread(this);
		workers[i]->Start();
	}

	debugFontRenderer = new Graphic::FontRenderer(Core::System::SystemConfiguration::Get<std::string>("DefaultFontPath"), this, 12);
}

EngineManager::~EngineManager()
{
	for (unsigned int i = 0; i < SystemConfiguration::Get<unsigned int>("WorkerThreadCount"); ++i)
		workers[i]->Stop();

	StartWorker();

	for (unsigned int i = 0; i < SystemConfiguration::Get<unsigned int>("WorkerThreadCount"); ++i)
	{
		workers[i]->Join();
		delete workers[i];
	}
			
	delete workers;
	delete debugFontRenderer;
}

void EngineManager::RenderFrame()
{
	clock.Tick();
	frameId++;

	if (renderControl->ShowTimes() && clock.AbsoluteTime() > nextGuiFrame)
	{
		frameTimeSW.Start();
		updateSW.Start();
	}

	StartWorker();
}

void EngineManager::WaitForFrame()
{
	WaitWorker();

	for (auto it = world->Cameras().begin(); it != world->Cameras().end(); ++it)
	{
		if (*it != world->ScreenCamera())
			(*it)->Render();
	}

	world->ScreenCamera()->Render();	

	glFlush();

	//Debugoutput
	renderControl->Context()->BackBuffer().Bind();	

	if (renderControl->ShowTimes()) //Render debug information only if enabled
	{
		
		if (clock.AbsoluteTime() > nextGuiFrame)
		{
			frameTimeSW.Stop();
			screenCameraTime.str("");
			screenCameraTime << "Screencamera OpenGL/C++: " << ((Scene::Cameras::RenderCamera*)world->ScreenCamera()->Camera())->OpenGLTime();
			screenCameraTime << " / " << ((Scene::Cameras::RenderCamera*)world->ScreenCamera()->Camera())->CppTime();
		}

		std::string& asdf = frameTimeSW.Display();
		debugFontRenderer->RenderString(asdf, 5, 16, Math::Vec3f(0.8f, 0.8f, 0.8f));
		
		
		debugFontRenderer->RenderString(updateSW.Display(), 5, 30, Math::Vec3f(0.8f, 0.8f, 0.8f));
		
		debugFontRenderer->RenderString(screenCameraTime.str(), 5, 44, Math::Vec3f(0.8f, 0.8f, 0.8f));

		cullingCountString.str("");
		cullingCountString << "Visible Objects: " << ((Scene::Cameras::RenderCamera*)world->ScreenCamera()->Camera())->VisibleObjects();
		debugFontRenderer->RenderString(cullingCountString.str(), 5, 62, Math::Vec3f(0.8f, 0.8f, 0.8f));
	}

	
	if (Logger::HasMessages())
	{
		Math::Vec3f loggerMessageColor(0.8f, 0.8f, 0.8f);

		if (Logger::MaxMessageType() >= LoggerOutputType::WarningLoggerOutputTye)
			loggerMessageColor = Math::Vec3f(0.94f, 0.8f, 0.0f);
		if (Logger::MaxMessageType() >= LoggerOutputType::ErrorLoggerOutputTye)
			loggerMessageColor = Math::Vec3f(1.0f, 0.0f, 0.0f);

		debugFontRenderer->RenderString("Logger reported something! Look at the console", 5, 80, loggerMessageColor);
	}

	if (clock.AbsoluteTime() > nextGuiFrame)
	{
		nextGuiFrame += 1.0;
	}
}

inline Scene::World* EngineManager::World() const
{
	return world;
}

inline Core::Control::RenderControl* EngineManager::RenderControl() const
{
	return renderControl;
}

inline Util::Clock& EngineManager::Clock()
{
	return clock;
}

inline unsigned int EngineManager::FrameId() const
{
	return frameId;
}

void EngineManager::StartWorker()
{
	//Counter und Events rcksetzen
	finishedPositionUpdateCounter = 0;
	frameStartCounter = 0;
	lastUpdateThreadCounter = 0;
	finishedCullingCounter = 0;
	finishedUpdateCounter = 0;
	positionUpdateFinished.Reset();
	cullingFinished.Reset();
	updateFinished.Reset();

	//Frame starten
	frameStart.Set();
}

void EngineManager::WaitWorker()
{
	updateFinished.Wait();

	if (renderControl->ShowTimes() && clock.AbsoluteTime() > nextGuiFrame)
			updateSW.Stop();
}

void EngineManager::RenderControl_SizeChanged(void* _sender, Core::EventHandling::EventArgs& _args)
{
	world->Resize(renderControl->Width(), renderControl->Height());
}

inline Engine::Core::Concurrency::Atomic<unsigned int>& EngineManager::FrameStartCounter()
{
	return frameStartCounter;
}
inline Engine::Core::Concurrency::Atomic<unsigned int>& EngineManager::FinishedPositionUpdateCounter()
{
	return finishedPositionUpdateCounter;
}
inline Engine::Core::Concurrency::Atomic<unsigned int>& EngineManager::FinishedCullingCounter()
{
	return finishedCullingCounter;
}
inline Engine::Core::Concurrency::Atomic<unsigned int>& EngineManager::FinishedUpdateCounter()
{
	return finishedUpdateCounter;
}
inline Engine::Core::Concurrency::Atomic<unsigned int>& EngineManager::LastUpdateThreadCounter()
{
	return lastUpdateThreadCounter;
}
inline Engine::Core::Concurrency::Event& EngineManager::FrameStart()
{
	return frameStart;
}
inline Engine::Core::Concurrency::Event& EngineManager::PositionUpdateFinished()
{
	return positionUpdateFinished;
}
inline Engine::Core::Concurrency::Event& EngineManager::CullingFinished()
{
	return cullingFinished;
}
inline Engine::Core::Concurrency::Event& EngineManager::UpdateFinished()
{
	return updateFinished;
}

inline FontRenderer* EngineManager::DebugFontRenderer() const
{
	return debugFontRenderer;
}

inline bool EngineManager::IsDebugGUIFrame() const
{
	return (clock.AbsoluteTime() > nextGuiFrame);
}