#define DLL_EXPORT

#include "TonemapEffect.hpp"
#include "../BasicBodies/Plane.hpp"
#include "../GraphicContext.hpp"

#include "../../Core/System/SystemConfiguration.hpp"

using namespace Engine;
using namespace Engine::Graphic;
using namespace Engine::Graphic::Shaders;
using namespace Engine::Graphic::Buffers;
using namespace Engine::Graphic::FrameBuffers;
using namespace Engine::Graphic::Effects;
using namespace Engine::Core;
using namespace Engine::Core::System;

ShaderPtr* TonemapEffect::luminanceUpdateShaderPtr = nullptr;
ShaderPtr* TonemapEffect::tonemapShaderPtr = nullptr;

TonemapEffect::TonemapEffect(GraphicContext* _context, unsigned int _width, unsigned int _height, const Buffers::Texture2D& _inputTexture, const FrameBuffers::FrameBuffer& _renderBuffer) :
	inputTexture(_inputTexture), context(_context), targetBuffer(0), renderBuffer(_renderBuffer),
	adaptionSpeed(1.0f), maxLuminance(16.0f), middleGray(0.5f), mipmapLevel(0)
{
	if (luminanceUpdateShaderPtr == nullptr || luminanceUpdateShaderPtr->GetCounter() == nullptr)
	{
		if (luminanceUpdateShaderPtr != nullptr)
		{
			delete luminanceUpdateShaderPtr;
			delete tonemapShaderPtr;
		}

		Shader* shader = new Shader(GL_TRIANGLES, _context);
		shader->AttachShaderFile(SystemConfiguration::Get<std::string>("LuminanceUpdateVertexShader"), GL_VERTEX_SHADER);
		shader->AttachShaderFile(SystemConfiguration::Get<std::string>("LuminanceUpdateFragmenShader"), GL_FRAGMENT_SHADER);
		shader->BindAttribute("position", SHADER_POSITION);
		shader->Link();

		luminanceUpdateShaderPtr = new ShaderPtr(shader);
		--(*luminanceUpdateShaderPtr->GetCounter());

		shader = new Shader(GL_TRIANGLES, _context);
		shader->AttachShaderFile(SystemConfiguration::Get<std::string>("TonemapVertexShader"), GL_VERTEX_SHADER);
		shader->AttachShaderFile(SystemConfiguration::Get<std::string>("TonemapFragmenShader"), GL_FRAGMENT_SHADER);
		shader->BindAttribute("position", SHADER_POSITION);
		shader->Link();

		tonemapShaderPtr = new ShaderPtr(shader);
		--(*tonemapShaderPtr->GetCounter());
	}

	luminanceUpdateShader = *luminanceUpdateShaderPtr;
	tonemapShader = *tonemapShaderPtr;

	avgTextures[0] = new Texture2D(GL_R16F, GL_RED, false);
	avgTextures[1] = new Texture2D(GL_R16F, GL_RED, false);

	avgBuffers[0] = new FrameBuffer(_context, 1, 1);
	avgBuffers[0]->AttachTexture(avgTextures[0], GL_COLOR_ATTACHMENT0);
	avgBuffers[0]->Init();
	avgBuffers[1] = new FrameBuffer(_context, 1, 1);
	avgBuffers[1]->AttachTexture(avgTextures[1], GL_COLOR_ATTACHMENT0);
	avgBuffers[1]->Init();

	Resize(_width, _height);
}

TonemapEffect::~TonemapEffect()
{
	delete avgTextures[0];
	delete avgTextures[1];
	delete avgBuffers[0];
	delete avgBuffers[1];
}

void TonemapEffect::Render(float _frameTime)
{
	//Change read and write buffer
	int readBuffer = targetBuffer;
	targetBuffer = (targetBuffer + 1) % 2;

	context->DepthTest().Disable();
	context->Blend().Disable();
	
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	
	inputTexture.GenerateMipmaps();
	inputTexture.Bind(1);
	avgTextures[readBuffer]->Bind(0);



	//Compute new luminance average
	avgBuffers[targetBuffer]->Bind();

	luminanceUpdateShader->Activate();
	luminanceUpdateShader->UniformVariable<GLint>("mipmapLevel") = mipmapLevel;
	luminanceUpdateShader->UniformVariable<GLfloat>("adaptionSpeed") = adaptionSpeed;
	luminanceUpdateShader->UniformVariable<GLfloat>("frameTime") = _frameTime;
	luminanceUpdateShader->ChangeValues();
	context->ScreenPlane()->Draw(luminanceUpdateShader->PrimitiveType());

	//Tonemapping
	renderBuffer.Bind();
	
	tonemapShader->Activate();
	tonemapShader->UniformVariable<GLfloat>("middleGray") = middleGray;
	tonemapShader->UniformVariable<GLfloat>("maxLuminance") = maxLuminance;
	tonemapShader->ChangeValues();
	avgTextures[targetBuffer]->Bind(0);

	context->ScreenPlane()->Draw(tonemapShader->PrimitiveType());
}

void TonemapEffect::Resize(unsigned int _width, unsigned int _height)
{
	mipmapLevel  = (unsigned int)ceilf(logf((float)max(_width, _height)) / logf(2));
}