#define DLL_EXPORT

#include "FontRenderer.hpp"

#include "../EngineManager.hpp"
#include "../Core/EventHandling/EventHandler.hpp"
#include "../Util/Logger.hpp"
#include "../Core/Control/RenderControl.hpp"
#include "Shaders\Shader.hpp"
#include "GraphicContext.hpp"

#include <ft2build.h>
#include FT_FREETYPE_H

#define FIRSTCHARCODE 32
#define BUFFERLENGTH 10

using namespace Engine;
using namespace Engine::Graphic;
using namespace Engine::Graphic::Shaders;
using namespace Engine::Util;
using namespace Engine::Math;
using namespace Engine::Core::System;
using namespace Engine::Core::EventHandling;

FontRenderer::FontRenderer(const std::string _fontPath, EngineManager* _engineManager, unsigned int _maxCharSize) : charSize(_maxCharSize), context(_engineManager->RenderControl()->Context())
{
	_engineManager->RenderControl()->SizeChanged += new EventHandler<FontRenderer, EventArgs>(this, &FontRenderer::RenderControl_SizeChanged);

	//Texture anlegen und mit Glyphen fllen
	glGenTextures(1, &fontTextureHandle);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, fontTextureHandle);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_R8, _maxCharSize * 16, _maxCharSize * 16, 0, GL_RED, GL_UNSIGNED_BYTE, nullptr);

	//Load font
	FT_Library library;

	FT_Error error = FT_Init_FreeType( &library );
	if (error)
	{
		Logger::Fatal("Cannot init freetype2");
		system("PAUSE");
	}
	
	FT_Face face;
	error = FT_New_Face( library, _fontPath.c_str(), 0, &face );
	if ( error == FT_Err_Unknown_File_Format )
	{
		Logger::Fatal("Unknown font file format");
		system("PAUSE");
	}
	else if ( error )
	{
		Logger::Fatal("Error while loading the font file");
		system("PAUSE");
	}

	error = FT_Set_Char_Size(face, 0, 16*64, 96, 96);
	if ( error )
		Logger::Fatal("Set fontsize failed");

	 error = FT_Set_Pixel_Sizes(face, _maxCharSize, 0);

	 unsigned char* buffer = new unsigned char[_maxCharSize * _maxCharSize];

	 for (short charcodeCounter = 0; charcodeCounter < 256; charcodeCounter++)
	 {
		//char charcode = (char)charcodeCounter;
		error = FT_Load_Char(face, charcodeCounter + FIRSTCHARCODE, FT_LOAD_RENDER);
		if ( error )
			Logger::Fatal("Error while loading char");

		charLeft[charcodeCounter] = face->glyph->bitmap_left;
		charTop[charcodeCounter] = face->glyph->bitmap_top;
		advanceLeft[charcodeCounter] = face->glyph->advance.x >> 6;
		advanceTop[charcodeCounter] = face->glyph->advance.y >> 6;

		int textureY = charcodeCounter / 16;
		int textureX = charcodeCounter % 16;

		for (unsigned int x = 0; x < _maxCharSize * _maxCharSize; x++)
			buffer[x] = 0;

		unsigned int bmpWidth = face->glyph->bitmap.width, bmpHeight = face->glyph->bitmap.rows;

		if ((unsigned int)face->glyph->bitmap.width > _maxCharSize)
			bmpWidth = _maxCharSize;
		if ((unsigned int)face->glyph->bitmap.rows > _maxCharSize)
			bmpHeight = _maxCharSize;


		for (unsigned int y = 0; y < bmpHeight; y++)
			for (unsigned int x = 0; x < bmpWidth; x++)
				buffer[x+_maxCharSize * (_maxCharSize - bmpHeight + y)] = face->glyph->bitmap.buffer[x + face->glyph->bitmap.width * (face->glyph->bitmap.rows - y - 1)];

		//Load glyph to texture
		glTexSubImage2D(GL_TEXTURE_2D, 0, textureX * _maxCharSize, textureY * _maxCharSize, _maxCharSize, _maxCharSize, GL_RED, GL_UNSIGNED_BYTE, buffer);
	}

	delete[] buffer;

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_R, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	 projectionMatrix = Math::Mat4f::Ortho(0.0f, (float)_engineManager->RenderControl()->Width(), (float)_engineManager->RenderControl()->Height(), 0.0f, 0.0f, 100.0f);

	 glGenBuffers(1, &vbo);
	 glBindBuffer(GL_ARRAY_BUFFER, vbo);
	 glBufferData(GL_ARRAY_BUFFER, BUFFERLENGTH * sizeof(CharacterData), nullptr, GL_STREAM_DRAW);

	 glGenVertexArrays(1, &vao);
}

FontRenderer::~FontRenderer()
{
	glDeleteBuffers(1, &vbo);
	glDeleteVertexArrays(1, &vao);
	glDeleteTextures(1, &fontTextureHandle);
}

void FontRenderer::RenderString(std::string _text, unsigned int _left, unsigned int _top, const Math::Vec3f& _color)
{
	context->DepthTest().Disable();
	context->Blend().Enable(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	context->CullFace().Disable();

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, fontTextureHandle);

	context->DefaultFontShader()->Activate();
	context->DefaultFontShader()->UniformVariable<Math::Vec3f>("color") = _color;
	context->DefaultFontShader()->UniformVariable<Math::Mat4f>("projectionMatrix") = projectionMatrix;
	context->DefaultFontShader()->UniformVariable<GLint>("glyphTexture") = 0;
	context->DefaultFontShader()->UniformVariable<GLfloat>("charSize") = (float)charSize;
	context->DefaultFontShader()->ChangeValues();

	glBindVertexArray(vao);

	float left = (float)_left;
	float top = (float)_top;

	//Immer BUFFERLENGTH zeichen in den Buffer kopieren und zeichnen
	for (size_t rangeStart = 0; rangeStart < _text.size(); rangeStart+=BUFFERLENGTH)
	{
		size_t range = BUFFERLENGTH;
		if (range > _text.size() - rangeStart)
			range = _text.size() - rangeStart;

		//Copy to buffer
		glBindBuffer(GL_ARRAY_BUFFER, vbo);
		CharacterData* data = (CharacterData*)glMapBufferRange(GL_ARRAY_BUFFER, 0, BUFFERLENGTH*sizeof(CharacterData), GL_MAP_WRITE_BIT | GL_MAP_INVALIDATE_RANGE_BIT);

		for (size_t i = 0; i < range; i++)
		{
			unsigned int charIndex = _text.c_str()[rangeStart + i] - FIRSTCHARCODE;
			if (charIndex < 0 || charIndex > 256)
			{
				Logger::Error("Character out of range, the buffer may now be currupted");
				return;
			}

			float thisLeft = left + charLeft[charIndex];
			float thisTop = top - charTop[charIndex];
			
			data[i].position.X(thisLeft).Y(thisTop);
			data[i].charIndex = charIndex;
			
			left += advanceLeft[charIndex];
		}

		if (!glUnmapBuffer(GL_ARRAY_BUFFER))
			Logger::Fatal("Failed to unmap ARRAY_BUFFER");

		//Draw Buffer
		glEnableVertexAttribArray(SHADER_POSITION);
		glVertexAttribPointer(SHADER_POSITION, 2, GL_FLOAT, GL_FALSE, sizeof(CharacterData), (GLvoid*)0);
		glEnableVertexAttribArray(SHADER_FONT_CHARINDEX);
		glVertexAttribIPointer(SHADER_FONT_CHARINDEX, 1, GL_UNSIGNED_INT, sizeof(CharacterData), (GLvoid*)(2 * sizeof(float)));

		context->DrawArrays(context->DefaultFontShader()->PrimitiveType(), 0, (GLsizei)range);

		glBindBuffer(GL_ARRAY_BUFFER, 0);
	}

	glBindVertexArray(0);

	context->DepthTest().TestAndWrite();
	context->Blend().Disable();
}

void FontRenderer::RenderControl_SizeChanged(void* _sender, Core::EventHandling::EventArgs& _args)
{
	Core::Control::RenderControl* control = (Core::Control::RenderControl*)_sender;
	projectionMatrix = Math::Mat4f::Ortho(0.0f, (float)control->Width(), (float)control->Height(), 0.0f, 0.0f, 100.0f);
}