#define DLL_EXPORT

#include "FrameBuffer.hpp"

#include "../../Util/Logger.hpp"
#include "../../Util/Convert.hpp"
#include "../Buffers/Texture2D.hpp"
#include "../GraphicContext.hpp"

using namespace Engine;
using namespace Engine::Graphic;
using namespace Engine::Graphic::Buffers;
using namespace Engine::Graphic::FrameBuffers;
using namespace Engine::Util;

FrameBuffer::FrameBuffer(GraphicContext* _context, unsigned int _width, unsigned int _height) : width(_width), height(_height), context(_context), maxAttachment(0)
{
	glGenFramebuffers(1, &handle);
}

FrameBuffer::FrameBuffer(GraphicContext* _context, unsigned int _width, unsigned int _height, GLuint _handle) : width(_width), height(_height), handle(_handle), context(_context), maxAttachment(0)
{ }

FrameBuffer::~FrameBuffer()
{
	if (handle != 0)
		glDeleteFramebuffers(1, &handle);
}

void FrameBuffer::Bind() const
{
	glBindFramebuffer(GL_FRAMEBUFFER, handle);

	GLenum targetArray[] = { GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1, GL_COLOR_ATTACHMENT2, GL_COLOR_ATTACHMENT3, GL_COLOR_ATTACHMENT4, GL_COLOR_ATTACHMENT5, 
						     GL_COLOR_ATTACHMENT6, GL_COLOR_ATTACHMENT7, GL_COLOR_ATTACHMENT8, GL_COLOR_ATTACHMENT9, GL_COLOR_ATTACHMENT10, GL_COLOR_ATTACHMENT11 };

	if (maxAttachment > 0)
		glDrawBuffers(maxAttachment - GL_COLOR_ATTACHMENT0 + 1, &targetArray[0]);

	glViewport(0, 0, width, height);
}

void FrameBuffer::Unbind() const
{
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

void FrameBuffer::Init()
{
	glBindFramebuffer(GL_FRAMEBUFFER, handle);
	Resize(width, height);

	//Check if something went wrong
	GLenum result = glCheckFramebufferStatus(GL_FRAMEBUFFER);
	if (result != GL_FRAMEBUFFER_COMPLETE)
		Logger::Fatal(std::string("Failed to create framebuffer: ") + Convert::ToString(result));

	Unbind();
}

void FrameBuffer::Clear(GLenum _clearBits) const
{
	context->Clear(_clearBits);
}

void FrameBuffer::Resize(unsigned int _width, unsigned int _height)
{
	width = _width;
	height = _height;

	for (auto it = attachedTextures.begin(); it != attachedTextures.end(); ++it)
	{
		(*it)->Bind(0);
		(*it)->BufferImage(width, height);
	}
}

void FrameBuffer::AttachTexture(Buffers::Texture2D* _texture, GLenum _target)
{
	glBindFramebuffer(GL_FRAMEBUFFER, handle);

	_texture->Bind(0);
	glFramebufferTexture2D(GL_FRAMEBUFFER, _target, GL_TEXTURE_2D, _texture->Handle(), 0);

	attachedTextures.push_back(_texture);

	if (_target != GL_DEPTH_ATTACHMENT && _target != GL_STENCIL_ATTACHMENT && _target != GL_DEPTH_STENCIL_ATTACHMENT)
	{
		if (maxAttachment < _target)
			maxAttachment = _target;
	}
}

void FrameBuffer::Copy(const FrameBuffer& _targetBuffer) const
{
	glBindFramebuffer(GL_READ_FRAMEBUFFER, Handle());
	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, _targetBuffer.Handle());

	glBlitFramebuffer(0, 0, Width(), Height(), 0, 0, Width(), Height(), GL_COLOR_BUFFER_BIT, GL_NEAREST);
}

inline GLuint FrameBuffer::Handle() const
{
	return handle;
}

inline unsigned int FrameBuffer::Height() const
{
	return height;
}

inline unsigned int FrameBuffer::Width() const
{
	return width;
}