#define DLL_EXPORT

#include "OctreeGenerationStage.hpp"

#include "../../GraphicContext.hpp"
#include "../../../EngineManager.hpp"
#include "../../../Scene/World.hpp"
#include "../../../Scene/Components/RenderComponent.hpp"
#include "../../../Scene/Components/VoxelOctreeComponent.hpp"
#include "../../../Util/Logger.hpp"
#include "../../../Core/System/SystemConfiguration.hpp"
#include "../../../Scene/Components/LightComponents/LightComponent.hpp"
#include "../../../Scene/Lights/SpotLight.hpp"
#include "../OctreeData.hpp"
#include "../../BasicBodies/Plane.hpp"

#include <fstream>
#include <iomanip>

using namespace Engine;
using namespace Engine::Math;
using namespace Engine::Graphic;
using namespace Engine::Graphic::Buffers;
using namespace Engine::Graphic::GI;
using namespace Engine::Graphic::GI::Generation;
using namespace Engine::Scene;
using namespace Engine::Scene::Components;
using namespace Engine::Scene::Lights;
using namespace Engine::Util;
using namespace Engine::Core::System;

/* Layout for linkbuffer
 * 
 * Child			UINT32
 * Transparency		FLOAT32
 * +X				UINT32
 * -X				UINT32
 * +Y				UINT32
 * -Y				UINT32
 * +Z				UINT32
 * -Z				UINT32
 */

OctreeGenerationStage::OctreeGenerationStage(VoxelOctreeComponent* _component, EngineManager* _manager, GraphicContext* _context, unsigned int _entryCount)
	: component(_component), manager(_manager), context(_context), octreeShader(GL_POINTS, _context), useCounter(0),
	deferredBuffer(), ptrShader(GL_POINTS, _context),
	lightDepth(GL_DEPTH_COMPONENT24, GL_DEPTH_COMPONENT, false), lightBuffer(_context, 256, 256),
	injectionShader(GL_POINTS, _context), emptyBuffer(_context, 1024, 1024), emptyTexture(GL_RED, GL_RED, false)
{
	//Allocate data class for this stagechain
	component->Data(new OctreeData());
	component->Data()->Resize(_entryCount);

	Resize(_entryCount);

	octreeShader.AttachShaderFile(SystemConfiguration::Get<std::string>("OctreeInsertShaderPath"), GL_VERTEX_SHADER);
	octreeShader.Link();

	ptrShader.AttachShaderFile("../../../Data/OctreeShader/octree_genptr.vert", GL_VERTEX_SHADER);
	ptrShader.Link();

	glGenVertexArrays(1, &octreeVAO);

	lightBuffer.AttachTexture(&lightDepth, GL_DEPTH_ATTACHMENT);
	lightBuffer.Init();

	emptyBuffer.AttachTexture(&emptyTexture, GL_COLOR_ATTACHMENT0);
	emptyBuffer.Init();

	lightDepth.SetFilter(GL_LINEAR, GL_LINEAR);
	lightDepth.SetWrap(GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE);
	lightDepth.SetCompareFunction(GL_COMPARE_REF_TO_TEXTURE, GL_LESS);

	injectionShader.AttachShaderFile("../../../Data/OctreeShader/octree_inject_spotlight.vert", GL_VERTEX_SHADER);
	injectionShader.Link();
}

OctreeGenerationStage::~OctreeGenerationStage()
{
	glDeleteVertexArrays(1, &octreeVAO);
}

void OctreeGenerationStage::Process()
{
	glFlush();
	glMemoryBarrier(GL_ALL_BARRIER_BITS);
	OctreeData* data = (OctreeData*)component->Data();

#ifdef _ENGINE_OCTREEE_TIMERS
	totalWatch.Start();
	initWatch.Start();
#endif

	for (int i = 0; i < 10; i++)
		data->LevelBuffers(i).Set(0, 8, 0, 0);

	linkCounter.Set(0);
	deferredCounter.Set(0, 1, 0, 0);

	

	GLint leafLevel = OCTREE_LEAFLEVEL;
	unsigned int resolution = (unsigned int)pow(2, leafLevel - 1);
	glViewport(0, 0, resolution, resolution);

	//The solution with clearing would be cooler, but does not work fast on Geforce 5xx cards
	//data->LockBuffer().Clear(nullptr, GL_RED_INTEGER, GL_UNSIGNED_INT);
	//data->LinkBuffer().Clear(nullptr, GL_R8, GL_RED_INTEGER, GL_UNSIGNED_INT);
	//This is (in combination with some shader code) the workaround
	unsigned int clearValues[] = { 0 };
	data->LockBuffer().BufferSubData(clearValues, 0, sizeof(unsigned int)* 1);
	unsigned int linkClearValues[] = { 0, 0, 0, 0, 0, 0, 0 };
	data->LinkBuffer().BufferSubData(linkClearValues, 0, sizeof(unsigned int)* 7);

	data->LinkBuffer().Bind(1);
	data->LockBuffer().Bind(2, GL_READ_WRITE);
	deferredBuffer.Bind(3);

	linkCounter.Bind(0);
	deferredCounter.BindAtomicCounter(1);

	context->CullFace().Disable();
	context->DepthTest().Disable();


#ifdef _ENGINE_OCTREEE_TIMERS
	initWatch.Stop();
	voxelizeWatch.Start();
#endif

	///
	/// Voxelisation Stage
	///


	World* world = manager->World();
	auto renderList = world->RenderObjects();
	for (auto it = renderList.begin(); it != renderList.end(); ++it)
	{
		(*it)->RenderOctree(component->MatrixX(), component->MatrixY(), component->MatrixZ());
	}

	context->CullFace().EnableBackFaceCulling();
	context->DepthTest().DepthFunc(GL_LEQUAL);

	glMemoryBarrier(GL_SHADER_IMAGE_ACCESS_BARRIER_BIT | GL_ATOMIC_COUNTER_BARRIER_BIT | GL_SHADER_STORAGE_BARRIER_BIT);


	context->Rasterizer().Disable();

#ifdef _ENGINE_OCTREEE_TIMERS
	voxelizeWatch.Stop();
#endif


	///
	/// Sparse Voxel Octree Building
	///

	glBindVertexArray(octreeVAO);
	for (int i = 0; i < leafLevel; i++)
	{
		if (i >= 2 && i < leafLevel - 1) //Set ptr only when there can be external ptrs and not in the last layer
		{
			ptrShader.Activate();
			ptrShader.ChangeValues();
			data->LevelBuffers(i-2).BindIndirectDraw();

			context->DrawArraysIndirect(GL_POINTS, 0);
		}

#ifdef _ENGINE_OCTREEE_TIMERS
	octreeBuildWatches[i].Start();
#endif

		Buffer::Copy(linkCounter, data->LevelBuffers(i), 0, sizeof(unsigned int) * 2, sizeof(unsigned int));

		deferredCounter.BindIndirectDraw();
		data->LevelBuffers(i).BindAtomicCounter(2);

		octreeShader.Activate();
		octreeShader.UniformVariable<GLint>("leafLevel") = leafLevel;
		octreeShader.UniformVariable<GLint>("maxLevel") = i + 1;
		octreeShader.ChangeValues();

		context->DrawArraysIndirect(octreeShader.PrimitiveType(), 0);

		glMemoryBarrier(GL_SHADER_IMAGE_ACCESS_BARRIER_BIT | GL_ATOMIC_COUNTER_BARRIER_BIT | GL_SHADER_STORAGE_BARRIER_BIT);

#ifdef _ENGINE_OCTREEE_TIMERS
	octreeBuildWatches[i].Stop();
#endif
	}

	glBindVertexArray(0);

	context->Rasterizer().Enable();



	///
	/// Light Injection Stage
	///

	data->LinkBuffer().Bind(1);
	context->CullFace().EnableFrontFaceCulling();

	for (auto it = world->Lights().begin(); it != world->Lights().end(); ++it)
	{
		if ((*it)->Light()->GetType() == LightType::Spot)
		{
			SpotLight* light = (SpotLight*)(*it)->Light();

			//Render Photonmap
			lightBuffer.Bind();
			lightBuffer.Clear();
			context->DepthTest().TestAndWrite();

			for (auto it = world->RenderObjects().begin(); it != world->RenderObjects().end(); ++it)
			{
				(*it)->RenderNormals(light->ShadowMatrix());
			}

			context->DepthTest().Disable();
			context->Rasterizer().Disable();

			//Inject Photonmap-Pixels to Octree
			lightDepth.Bind(6);
			injectionShader.Activate();
			injectionShader.UniformVariable<Mat4f>("shadowMatrix") = light->ShadowMatrix();
			injectionShader.UniformVariable<Mat4f>("invOctreeMatrix") = component->InverseMatrixZ();
			injectionShader.UniformVariable<Vec3f>("lightPosition") = light->WorldPosition();
			injectionShader.UniformVariable<Vec3f>("lightDirection") = light->WorldDirection();
			injectionShader.UniformVariable<GLfloat>("lightAngle") = light->Angle();
			injectionShader.UniformVariable<Vec3f>("lightIntensity") = light->Intensity();
			injectionShader.ChangeValues();

			glBindVertexArray(octreeVAO);
			data->LevelBuffers(8).BindIndirectDraw();
			context->DrawArraysIndirect(injectionShader.PrimitiveType(), 0);
			glBindVertexArray(0);

			injectionShader.Deactivate();
			context->Rasterizer().Enable();
		}
	}

	context->CullFace().EnableBackFaceCulling();


#ifdef _ENGINE_OCTREEE_TIMERS
	totalWatch.Stop();
#endif

	//WriteToFile("./debugoutput.txt");
}

void OctreeGenerationStage::Resize(unsigned int _entryCount)
{
	deferredBuffer.BufferData(nullptr, _entryCount * sizeof(GLuint) * 7, GL_STATIC_DRAW);
}

void OctreeGenerationStage::WriteToFile(const std::string _path) const
{
	/*OctreeData* data = (OctreeData*)component->Data();
	std::ofstream file(_path);
	
	if (file.good())
	{
		
		GLint bufferCount = linkCounter.Get();
		GLint deferredBufferCount = deferredCounter.Get();

		file << "Counter=" << bufferCount << "\n";
		file << "DeferredCounter=" << deferredBufferCount << "\n\n";

		GLuint* lock = (GLuint*)data->LockBuffer().MapMemory(GL_READ_ONLY);
		GLuint* link = (GLuint*)data->LinkBuffer().MapMemory(GL_READ_ONLY);

		for (int i = 0; i < (1 + bufferCount * 8); i++)
		{
			if ((i-1) % 8 == 0)
				file << "\n";

			file << "Line=" << std::setw(4) << i <<": ";
			for (int j = 0; j < 7; j++)
				file << link[i*7+j] << ", ";
			file << " [Lock=" << lock[i] << "]" << std::endl;
		}

		file.close();

		data->LockBuffer().UnMapMemory();
		data->LinkBuffer().UnMapMemory();
		DebugBreak();
	}
	else
		Logger::Error("Cannot open file" + _path);
	*/
}


#ifdef _ENGINE_OCTREEE_TIMERS

inline GLStopWatch& OctreeGenerationStage::TotalWatch()
{
	return totalWatch;
}

inline GLStopWatch& OctreeGenerationStage::InitWatch()
{
	return initWatch;
}

inline GLStopWatch& OctreeGenerationStage::VoxelizeWatch()
{
	return voxelizeWatch;
}

inline GLStopWatch* OctreeGenerationStage::OctreeBuildWatches()
{
	return octreeBuildWatches;
}

#endif