#define DLL_EXPORT

#include "IndirectIlluminationStage.hpp"
#include "../../../Scene/Components/VoxelOctreeComponent.hpp"
#include "../../../Scene/World.hpp"
#include "../../../Scene/SceneObject.hpp"
#include "../../GraphicContext.hpp"
#include "../../../Scene/Components/RenderComponent.hpp"
#include "../../BasicBodies/Plane.hpp"
#include "../../../Scene/Components/LightComponents/LightComponent.hpp"
#include "../../../Scene/Lights/SpotLight.hpp"
#include "../Generation/OctreeGenerationStage.hpp"
#include "../OctreeData.hpp"

using namespace Engine;
using namespace Engine::Math;
using namespace Engine::Graphic;
using namespace Engine::Graphic::GI;
using namespace Engine::Graphic::GI::Illumination;
using namespace Engine::Scene;
using namespace Engine::Scene::Components;
using namespace Engine::Scene::Lights;

IndirectIlluminationStage::IndirectIlluminationStage(VoxelOctreeComponent* _component, GraphicContext* _context) : context(_context), component(_component),
lightShader(GL_TRIANGLES, _context), lightDepth(GL_DEPTH_COMPONENT24, GL_DEPTH_COMPONENT, false), lightBuffer(_context, 1024, 1024), indirectShader(GL_TRIANGLES, _context),
diffuseIllumination(GL_RGB16F, GL_RGB, false), specularIllumination(GL_RGB16F, GL_RGB, false), lightAccumulationBuffer(_context, 0, 0)
{
	lightShader.AttachShaderFile("../../../Data/SystemShader/quad.vert", GL_VERTEX_SHADER);
	lightShader.AttachShaderFile("../../../Data/OctreeShader/directIllumination.frag", GL_FRAGMENT_SHADER);
	lightShader.Link();

	lightBuffer.AttachTexture(&lightDepth, GL_DEPTH_ATTACHMENT);
	lightBuffer.Init();

	lightDepth.SetFilter(GL_LINEAR, GL_LINEAR);
	lightDepth.SetWrap(GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE);
	lightDepth.SetCompareFunction(GL_COMPARE_REF_TO_TEXTURE, GL_LESS);

	indirectShader.AttachShaderFile("../../../Data/OctreeShader/octree_ambient_occlusion.vert", GL_VERTEX_SHADER);
	indirectShader.AttachShaderFile("../../../Data/OctreeShader/indirectIllumination.frag", GL_FRAGMENT_SHADER);
	indirectShader.Link();
}

IndirectIlluminationStage::~IndirectIlluminationStage()
{ }

void IndirectIlluminationStage::Process(const Math::Mat4f& _inverseCameraMatrix, const Math::Vec3f& _cameraPosition, const FrameBuffers::FrameBuffer& _illuminationBuffer)
{
	World* world = component->SceneObject()->World();

	//Direct Illumination

	for (auto it = world->Lights().begin(); it != world->Lights().end(); ++it)
	{
		if ((*it)->Light()->GetType() == LightType::Spot)
		{
			context->DepthTest().TestAndWrite();
			lightBuffer.Bind();
			lightBuffer.Clear();

			context->CullFace().EnableFrontFaceCulling();

			SpotLight* light = (SpotLight*)(*it)->Light();
			for (auto it = world->RenderObjects().begin(); it != world->RenderObjects().end(); ++it)
			{
				(*it)->RenderNormals(light->ShadowMatrix());
			}

			context->CullFace().EnableBackFaceCulling();

			//Do deferred shading for direct illumination
			context->Blend().Enable(GL_ONE, GL_ONE);
			context->DepthTest().Disable();
			
			_illuminationBuffer.Bind();

			lightDepth.Bind(6);
			lightShader.Activate();
			lightShader.UniformVariable<Mat4f>("inverseCameraMatrix") = _inverseCameraMatrix;
			lightShader.UniformVariable<Mat4f>("lightMatrix") = light->ShadowMatrix();
			lightShader.UniformVariable<Vec3f>("lightPosition") = light->WorldPosition().GetVec3f();
			lightShader.UniformVariable<GLfloat>("lightAngle") = light->Angle();
			lightShader.UniformVariable<Vec3f>("lightDirection") = light->WorldDirection();
			lightShader.UniformVariable<Vec3f>("lightIntensity") = light->Intensity();
			lightShader.UniformVariable<Vec3f>("cameraPosition") = _cameraPosition;
			lightShader.ChangeValues();

			context->ScreenPlane()->Draw(GL_TRIANGLES);

			context->Blend().Disable();
		}
	}

	//Deferred shading in the octree (indirect illumination)
	_illuminationBuffer.Bind();
	context->Blend().Enable(GL_ONE, GL_ONE);

	OctreeData* data = (OctreeData*)component->Data();
	data->LinkBuffer().Bind(1);
	data->LockBuffer().Bind(2, GL_READ_ONLY);
	data->BrickTexture().Bind(0);

	indirectShader.Activate();
	indirectShader.UniformVariable<Mat4f>("inverseCameraMatrix") = _inverseCameraMatrix;
	indirectShader.UniformArrays<Math::Mat4f>("cameraMatrixArray")[2] = component->MatrixZ();
	indirectShader.UniformVariable<GLint>("steps") = 10;
	indirectShader.UniformVariable<Vec3f>("cameraPosition") = _cameraPosition;
	indirectShader.ChangeValues();

	context->ScreenPlane()->Draw(indirectShader.PrimitiveType());

	indirectShader.Deactivate();
	context->Blend().Disable();
}