#define DLL_EXPORT

#include "OctreePreparationStage.hpp"
#include <cmath>
#include "../../../EngineManager.hpp"
#include "../../../Core/System/SystemConfiguration.hpp"
#include "../../GraphicContext.hpp"
#include "../../../Core/Control/RenderControl.hpp"
#include "../../../Core/IO/KeyboardInput.hpp"
#include "../OctreeData.hpp"
#include "../../../Scene/Components/VoxelOctreeComponent.hpp"
#include <fstream>

using namespace Engine;
using namespace Engine::Graphic;
using namespace Engine::Graphic::GI;
using namespace Engine::Graphic::GI::Preparation;
using namespace Engine::Core::System;
using namespace Engine::Core::IO;

OctreePreparationStage::OctreePreparationStage(Scene::Components::VoxelOctreeComponent* _component, EngineManager* _manager, GraphicContext* _context) :
	mipmapShader(GL_POINTS, _context), context(_context), transferShader(GL_POINTS, _context), doTransfer(true), manager(_manager), component(_component)
{
	mipmapShader.AttachShaderFile(SystemConfiguration::Get<std::string>("OctreeMipmapShaderPath"), GL_VERTEX_SHADER);
	mipmapShader.Link();

	transferShader.AttachShaderFile(SystemConfiguration::Get<std::string>("OctreeTransferShaderPath"), GL_VERTEX_SHADER);
	transferShader.Link();

	glGenVertexArrays(1, &octreeVAO);
}

OctreePreparationStage::~OctreePreparationStage()
{
	glDeleteVertexArrays(1, &octreeVAO);
}

void OctreePreparationStage::Process()
{
	OctreeData* data = (OctreeData*)component->Data();

	#ifdef _ENGINE_OCTREEE_TIMERS
	totalWatch.Start();
	#endif

	context->Rasterizer().Disable();
	glBindVertexArray(octreeVAO);

	for (int level = 8; level >= 1; level --)
	{
		data->LevelBuffers(level-1).BindIndirectDraw();

		#ifdef _ENGINE_OCTREEE_TIMERS
		mipmapWatches[level-1].Start();
		#endif

		//Mipmap from previouse stage
		mipmapShader.Activate();

		data->LinkBuffer().Bind(1);
		data->BrickTexture().BindImage(0);

		mipmapShader.UniformVariable<GLint>("level") = level;
		mipmapShader.ChangeValues();
		
		context->DrawArraysIndirect(GL_POINTS, 0);

		glMemoryBarrier(GL_SHADER_IMAGE_ACCESS_BARRIER_BIT);

		#ifdef _ENGINE_OCTREEE_TIMERS
		mipmapWatches[level-1].Stop();
		transferWatches[level-1].Start();
		#endif

		
		//Transfer data to adjacent bricks
		if (doTransfer)
		{
			transferShader.Activate();

			for (int i = 0; i < 3; i++)
			{
				transferShader.UniformVariable<GLint>("direction") = i;
				transferShader.ChangeValues();

				context->DrawArraysIndirect(GL_POINTS, 0);
				glMemoryBarrier(GL_SHADER_IMAGE_ACCESS_BARRIER_BIT);
			}
		}

		#ifdef _ENGINE_OCTREEE_TIMERS
		transferWatches[level-1].Stop();
		#endif
	}		

	glBindVertexArray(0);
	context->Rasterizer().Enable();

	glMemoryBarrier(GL_SHADER_IMAGE_ACCESS_BARRIER_BIT);

	//WriteToFile("./texture3d.txt");

	if (manager->RenderControl()->Keyboard().IsKeyDown(Keys::F11))
		doTransfer = !doTransfer;

	#ifdef _ENGINE_OCTREEE_TIMERS
		totalWatch.Stop();
	#endif
}

void OctreePreparationStage::WriteToFile(const std::string _path) const
{
	OctreeData* data = (OctreeData*)component->Data();

	GLint w, h, d;
	data->BrickTexture().Bind(0);

	glGetTexLevelParameteriv(GL_TEXTURE_3D, 0, GL_TEXTURE_WIDTH, &w);
	glGetTexLevelParameteriv(GL_TEXTURE_3D, 0, GL_TEXTURE_HEIGHT, &h);
	glGetTexLevelParameteriv(GL_TEXTURE_3D, 0, GL_TEXTURE_DEPTH, &d);

	float* values = new float[w*h*d*4];

	
	glGetTexImage(GL_TEXTURE_3D, 0, GL_RED, GL_FLOAT, values);

	std::ofstream file(".\\texture3d.txt");

	for (int z = 0; z < d; z++)
	{
		file  << "Layer "  << z << std::endl;

		for (int y = 0; y < h; y++)
		{
			file << y << ": ";

			for (int x = 0; x < w; x++)
			{
				file << values[(z*(w*h) + y * w + x) * 4] << " ";
			}
			file << std::endl;
		}

		file << std::endl;
	}

	file.close();

	delete values;
}

#ifdef _ENGINE_OCTREEE_TIMERS

Util::GLStopWatch& OctreePreparationStage::TotalWatch()
{
	return totalWatch;
}
Util::GLStopWatch* OctreePreparationStage::MipmapWatches()
{
	return mipmapWatches;
}
Util::GLStopWatch* OctreePreparationStage::TransferWatches()
{
	return transferWatches;
}

#endif