#define DLL_EXPORT

#include "OctreeDebugRenderStage.hpp"
#include "../OctreeData.hpp"
#include "../Generation/OctreeGenerationStage.hpp"
#include "../Preparation/OctreePreparationStage.hpp"
#include "../Illumination/AmbientOcclusionIlluminationStage.hpp"
#include "../../GraphicContext.hpp"
#include "../../FontRenderer.hpp"
#include "../../../Scene/Components/VoxelOctreeComponent.hpp"
#include "../../../EngineManager.hpp"
#include "../../../Core/Control/RenderControl.hpp"
#include "../../../Core/IO/KeyboardInput.hpp"

#include <iomanip>

using namespace Engine;
using namespace Engine::Math;
using namespace Engine::Graphic;
using namespace Engine::Graphic::GI;
using namespace Engine::Graphic::GI::Rendering;
using namespace Engine::Graphic::GI::Generation;
using namespace Engine::Graphic::GI::Preparation;
using namespace Engine::Graphic::GI::Illumination;
using namespace Engine::Scene;
using namespace Engine::Scene::Components;
using namespace Engine::Core::IO;

#define PointsPerCall

OctreeDebugRenderStage::OctreeDebugRenderStage(VoxelOctreeComponent* _component, GraphicContext* _context, EngineManager* _manager) : component(_component),
	context(_context), level(9), shader(GL_LINES, _context), manager(_manager), showVoxels(false)
#ifdef _ENGINE_OCTREEE_TIMERS
	, detailComponent(0)
#endif
{
	shader.AttachShaderFile("../../../Data/OctreeShader/octree_render_outlines.vert", GL_VERTEX_SHADER);
	shader.AttachShaderFile("../../../Data/OctreeShader/octree_render_outlines.geom", GL_GEOMETRY_SHADER);
	shader.AttachShaderFile("../../../Data/OctreeShader/octree_render_outlines.frag", GL_FRAGMENT_SHADER);
	shader.Link();

	glGenVertexArrays(1, &vao);
}
OctreeDebugRenderStage::~OctreeDebugRenderStage()
{
	glDeleteVertexArrays(1, &vao);
}

void OctreeDebugRenderStage::Process(const Math::Mat4f& _cameraMatrix)
{
#ifdef _ENGINE_OCTREEE_TIMERS
	totalWatch.Start();
#endif

	OctreeData* data = (OctreeData*)component->Data();
	OctreeGenerationStage* generationStage = (OctreeGenerationStage*)component->GenerationStage();
	OctreePreparationStage* preparationStage = (OctreePreparationStage*)component->PreparationStage();
	AmbientOcclusionIlluminationStage* illuminationStage = (AmbientOcclusionIlluminationStage*)component->IlluminationStage();

	if (showVoxels)
	{
		unsigned int numberVoxels = (unsigned int)pow(pow(2, level), 3);

		data->LevelBuffers(level - 1).BindIndirectDraw();

		data->LinkBuffer().Bind(1);
		data->LockBuffer().Bind(2, GL_READ_ONLY);
		data->BrickTexture().Bind(3);

		shader.Activate();
		shader.UniformVariable<Vec3f>("color") = Vec3f(0.0f, 1.0f, 0.0f);
		shader.UniformVariable<GLuint>("level") = level;
		shader.UniformVariable<GLuint>("leafLevel") = 9;
		shader.UniformVariable<Mat4f>("cameraMatrix") = _cameraMatrix;
		shader.UniformVariable<Mat4f>("invGIMatrix") = component->InverseMatrixZ();
		shader.UniformVariable<GLint>("brickTexture") = 3;
		shader.ChangeValues();

		glBindVertexArray(vao);
		context->DrawArraysIndirect(GL_POINTS, 0);
		glBindVertexArray(0);

		if (manager->RenderControl()->Keyboard().IsKeyDown(Keys::Add) && level <= 8)
			level++;
		if (manager->RenderControl()->Keyboard().IsKeyDown(Keys::Subtract) && level >= 2)
			level--;
	}

#ifdef _ENGINE_OCTREEE_TIMERS
	totalWatch.Stop();
#endif


#ifdef _ENGINE_OCTREEE_TIMERS
	//Output timer query results
	unsigned int lineh = 16;
	unsigned int xoffset = 250;
	unsigned int height = manager->RenderControl()->Height();
	Vec3f color(0.6f, 0.6f, 0.6f);

	if (manager->IsDebugGUIFrame())
	{
		cameraTotal.str("");
		cameraTotal		 << "Total C++/GL: " << std::setw(8) << std::fixed << std::setprecision(5) << component->CppTime() << " / " <<
			std::setw(8) << std::fixed << std::setprecision(5)  << component->OpenGLTime();

		renderTotal.str("");
		renderTotal		 << "Debug Output: " << std::setw(8) << std::fixed << std::setprecision(5) << generationStage->VoxelizeWatch().Time();

		generationTotal.str("");
		generationTotal	 << "Generation:   " << std::setw(8) << std::fixed << std::setprecision(5) << generationStage->TotalWatch().Time();
		generationInit.str("");
		generationInit				 << "Init and Clear: " << std::setw(8) << std::fixed << std::setprecision(5) << generationStage->InitWatch().Time();
		generationVoxelize.str("");
		generationVoxelize			 << "Voxelize:       " << std::setw(8) << std::fixed << std::setprecision(5) << generationStage->VoxelizeWatch().Time();

		for (int i = 0; i < 10; i++)
		{
			generationBuildOctree[i].str("");
			generationBuildOctree[i] << "Build [Level=" << i << "]: " << std::setw(8) << std::fixed << std::setprecision(5) << generationStage->OctreeBuildWatches()[i].Time();
		}

		preparationTotal.str("");
		preparationTotal << "Preparation:  " << std::setw(8) << std::fixed << std::setprecision(5) << preparationStage->TotalWatch().Time();

		for (int i = 0; i < 8; i++)
		{
			preparationMipmap[i].str("");
			preparationMipmap[i]   << "Mipmap   [Level=" << i + 1 << "]: " << std::setw(8) << std::fixed << std::setprecision(5) << preparationStage->MipmapWatches()[i].Time();
			preparationTransfer[i].str("");
			preparationTransfer[i] << "Transfer [Level=" << i + 1 << "]: " << std::setw(8) << std::fixed << std::setprecision(5) << preparationStage->TransferWatches()[i].Time();
		}

		illuminationTotal.str("");
		illuminationTotal << "Illumination: " << std::setw(8) << std::fixed << std::setprecision(5) << illuminationStage->TotalWatch().Time();	
	}
	
	manager->DebugFontRenderer()->RenderString("Octree GI (Details with F1)", 10, height - 6 * lineh, Math::Vec3f(0.8f, 0.8f, 1.0f));
	manager->DebugFontRenderer()->RenderString(cameraTotal.str(), 10, height - 5 * lineh, color);
	manager->DebugFontRenderer()->RenderString(generationTotal.str(), 10, height - 4 * lineh, color);
	manager->DebugFontRenderer()->RenderString(preparationTotal.str(), 10, height - 3 * lineh, color);
	manager->DebugFontRenderer()->RenderString(illuminationTotal.str(), 10, height - 2 * lineh, color);
	manager->DebugFontRenderer()->RenderString(renderTotal.str(), 10, height - 1 * lineh, color);

	if (detailComponent == 1)
	{
		manager->DebugFontRenderer()->RenderString("Generation Stage Details", xoffset, height - 14 * lineh, Math::Vec3f(0.8f, 0.8f, 1.0f));
		manager->DebugFontRenderer()->RenderString(generationInit.str(), xoffset, height - 13 * lineh, color);
		manager->DebugFontRenderer()->RenderString(generationVoxelize.str(), xoffset, height - 12 * lineh, color);
		for (int i = 0; i < 10; i++)
			manager->DebugFontRenderer()->RenderString(generationBuildOctree[i].str(), xoffset, height - (11 - i) * lineh, color);
	}
	else if (detailComponent == 2)
	{
		manager->DebugFontRenderer()->RenderString("Preparation Stage Details", xoffset, height - 20 * lineh, Math::Vec3f(0.8f, 0.8f, 1.0f));

		for (int i = 0; i < 9; i++)
		{
			manager->DebugFontRenderer()->RenderString(preparationMipmap[i].str(), xoffset, height - (19 - 2 * i) * lineh, color);
			manager->DebugFontRenderer()->RenderString(preparationTransfer[i].str(), xoffset, height - (19 - 2 * i - 1) * lineh, color);
		}
	}

	if (manager->RenderControl()->Keyboard().IsKeyDown(Keys::F1))
		detailComponent = (detailComponent + 1) % 3;

#endif

	if (manager->RenderControl()->Keyboard().IsKeyDown(Keys::F3))
		showVoxels = !showVoxels;
}