#define DLL_EXPORT

#include "GraphicContext.hpp"

#include "../Core/System/SystemConfiguration.hpp"
#include "../Util/Convert.hpp"
#include "../Util/Logger.hpp"
#include "../Core/IO/PngLoader.hpp"
#include "BasicBodies\Plane.hpp"
#include "BasicBodies\Cone.hpp"
#include "BasicBodies\IcoSphere.hpp"
#include "BasicBodies\LineSphere.hpp"
#include "BasicBodies\LineCone.hpp"
#include "BasicBodies\LineCube.hpp"
#include "Shaders\Shader.hpp"
#include "IconRenderer.hpp"
#include "Buffers/Texture2D.hpp"
#include "../Util/Debugging/InternalDebugger.hpp"

#include <cassert>

using namespace Engine;
using namespace Engine::Graphic;
using namespace Engine::Graphic::Buffers;
using namespace Engine::Graphic::GraphicStates;
using namespace Engine::Graphic::Shaders;
using namespace Engine::Graphic::BasicBodies;
using namespace Engine::Graphic::DelayedChange;
using namespace Engine::Util;
using namespace Engine::Util::Debugging;
using namespace Engine::Core::System;
using namespace Engine::Math;

GraphicContext::GraphicContext(HWND _hwnd, unsigned int _width, unsigned int _height) : hwnd(_hwnd), hdc(nullptr), hrc(nullptr),
	blend(this), cullFace(this), depthTest(this), polygonOffset(this), stencilTest(this), currentShader(0), lineWidth(this), unpackAlignement(this), backBuffer(this, _width, _height, 0),
	guiProjectionMatrix(Mat4f::Ortho(0.0f, (float)_width, (float)_height, 0.0f, 0.0f, 100.0f)), rasterizer(this)
{
	//min. 2.1 Context erstellen (bentigt zum initialisieren)
	hdc = GetDC(_hwnd);

	//Pixelformat setzen
	PIXELFORMATDESCRIPTOR pfd;
	ZeroMemory(&pfd, sizeof(pfd));
	pfd.nSize = sizeof(pfd);
	pfd.nVersion = 1;
	pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL |
              PFD_DOUBLEBUFFER;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = 32;
	pfd.cDepthBits = 24;
	int iFormat = ChoosePixelFormat(hdc, &pfd);
	SetPixelFormat(hdc, iFormat, &pfd);

	HGLRC tempHrc = wglCreateContext(hdc);

	//Originalcontexte speichern
	oldHDC = wglGetCurrentDC();
	oldHRC = wglGetCurrentContext();

	//Rendercontext aktuell setzen
	wglMakeCurrent(hdc, tempHrc);

	//GLEW initialisieren
	GLenum result = glewInit();

	if (result != GLEW_OK)
	{
		Logger::Fatal("Glew initialization failed");
		Logger::InfoMessage((const char*)glewGetErrorString(result));
	}

	//Versuchen einen 4.0 Kontext zu erstellen
	int attributes[] = {
		WGL_CONTEXT_MAJOR_VERSION_ARB, SystemConfiguration::Get<int>("OpenGLMajorVersion"),
		WGL_CONTEXT_MINOR_VERSION_ARB, SystemConfiguration::Get<int>("OpenGLMinorVersion"),
		WGL_CONTEXT_PROFILE_MASK_ARB, WGL_CONTEXT_CORE_PROFILE_BIT_ARB,
		0, 0, //Placeholder for debug bits
		0
	};

	bool useDebugContext = SystemConfiguration::Get<bool>("OpenGLDebugContext");
	if (useDebugContext)
	{
		attributes[6] = WGL_CONTEXT_FLAGS_ARB;
		attributes[7] = WGL_CONTEXT_DEBUG_BIT_ARB;
	}

	if (wglewIsSupported("WGL_ARB_create_context") == 1)
	{
		int glVersion[2] = {-1, -1}; // Set some default values for the version

		hrc = wglCreateContextAttribsARB(hdc, NULL, attributes);
		wglMakeCurrent(NULL, NULL);
		wglDeleteContext(tempHrc);
		wglMakeCurrent(hdc, hrc);

		glGetIntegerv(GL_MAJOR_VERSION, &glVersion[0]); // Get back the OpenGL MAJOR version we are using
		glGetIntegerv(GL_MINOR_VERSION, &glVersion[1]); // Get back the OpenGL MAJOR version we are using

		if (glVersion[0] != SystemConfiguration::Get<int>("OpenGLMajorVersion") || glVersion[1] != SystemConfiguration::Get<int>("OpenGLMinorVersion"))
		{
			Logger::Fatal("Context creation failed, Version should be " + Convert::ToString(SystemConfiguration::Get<int>("OpenGLMajorVersion")) + "." + Convert::ToString(SystemConfiguration::Get<int>("OpenGLMinorVersion")) +
				"but is " + Convert::ToString(glVersion[0]) + "." + Convert::ToString(glVersion[1]));
			system("PAUSE");
			exit(-1);
		}
	}
	else
	{
		wglMakeCurrent(oldHDC, oldHRC);
		wglDeleteContext(tempHrc);
		Logger::Fatal("WGL_ARB_create_context not found");
	}

	//Set VSync
	wglSwapIntervalEXT(SystemConfiguration::Get<bool>("EnableVSync"));

	glEnable(GL_FRAMEBUFFER_SRGB);

	//Register DebugCallback
	if (useDebugContext)
	{
		glDebugMessageCallbackARB(&GraphicContext::DebugCallback, this);
		glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS_ARB);
	}

	//Daten laden
	plane = new Plane(this);
	unitSphere = new IcoSphere(this);
	unitCone = new Cone(this);
	
	lineCone = new BasicBodies::LineCone(this);
	lineSphere = new BasicBodies::LineSphere(this);
	lineCube = new BasicBodies::LineCube(this);

	/*directionalLightShader = new LightEffectShader(this);
	directionalLightShader->AttachShaderFile(SystemConfiguration::Get<std::string>("DirectionalLightVertexShader"), GL_VERTEX_SHADER);
	directionalLightShader->AttachShaderFile(SystemConfiguration::Get<std::string>("DirectionalLightFragmentShader"), GL_FRAGMENT_SHADER);
	directionalLightShader->Link();

	pointLightShader = new LightEffectShader(this);
	pointLightShader->AttachShaderFile(SystemConfiguration::Get<std::string>("PointLightVertexShader"), GL_VERTEX_SHADER);
	pointLightShader->AttachShaderFile(SystemConfiguration::Get<std::string>("PointLightFragmentShader"), GL_FRAGMENT_SHADER);
	pointLightShader->Link();

	spotLightShader = new LightEffectShader(this);
	spotLightShader->AttachShaderFile(SystemConfiguration::Get<std::string>("SpotLightVertexShader"), GL_VERTEX_SHADER);
	spotLightShader->AttachShaderFile(SystemConfiguration::Get<std::string>("SpotLightFragmentShader"), GL_FRAGMENT_SHADER);
	spotLightShader->Link();*/

	

	fontShader = new Shader(GL_POINTS, this);
	fontShader->AttachShaderFile(SystemConfiguration::Get<std::string>("FontVertexShader"), GL_VERTEX_SHADER);
	fontShader->AttachShaderFile(SystemConfiguration::Get<std::string>("FontGeometryShader"), GL_GEOMETRY_SHADER);
	fontShader->AttachShaderFile(SystemConfiguration::Get<std::string>("FontFragmentShader"), GL_FRAGMENT_SHADER);
	fontShader->Link();

	iconRenderer = new Graphic::IconRenderer(this, _width, _height);
	iconPointLight = Core::IO::PngLoader::LoadFile(SystemConfiguration::Get<std::string>("IconPointLight"), this);
	iconSpotLight = Core::IO::PngLoader::LoadFile(SystemConfiguration::Get<std::string>("IconSpotLight"), this);

	lineShader = new Shader(GL_LINES, this);
	lineShader->AttachShaderFile(SystemConfiguration::Get<std::string>("LineVertexShader"), GL_VERTEX_SHADER);
	lineShader->AttachShaderFile(SystemConfiguration::Get<std::string>("LineFragmentShader"), GL_FRAGMENT_SHADER);
	lineShader->Link();

	/*gaussShader = new Shaders::PostprocessingShader(this);
	gaussShader->AttachShaderFile(SystemConfiguration::Get<std::string>("GaussVertexShader"), GL_VERTEX_SHADER);
	gaussShader->AttachShaderFile(SystemConfiguration::Get<std::string>("GaussFragmenShader"), GL_FRAGMENT_SHADER);
	gaussShader->Link();

	adaptiveLumShader = new Shaders::PostprocessingShader(this);
	adaptiveLumShader->AttachShaderFile(SystemConfiguration::Get<std::string>("AdaptiveLumVertexShader"), GL_VERTEX_SHADER);
	adaptiveLumShader->AttachShaderFile(SystemConfiguration::Get<std::string>("AdaptiveLumFragmenShader"), GL_FRAGMENT_SHADER);
	adaptiveLumShader->Link();

	tonemapShader = new Shaders::PostprocessingShader(this);
	tonemapShader->AttachShaderFile(SystemConfiguration::Get<std::string>("TonemapVertexShader"), GL_VERTEX_SHADER);
	tonemapShader->AttachShaderFile(SystemConfiguration::Get<std::string>("TonemapFragmenShader"), GL_FRAGMENT_SHADER);
	tonemapShader->Link();

	fxaaShader = new Shaders::PostprocessingShader(this);
	fxaaShader->AttachShaderFile(SystemConfiguration::Get<std::string>("FxaaVertexShader"), GL_VERTEX_SHADER);
	fxaaShader->AttachShaderFile(SystemConfiguration::Get<std::string>("FxaaFragmenShader"), GL_FRAGMENT_SHADER);
	fxaaShader->Link();*/
}

GraphicContext::~GraphicContext()
{
	delete plane;
	delete unitSphere;
	delete lineCone;
	delete lineSphere;
	delete lineCube;

	/*delete pointLightShader;
	delete directionalLightShader;
	delete spotLightShader;*/
	delete fontShader;
	delete lineShader;
	/*delete gaussShader;
	delete adaptiveLumShader;
	delete tonemapShader;
	delete fxaaShader;*/

	delete iconRenderer;
	delete iconPointLight;
	delete iconSpotLight;

	if (hdc != nullptr && hrc != nullptr)
	{
		wglMakeCurrent(NULL, NULL);
		wglDeleteContext(hrc);
		ReleaseDC(hwnd, hdc);
	}
}

void GraphicContext::Resize(unsigned int _width, unsigned int _height)
{
	iconRenderer->Resize(_width, _height);
	backBuffer.Resize(_width, _height);
	guiProjectionMatrix = Mat4f::Ortho(0.0f, (float)_width, (float)_height, 0.0f, 0.0f, 100.0f);
}

void GraphicContext::SwapBuffer()
{
	this->ChangeValues();
	SwapBuffers(hdc);
	backBuffer.Bind();
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	Clear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
}

void GraphicContext::Activate()
{
	//Call this method only if it is REALLY necessary
	wglMakeCurrent(hdc, this->hrc);
}


#pragma region DrawMethods

void GraphicContext::DrawArrays(GLenum _mode, GLint _first, GLsizei _count)
{
	this->ChangeValues();
	glDrawArrays(_mode, _first, _count);
}

void GraphicContext::DrawElements(GLenum _mode, GLsizei _count, GLenum _type, const GLvoid* _indices)
{
	this->ChangeValues();
	glDrawElements(_mode, _count, _type, _indices);
}

void GraphicContext::DrawElementsInstanced(GLenum _mode, GLsizei _count, GLenum _type, const GLvoid* _indices, GLsizei _primcount)
{
	this->ChangeValues();
	glDrawElementsInstanced(_mode, _count, _type, _indices, _primcount);
}

void GraphicContext::DrawArraysInstanced(GLenum _mode,  GLint _first,  GLsizei _count,  GLsizei _primcount)
{
	this->ChangeValues();
	glDrawArraysInstanced(_mode, _first, _count, _primcount);
}

void GraphicContext::DrawArraysIndirect(GLenum _mode, const void *_indirect)
{
	this->ChangeValues();
	glDrawArraysIndirect(_mode, _indirect);
}

void GraphicContext::Clear(GLbitfield _mask)
{
	this->ChangeValues();
	glClear(_mask);
}

#pragma endregion


#pragma region States

inline BlendingState& GraphicContext::Blend()
{
	return blend;
}

inline FaceCullingState& GraphicContext::CullFace()
{
	return cullFace;
}

inline DepthTestState& GraphicContext::DepthTest()
{
	return depthTest;
}

inline LineWidthState& GraphicContext::LineWidth()
{
	return lineWidth;
}

inline PolygonOffsetState& GraphicContext::PolygonOffset()
{
	return polygonOffset;
}

inline StencilTestState& GraphicContext::StencilTest()
{
	return stencilTest;
}

inline GraphicStates::UnpackAlignmentState& GraphicContext::UnpackAlignement()
{
	return unpackAlignement;
}

inline GraphicStates::RasterizerState& GraphicContext::Rasterizer()
{
	return rasterizer;
}

#pragma endregion


#pragma region Preloaded Primitives and Shaders

const FrameBuffers::FrameBuffer& GraphicContext::BackBuffer() const
{
	return backBuffer;
}

BasicBodies::Plane* GraphicContext::ScreenPlane() const
{
	return plane;
}

BasicBodies::IcoSphere* GraphicContext::UnitSphere() const
{
	return unitSphere;
}

/*Shaders::LightEffectShader* GraphicContext::DirectionalLightShader() const
{
	return directionalLightShader;
}

Shaders::LightEffectShader* GraphicContext::PointLightShader() const
{
	return pointLightShader;
}

Shaders::LightEffectShader* GraphicContext::SpotLightShader() const
{
	return spotLightShader;
}*/

Shaders::Shader* GraphicContext::DefaultFontShader() const
{
	return fontShader;
}

/*Shaders::PostprocessingShader* GraphicContext::GaussShader() const
{
	return gaussShader;
}

Shaders::PostprocessingShader* GraphicContext::AdaptiveLumShader() const
{
	return adaptiveLumShader;
}

Shaders::PostprocessingShader* GraphicContext::TonemapShader() const
{
	return tonemapShader;
}

Shaders::PostprocessingShader* GraphicContext::FxaaShader() const
{
	return fxaaShader;
}*/

Graphic::IconRenderer* GraphicContext::IconRenderer() const
{
	return iconRenderer;
}

Texture2D* GraphicContext::IconPointLight() const
{
	return iconPointLight;
}

Texture2D* GraphicContext::IconSpotLight() const
{
	return iconSpotLight;
}

Shaders::Shader* GraphicContext::DebugLineShader() const
{
	return lineShader;
}

BasicBodies::LineCone* GraphicContext::LineCone() const
{
	return lineCone;
}

BasicBodies::LineSphere* GraphicContext::LineSphere() const
{
	return lineSphere;
}

BasicBodies::LineCube* GraphicContext::LineCube() const
{
	return lineCube;
}

BasicBodies::Cone* GraphicContext::UnitCone() const
{
	return unitCone;
}

inline const Mat4f& GraphicContext::GuiProjectionMatrix() const
{
	return guiProjectionMatrix;
}

#pragma endregion


bool GraphicContext::IsCurrentShader(const Shaders::Shader* _shader) const
{
	if (_shader != nullptr)
		return (_shader->Handle() == currentShader);
	else
		return (currentShader == 0);
}

void GraphicContext::BindShader(const Shaders::Shader* _shader)
{
	if (_shader != nullptr)
		currentShader = _shader->Handle();
	else
		currentShader = 0;
}

void APIENTRY GraphicContext::DebugCallback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const GLchar* message, GLvoid* userParam)
{
	std::string fullmsg = std::string(message);
	bool halt = false;
	
	switch (severity)
	{
	case GL_DEBUG_SEVERITY_HIGH:
		halt = true;
		Logger::Error(fullmsg);
		break;
	case GL_DEBUG_SEVERITY_MEDIUM:
	case GL_DEBUG_SEVERITY_LOW:
		Logger::Warning(fullmsg);
		break;
	}

	
	Logger::Debug(std::string("Source:   ") + Convert::ToString(source));
	Logger::Debug(std::string("Type:     ") + Convert::ToString(type));
	Logger::Debug(std::string("Id:       ") + Convert::ToString(id));
	Logger::Debug(std::string("Severity: ") + Convert::ToString(severity));

	InternalDebugger::PrintStackTrace();

	if (halt)
	{
		system("PAUSE");
	}
}