#define DLL_EXPORT

#include "Material.hpp"
#include "../../Util/Logger.hpp"

using namespace Engine;
using namespace Engine::Graphic;
using namespace Engine::Graphic::Materials;
using namespace Engine::Graphic::Shaders;
using namespace Engine::Util;

Material::Material(ShaderPtr _shader, ShaderPtr _forwardShader, ShaderPtr _normalShader,
				   ShaderPtr _octreeShader) :
	shader(_shader), forwardShader(_forwardShader), normalShader(_normalShader), octreeShader(_octreeShader), isShadeless(false), isTransparent(false)
{ }

Material::~Material()
{
	for (auto it = parts.begin(); it != parts.end(); ++it)
	{
		delete (*it);
	}
}

inline std::vector<IMaterialPart*>& Material::Parts()
{
	return parts;
}

inline ShaderPtr Material::Shader() const
{
	return shader;
}

inline void Material::Shader(ShaderPtr _shader)
{
	shader = _shader;
}

inline ShaderPtr Material::ForwardShader() const
{
	return forwardShader;
}

inline void Material::ForwardShader(ShaderPtr _shader)
{
	forwardShader = _shader;
}

inline ShaderPtr Material::NormalShader() const
{
	return normalShader;
}
inline void Material::NormalShader(ShaderPtr _shader)
{
	normalShader = _shader;
}

inline ShaderPtr Material::OctreeShader() const
{
	return octreeShader;
}
inline void Material::OctreeShader(ShaderPtr _shader)
{
	octreeShader = _shader;
}

void Material::StartActivate(ShadeMode _mode) const
{
	Shaders::Shader* sh;
	
	switch (_mode)
	{
	case ShadeMode::Normals:
		sh = normalShader.Get();
		break;
	case ShadeMode::Deferred:
		sh = shader.Get();
		break;
	case ShadeMode::Forward:
		sh = forwardShader.Get();
		break;
	case ShadeMode::Octree:
		sh = octreeShader.Get();
		break;
	}

	sh->Activate();
	sh->ResetTextureUnit(3); //0 and 1 are used for light Buffers

	Logger::DisableWarnings();

	for (auto it = parts.begin(); it != parts.end(); ++it)
	{
		(*it)->ToShader(sh);
	}

	sh->UniformVariable<GLboolean>("isShadeless") = isShadeless;
	sh->UniformVariable<GLboolean>("isTransparent") = isTransparent;

	Logger::EnableWarnings();
}

void Material::FinishActivate(ShadeMode _mode) const
{
	Shaders::Shader* sh;
	
	switch (_mode)
	{
	case ShadeMode::Normals:
		sh = normalShader.Get();
		break;
	case ShadeMode::Deferred:
		sh = shader.Get();
		break;
	case ShadeMode::Forward:
		sh = forwardShader.Get();
		break;
	case ShadeMode::Octree:
		sh = octreeShader.Get();
		break;
	}

	sh->ChangeValues();
}

void Material::Deactivate() const
{ }

inline bool Material::IsShadeless() const
{
	return isShadeless;
}

inline void Material::IsShadeless(bool _value)
{
	isShadeless = _value;
}

inline bool Material::IsTransparent() const
{
	return isTransparent;
}

inline void Material::IsTransparent(bool _value)
{
	isTransparent = _value;
}