#define DLL_EXPORT

#include "BSPTriangle.hpp"
#include "../Shaders/Shader.hpp"

using namespace Engine;
using namespace Engine::Math;
using namespace Engine::Graphic::Meshes;
using namespace Engine::Graphic::Shaders;
using namespace Engine::Scene::Boundings;

BSPTriangle::BSPTriangle(unsigned int _i1, unsigned int _i2, unsigned int _i3, const Math::Vec4f& _p1, const Math::Vec4f& _p2, const Math::Vec4f& _p3, const Math::Vec4f& _plane) : plane(_plane)
{
	indices[0] = _i1;
	indices[1] = _i2;
	indices[2] = _i3;

	positions[0] = _p1;
	positions[1] = _p2;
	positions[2] = _p3;
}
BSPTriangle::~BSPTriangle() { }

SplittingResult BSPTriangle::IntersectPlane(const Vec4f& _plane) const
{
	float a,b,c;
	return IntersectPlane(_plane, a, b, c);
}

SplittingResult BSPTriangle::IntersectPlane(const Vec4f& _plane, float& _d1, float& _d2, float& _d3) const
{
	_d1 = Vec4f::Dot(_plane, positions[0]);
	_d2 = Vec4f::Dot(_plane, positions[1]);
	_d3 = Vec4f::Dot(_plane, positions[2]);

	if (_d1 > -0.01f && _d2 > -0.01f && _d3 > -0.01f)
		return SplittingResult::Front;
	else if(_d1 < 0.01f && _d2 < 0.01f && _d3 < 0.01f)
		return SplittingResult::Back;
	else
		return SplittingResult::Split;
}

void BSPTriangle::Split(float _d1, float _d2, float _d3, std::vector<BSPTriangle>& _front, std::vector<BSPTriangle>& _back, std::vector<float*>& _tdata,
						const std::vector<VertexStreamElement>& _mapping, int _floatPerVertex)
{
	float* data1 = _tdata[indices[0]];
	float* data2 = _tdata[indices[1]];
	float* data3 = _tdata[indices[2]];

	float abs1, abs2, abs3;
	abs1 = fabs(_d1);
	abs2 = fabs(_d2);
	abs3 = fabs(_d3);

	if (sgn(_d1) != sgn(_d2)) //Can be case 2 or 3
	{
		if (sgn(_d1) != sgn(_d3)) //Case 3
		{
			//Interpolation on edges 12 and 13
			float* data12 = Interpolate(abs1, abs2, data1, data2, _mapping, _floatPerVertex);
			float* data13 = Interpolate(abs1, abs3, data1, data3, _mapping, _floatPerVertex);
			_tdata.push_back(data12);
			_tdata.push_back(data13);

			unsigned int i12 = (unsigned int)_tdata.size() - 2;
			unsigned int i13 = (unsigned int)_tdata.size() - 1;

			Vec4f p12 = Vec4f::Interpolate(positions[0], positions[1], abs1, abs2);
			Vec4f p13 = Vec4f::Interpolate(positions[0], positions[2], abs1, abs3);

			BSPTriangle tr1(indices[0], i12, i13, positions[0], p12, p13, plane);
			BSPTriangle tr2(i12, indices[1], i13, p12, positions[1], p13, plane);
			BSPTriangle tr3(indices[1], indices[2], i13, positions[1], positions[2], p13, plane);

			if (_d1 > 0.0f)
			{
				_front.push_back(tr1);
				_back.push_back(tr2);
				_back.push_back(tr3);
			}
			else
			{
				_back.push_back(tr1);
				_front.push_back(tr2);
				_front.push_back(tr3);
			}
		}
		else //Case 2
		{
			//Interpolation on edges 12 and 23
			float* data12 = Interpolate(abs1, abs2, data1, data2, _mapping, _floatPerVertex);
			float* data23 = Interpolate(abs2, abs3, data2, data3, _mapping, _floatPerVertex);
			_tdata.push_back(data12);
			_tdata.push_back(data23);

			unsigned int i12 = (unsigned int)_tdata.size() - 2;
			unsigned int i23 = (unsigned int)_tdata.size() - 1;

			Vec4f p12 = Vec4f::Interpolate(positions[0], positions[1], abs1, abs2);
			Vec4f p23 = Vec4f::Interpolate(positions[1], positions[2], abs2, abs3);

			BSPTriangle tr1(indices[0], i12, i23, positions[0], p12, p23, plane);
			BSPTriangle tr2(indices[0], i23, indices[2], positions[0], p23, positions[2], plane);
			BSPTriangle tr3(i12, indices[1], i23, p12, positions[1], p23, plane);

			if (_d1 > 0.0f)
			{
				_front.push_back(tr1);
				_front.push_back(tr2);
				_back.push_back(tr3);
			}
			else
			{
				_back.push_back(tr1);
				_back.push_back(tr2);
				_front.push_back(tr3);
			}
		}
	}
	else //Must be case 1
	{
		//Interpolation on edges 13 and 23
		float* data13 = Interpolate(abs1, abs3, data1, data3, _mapping, _floatPerVertex);
		float* data23 = Interpolate(abs2, abs3, data2, data3, _mapping, _floatPerVertex);
		_tdata.push_back(data13);
		_tdata.push_back(data23);

		unsigned int i13 = (unsigned int)_tdata.size() - 2;
		unsigned int i23 = (unsigned int)_tdata.size() - 1;

		Vec4f p13 = Vec4f::Interpolate(positions[0], positions[2], abs1, abs3);
		Vec4f p23 = Vec4f::Interpolate(positions[1], positions[2], abs2, abs3);

		BSPTriangle tr1(indices[0], indices[1], i13, positions[0], positions[1], p13, plane);
		BSPTriangle tr2(indices[1], i23, i13, positions[1], p23, p13, plane);
		BSPTriangle tr3(i13, i23, indices[2], p13, p23, positions[2], plane);

		if (_d1 > 0.0f)
		{
			_front.push_back(tr1);
			_front.push_back(tr2);
			_back.push_back(tr3);
		}
		else
		{
			_back.push_back(tr1);
			_back.push_back(tr2);
			_front.push_back(tr3);
		}
	}
}

float* BSPTriangle::Interpolate(float _d1, float _d2, float* _tdata1, float* _tdata2, const std::vector<VertexStreamElement>& _mapping, int _floatPerVertex)
{
	float* output = new float[_floatPerVertex];

	float d = _d1 + _d2;

	_d1 /= d;
	_d2 /= d;

	for (auto it = _mapping.begin(); it != _mapping.end(); ++it)
	{
		int foffset = it->Offset() / sizeof(float);
		
		switch (it->Size())
		{
		case 1:
			output[foffset] = _d1 * _tdata1[foffset] + _d2 * _tdata2[foffset];
			break;
		case 2:
			{
				Vec2f v2 = Vec2f::Interpolate(Vec2f(_tdata1 + foffset), Vec2f(_tdata2 + foffset), _d1, _d2);
				memcpy(output + foffset, v2.GetGlVector(), 2 * sizeof(float));
			}
			break;
		case 3:
			{
				Vec3f v3 = Vec3f::Interpolate(Vec3f(_tdata1 + foffset), Vec3f(_tdata2 + foffset), _d1, _d2);

				if (it->AttribHandle() == SHADER_MATERIAL_NORMAL)
					v3.Normalize();

				memcpy(output + foffset, v3.GetGlVector(), 3 * sizeof(float));
			}
			break;
		case 4:
			{
				Vec4f v4 = Vec4f::Interpolate(Vec4f(_tdata1 + foffset), Vec4f(_tdata2 + foffset), _d1, _d2);
				memcpy(output + foffset, v4.GetGlVector(), 4 * sizeof(float));
			}
			break;
		}
	}

	return output;
}

inline const unsigned int* BSPTriangle::Indices() const
{
	return indices;
}
inline const Vec4f* BSPTriangle::Positions() const
{
	return positions;
}
inline const Vec4f& BSPTriangle::Plane() const
{
	return plane;
}