#define DLL_EXPORT

#include "Shader.hpp"

#include "Uniform.hpp"
#include "UniformArray.hpp"
#include "../../Util/Logger.hpp"
#include "../../Util/Convert.hpp"
#include "../GraphicContext.hpp"
#include "../../Core/System/SystemConfiguration.hpp"

#include <cassert>
#include <fstream>
#include <regex>

using namespace std;
using namespace Engine::Graphic;
using namespace Engine::Graphic::Shaders;
using namespace Engine::Util;
using namespace Engine::Core::Collections;

Shader::Shader(GLenum _primitiveType, GraphicContext* _context) : handle(0), primitiveType(_primitiveType), context(_context)
{
	//Programm erstellen
	handle = glCreateProgram();

	if (handle == 0)
	{
		Logger::GLErrors();
		Logger::Fatal("Shader Program creation failed");
		return;
	}
}

void Shader::Activate() const
{
	if (!context->IsCurrentShader(this))
	{
		glUseProgram(Handle());
		context->BindShader(this);
	}
}

void Shader::Deactivate() const
{
	if (!context->IsCurrentShader(nullptr))
	{
		glUseProgram(0);
		context->BindShader(nullptr);
	}
}

Shader::~Shader()
{
	if (handle != 0)
		glDeleteProgram(handle);
}


#pragma region Load and Link

bool Shader::Link(std::string* _error)
{
	//Linken
	glLinkProgram(handle);
	
	//Fehler berprfen
	GLint succeded;
	glGetProgramiv(handle, GL_LINK_STATUS, &succeded);

	if (!succeded)
	{
		//Fehler ausgeben
		GLint logSize;
		glGetProgramiv(handle, GL_INFO_LOG_LENGTH, &logSize);
		
		GLchar* message = new char[logSize];
		glGetProgramInfoLog(handle, logSize, NULL, message);

		Logger::Fatal("Shader program linker error");
		Logger::InfoMessage(message);
		
		if (_error != nullptr)
			(*_error) = std::string(message);

		delete[] message;
		return false;
	}

	//Uniform-Locations automatisch suchen
	QueryActiveUniforms();

	return true;
}

bool Shader::AttachShaderFile(const std::string _shaderPath, GLenum _shaderType, std::string _appendCode, std::string* _error)
{
	if(_shaderType == GL_GEOMETRY_SHADER && (Core::System::SystemConfiguration::Get<int>("OpenGLMajorVersion") < 3 || 
		(Core::System::SystemConfiguration::Get<int>("OpenGLMajorVersion") == 3 && Core::System::SystemConfiguration::Get<int>("OpenGLMinorVersion") < 2)))
	{
		Logger::Error(std::string("Shader not supported for this OpenGL Version. Shader: ") + _shaderPath);
		return false;
	}

	if(_shaderType == GL_TESS_CONTROL_SHADER && Core::System::SystemConfiguration::Get<int>("OpenGLMajorVersion") < 4)
	{
		Logger::Error(std::string("Shader not supported for this OpenGL Version. Shader: ") + _shaderPath);
		return false;
	}

	if(_shaderType == GL_TESS_EVALUATION_SHADER && Core::System::SystemConfiguration::Get<int>("OpenGLMajorVersion") < 4)
	{
		Logger::Error(std::string("Shader not supported for this OpenGL Version. Shader: ") + _shaderPath);
		return false;
	}

	std::string code = LoadShaderFile(_shaderPath);

	if (code != "")
	{
		code.append(_appendCode);

		bool result = AttachShader(code.c_str(), _shaderType, _error);
		if (!result)
			 Logger::InfoMessage(std::string("Path: ") + _shaderPath);


		return result;
	}
	return false;
}

bool Shader::AttachShader(const char* _shaderCode, GLenum _shaderType, std::string* _error)
{
	assert(handle != 0);

	//Attach version Prefix
	std::string code = OpenGLShaderPrefix() + std::string(_shaderCode);

	//Change Primitive if necessary
	if (primitiveType == GL_TRIANGLES && (_shaderType == GL_TESS_CONTROL_SHADER || _shaderType == GL_TESS_EVALUATION_SHADER))
		primitiveType = GL_PATCHES;

	//Shader erstellen
	GLuint shader = glCreateShader(_shaderType);

	if (shader == 0)
	{
		Logger::GLErrors();
		Logger::Fatal("Shader creation failed");
		return false;
	}

	const char* codePointer = code.c_str();
	glShaderSource(shader, 1, &codePointer, NULL);
	glCompileShader(shader);

	//Testen ob es Compilierfehler gab
	GLint succeded;
	glGetShaderiv(shader, GL_COMPILE_STATUS, &succeded);

	if (succeded == GL_FALSE || !glIsShader(shader))
	{
		//Log auslesen und ausgeben
		GLint logSize;
		glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &logSize);
		GLchar* message = new char[logSize];

		glGetShaderInfoLog(shader, logSize, NULL, message);

		if (_error != nullptr)
			(*_error) = std::string(message);

		Logger::Fatal(std::string("Compilation error while loading ") + Convert::ToString(_shaderType));
		Logger::InfoMessage(std::string(message));

		delete[] message;
		return false;
	}

	//Attachen
	glAttachShader(handle, shader);

	//Als gelscht markieren
	glDeleteShader(shader);
	return true;
}

std::string Shader::LoadShaderFile(const std::string _shaderPath)
{
	std::ifstream sfile(_shaderPath);
	
	if (sfile.good())
	{
		std::string code = std::string(std::istreambuf_iterator<char>(sfile), std::istreambuf_iterator<char>());
		sfile.close();

		ParseFileCode(code, _shaderPath);
		return code;
	}
	else
	{
		Logger::Fatal("Cannot open file: " + _shaderPath);
		return "";
	}
}

void Shader::ParseFileCode(std::string& _code, std::string _path)
{
	// Supported Syntax:
	// #include "_filename_"

	std::regex includeRegex("#include\\s+\\\"([^\\\"]+)\\\"");
	std::smatch includeMatch;

	while (std::regex_search(_code, includeMatch, includeRegex))
	{
		//Include Path
		std::string includePath = _path + std::string("\\..\\") + includeMatch[1].str();

		//Load File
		std::ifstream sfile(includePath);
		std::string includeCode = "";
	
		if (sfile.good())
		{
			includeCode = std::string(std::istreambuf_iterator<char>(sfile), std::istreambuf_iterator<char>());

			ParseFileCode(includeCode, includePath);
		}
		else
		{
			Logger::Fatal("Cannot open include file: " + includePath);
		}

		//Replace include with file's code
		_code.replace(includeMatch.position(0), includeMatch.length(0), includeCode, 0, string::npos);
	}
}

#pragma endregion


#pragma region GLSL-Variables

void Shader::QueryActiveUniforms()
{
	GLint uniformCount, maxUniformLength;
	glGetProgramiv(handle, GL_ACTIVE_UNIFORMS, &uniformCount);
	glGetProgramiv(handle, GL_ACTIVE_UNIFORM_MAX_LENGTH, &maxUniformLength);

	assert(maxUniformLength < 100);

	GLchar buffer[100];
	GLsizei nameLength;
	GLint size;
	GLenum type;
	
	for (GLuint i = 0; i < (GLuint)uniformCount; i++)
	{
		//Query and store
		glGetActiveUniform(handle, i, 100, &nameLength, &size, &type, buffer);

		if (size == 1) // Simple uniform
		{
			IUniform* iuniform = BuildUniform(type, buffer);

			if (iuniform != nullptr)
				uniformVariables[HashString((const char*)buffer)] = iuniform;
		}
		else
		{
			if(buffer[nameLength - 1] == ']')
			{
				//Remove [0] at end of name; for Nvidea Driver
				buffer[nameLength - 3] = '\0';
			}

			//Uniform Array
			IUniformArray* iarray = BuildUniformArray(type, buffer, size);

			if (iarray != nullptr)
				uniformArrays[HashString((const char*)buffer)] = iarray;
		}
	}
}

IUniformArray* Shader::BuildUniformArray(GLenum _type, const char* _name, GLsizei _size)
{
	assert(_size < 1000);

	IUniformArray* iarray = nullptr;

	switch (_type)
	{
	case GL_FLOAT:
		iarray = new UniformArray<GLfloat>(_size);
		break;
	case GL_INT:
	case GL_SAMPLER_1D:
	case GL_SAMPLER_2D:
	case GL_SAMPLER_3D:
		iarray = new UniformArray<GLint>(_size);
		break;
	case GL_UNSIGNED_INT:
		iarray = new UniformArray<GLuint>(_size);
		break;
	case GL_FLOAT_VEC2:
		iarray = new UniformArray<Math::Vec2f>(_size);
		break;
	case GL_FLOAT_VEC3:
		iarray = new UniformArray<Math::Vec3f>(_size);
		break;
	case GL_FLOAT_VEC4:
		iarray = new UniformArray<Math::Vec4f>(_size);
		break;
	case GL_BOOL:
		iarray = new UniformArray<GLboolean>(_size);
		break;
	case GL_FLOAT_MAT3:
		iarray = new UniformArray<Math::Mat3f>(_size);
		break;
	case GL_FLOAT_MAT4:
		iarray = new UniformArray<Math::Mat4f>(_size);
		break;
	default:
		Logger::Fatal("Unknown uniform array type");
		break;
	}

	if (iarray != nullptr)
	{
		for (size_t i = 0; i < iarray->Count(); i++)
		{
			stringstream sstr;
			sstr << _name << "[" << i << "]";
			IUniform* iuniform = BuildUniform(_type, sstr.str().c_str());

			if (iuniform != nullptr)
			{
				iarray->SetUniform(i, iuniform);
			}
		}
	}

	return iarray;
}

IUniform* Shader::BuildUniform(GLenum _type, const char* _name)
{
	IUniform* iuniform = nullptr;

	switch (_type)
	{
	case GL_FLOAT:
		iuniform = new Uniform<GLfloat>(this);
		break;
	case GL_INT:
	case GL_SAMPLER_1D:
	case GL_SAMPLER_2D:
	case GL_SAMPLER_3D:
	case GL_INT_IMAGE_BUFFER:
	case GL_UNSIGNED_INT_IMAGE_2D:
	case GL_UNSIGNED_INT_IMAGE_BUFFER:
	case GL_UNSIGNED_INT_SAMPLER_2D:
	case GL_INT_SAMPLER_1D:
	case GL_INT_SAMPLER_2D:
	case GL_INT_SAMPLER_3D:
	case GL_SAMPLER_1D_SHADOW:
	case GL_SAMPLER_2D_SHADOW:
	case GL_IMAGE_1D:
	case GL_IMAGE_2D:
	case GL_IMAGE_3D:
		iuniform = new Uniform<GLint>(this);
		break;
	case GL_UNSIGNED_INT:
		iuniform = new Uniform<GLuint>(this);
		break;
	case GL_FLOAT_VEC2:
		iuniform = new Uniform<Math::Vec2f>(this);
		break;
	case GL_FLOAT_VEC3:
		iuniform = new Uniform<Math::Vec3f>(this);
		break;
	case GL_FLOAT_VEC4:
		iuniform = new Uniform<Math::Vec4f>(this);
		break;
	case GL_BOOL:
		iuniform = new Uniform<GLboolean>(this);
		break;
	case GL_FLOAT_MAT3:
		iuniform = new Uniform<Math::Mat3f>(this);
		break;
	case GL_FLOAT_MAT4:
		iuniform = new Uniform<Math::Mat4f>(this);
		break;
	case GL_UNSIGNED_INT_ATOMIC_COUNTER:
		//This are uniforms that cannot be set with glUniform...
		break;
	default:
		Logger::Fatal("Unknown uniform type");
		break;
	}

	if (iuniform != nullptr)
		iuniform->Bind(_name);

	return iuniform;
}

void Shader::BindOutput(std::string _name, GLuint _index) const
{
	glBindFragDataLocation(handle, _index, _name.c_str());
}

void Shader::BindAttribute(std::string _name, GLuint _index) const
{
	glBindAttribLocation(handle, _index, _name.c_str());
}

#pragma endregion


#pragma region Getter & Setter

inline GLuint Shader::Handle() const
{
	return handle;
}

inline GLenum Shader::PrimitiveType() const
{
	return primitiveType;
}

const std::string Shader::OpenGLShaderPrefix() const
{
	return std::string(
		"#version " + Convert::ToString(Core::System::SystemConfiguration::Get<int>("OpenGLMajorVersion") * 100 + Core::System::SystemConfiguration::Get<int>("OpenGLMinorVersion") * 10) + "\n" +
		"#define _OPEN_GL_MAJOR_VERSION_NUMBER " + Convert::ToString(Core::System::SystemConfiguration::Get<int>("OpenGLMajorVersion")) + "\n" +
		"#define _OPEN_GL_MINOR_VERSION_NUMBER " + Convert::ToString(Core::System::SystemConfiguration::Get<int>("OpenGLMinorVersion")) + "\n"
		);
}

GLint Shader::GetTextureUnit()
{
    GLint unit = maxTextureUnit;
    maxTextureUnit++;
    return unit;
}

void Shader::ResetTextureUnit(GLint _startUnit)
{
    maxTextureUnit = _startUnit;
}


#pragma endregion