#ifndef _ENGINE_GRAPHIC_SHADER_UNIFORM_HPP
#define _ENGINE_GRAPHIC_SHADER_UNIFORM_HPP

#include <string>
#include <gl\glew.h>
#include "IUniform.hpp"
#include "../../Math/Math.hpp"


namespace Engine
{
	namespace Graphic
	{
		namespace Shaders
		{
			class Shader;

			template <typename T>
			class Uniform : public IUniform
			{
			public:
				Uniform(Shader* _shader) : IUniform(_shader) { }

				virtual ~Uniform() { }

				void SetValue(T _value)
				{
					if (HasHandle() && currentValue != _value && Observer() != nullptr)
					{
						currentValue = _value;
						OnNotifyValueChanged();
					}
				}

				Uniform& operator = (const T& _value)
				{
					SetValue(_value);
					return *this;
				}

				static Uniform<T>& Default()
				{
					return defaultUniform;
				}

			protected:

				T& CurrentValue()
				{
					return currentValue;
				}

				void DoChangeValue() { static_assert(false, "This parameter type is currently not supported for Uniforms"); }

			private:
				Uniform() : IUniform(nullptr) { }

				static Uniform<T> defaultUniform;

				T currentValue;
				
			};

			template<typename T> Uniform<T> Uniform<T>::defaultUniform;

			//Spezialisations for the different DoChange Methods
			#pragma region DoChange-Spezialisations

			template <> void Uniform<GLfloat>::DoChangeValue() { glUniform1f(Handle(), CurrentValue()); }

			template <> void Uniform<GLint>::DoChangeValue() { glUniform1i(Handle(), CurrentValue()); }

			template <> void Uniform<GLuint>::DoChangeValue() { glUniform1ui(Handle(), this->CurrentValue()); }

			template <> void Uniform<GLboolean>::DoChangeValue() { glUniform1i(Handle(), CurrentValue()); }

			template <> Uniform<Math::Vec2f>::Uniform(Shader* _shader) : IUniform((Engine::Graphic::DelayedChange::IDelayedValueObserver*)_shader), currentValue(0.0f, 0.0f) {	}
			template <> void Uniform<Math::Vec2f>::DoChangeValue() { glUniform2fv(Handle(), 1, CurrentValue().GetGlVector()); }

			template <> Uniform<Math::Vec3f>::Uniform(Shader* _shader) : IUniform((Engine::Graphic::DelayedChange::IDelayedValueObserver*)_shader), currentValue(0.0f, 0.0f, 0.0f) {	}
			template <> void Uniform<Math::Vec3f>::DoChangeValue() { glUniform3fv(Handle(), 1, CurrentValue().GetGlVector()); }
			
			template <> Uniform<Math::Vec4f>::Uniform(Shader* _shader) :  IUniform((Engine::Graphic::DelayedChange::IDelayedValueObserver*)_shader), currentValue(0.0f, 0.0f, 0.0f, 0.0f) {	}
			template <> void Uniform<Math::Vec4f>::DoChangeValue() { glUniform4fv(Handle(), 1, CurrentValue().GetGlVector()); }

			template <> Uniform<Math::Mat3f>::Uniform(Shader* _shader) :  IUniform((Engine::Graphic::DelayedChange::IDelayedValueObserver*)_shader), currentValue(0.0f) {	}
			template <> void Uniform<Math::Mat3f>::DoChangeValue() { glUniformMatrix3fv(Handle(), 1, !Math::Mat4f::IsColumnMajor(), CurrentValue().GetGlMatrix()); }

			template <> Uniform<Math::Mat4f>::Uniform(Shader* _shader) :  IUniform((Engine::Graphic::DelayedChange::IDelayedValueObserver*)_shader), currentValue(0.0f) {	}
			template <> void Uniform<Math::Mat4f>::DoChangeValue() { glUniformMatrix4fv(Handle(), 1, !Math::Mat4f::IsColumnMajor(), CurrentValue().GetGlMatrix()); }

			#pragma endregion
		}
	}
}

#endif