#define DLL_EXPORT

#include "Vec3f.hpp"

#include <math.h>
#include <algorithm>
#include "Vec4f.hpp"

using namespace Engine::Math::CPPonly;

Vec3f::Vec3f(void)
{
	values[0] = 0.0f;
	values[1] = 0.0f;
	values[2] = 0.0f;
}

Vec3f::Vec3f(float _x, float _y, float _z)
{
	values[0] = _x;
	values[1] = _y;
	values[2] = _z;
}

Vec3f::Vec3f(const Vec3f &_refToVec) 
{
	for (int i = 0; i < 3; i++)
	{
		values[i] = _refToVec.values[i];
	}
}

Vec3f::Vec3f(float* _values)
{
	for (int i = 0; i < 3; i++)
		values[i] = _values[i];
}

Vec3f::Vec3f(const Vec4f &_vec4)
{
	values[0] = _vec4.X();
	values[1] = _vec4.Y();
	values[2] = _vec4.Z();
}

Vec3f::~Vec3f(void)
{ }

float Vec3f::X(void) const
{
	return this->values[0];
}

Vec3f& Vec3f::X(float _x)
{
	this->values[0] = _x;
	return *this;
}

float Vec3f::Y(void) const
{
	return this->values[1];
}

Vec3f& Vec3f::Y(float _y)
{
	this->values[1] = _y;
	return *this;
}

float Vec3f::Z(void) const
{
	return this->values[2];
}

Vec3f& Vec3f::Z(float _z)
{
	this->values[2] = _z;
	return *this;
}

float* Vec3f::GetGlVector() const
{ 
	return (float*)(&values[0]); 
}

Vec4f Vec3f::GetVec4f() const
{
	return Vec4f(X(), Y(), Z());
}

Vec3f& Vec3f::Normalize() 
{ 
	float length = this->Length();

	X(X() / length);
	Y(Y() / length);
	Z(Z() / length);

	return *this; 
}

Vec3f Vec3f::Normalize(const Vec3f &_vector)
{ 
	Vec3f tmp;
	float length = _vector.Length();
	tmp.X(_vector.X() / length);
	tmp.Y(_vector.Y() / length);
	tmp.Z(_vector.Z() / length);
	return tmp;
}

float Vec3f::Length() const
{ 
	return sqrt(pow(X(),2) + pow(Y(),2) + pow(Z(),2)); 
}

float Vec3f::Length(Vec3f &_vector) 
{ 
	return sqrt(pow(_vector.X(),2) + pow(_vector.Y(),2) + pow(_vector.Z(),2)); 
}

float Vec3f::Dot(const Vec3f &_left, const Vec3f &_right)
{ 
	return (_left.X() * _right.X() + _left.Y() * _right.Y() + _left.Z() * _right.Z());
}

Vec3f Vec3f::Cross(const Vec3f &_left, const Vec3f &_right)
{ 
	Vec3f tmp; 
	tmp.X(_left.Y() * _right.Z() - _right.Y() * _left.Z());
	tmp.Y(-(_left.X() * _right.Z() - _right.X() * _left.Z()));
	tmp.Z(_left.X() * _right.Y() - _right.X() * _left.Y());
	return tmp; 
}

Vec3f Vec3f::ComponentMax(const Vec3f& _v1, const Vec3f& _v2)
{
	return Vec3f(std::max(_v1.X(), _v2.X()), std::max(_v1.Y(), _v2.Y()), std::max(_v1.Z(), _v2.Z()));
}

Vec3f Vec3f::ComponentMin(const Vec3f& _v1, const Vec3f& _v2)
{
	return Vec3f(std::min(_v1.X(), _v2.X()), std::min(_v1.Y(), _v2.Y()), std::min(_v1.Z(), _v2.Z()));
}

Vec3f Vec3f::Interpolate(const Vec3f& _v1, const Vec3f& _v2, float _a1, float _a2)
{
	float fac = _a1 / (_a1 + _a2);
	Vec3f diff = _v2 - _v1;

	return Vec3f(_v1.X() + fac * diff.X(), _v1.Y() + fac * diff.Y(), _v1.Z() + fac * diff.Z());
}

Vec3f Vec3f::operator + (const Vec3f &_toAdd) const
{ 
	return Vec3f(X() + _toAdd.X(), Y() + _toAdd.Y(), Z() + _toAdd.Z());
}

Vec3f Vec3f::operator + (const Vec4f &_toAdd) const
{ 
	return Vec3f(X() + _toAdd.X(), Y() + _toAdd.Y(), Z() + _toAdd.Z());
}

Vec3f Vec3f::operator - (const Vec3f &_toSubtract) const
{ 
	Vec3f tmp; 
	tmp.X(X() - _toSubtract.X());
	tmp.Y(Y() - _toSubtract.Y());
	tmp.Z(Z() - _toSubtract.Z());
	return tmp; 
}

Vec3f Vec3f::operator - (const Vec4f &_toSubtract) const
{ 
	Vec3f tmp; 
	tmp.X(X() - _toSubtract.X());
	tmp.Y(Y() - _toSubtract.Y());
	tmp.Z(Z() - _toSubtract.Z());
	return tmp; 
}

Vec3f Vec3f::operator * (float _toMultiply) const
{ 
	Vec3f tmp; 
	tmp.X(X() * _toMultiply);
	tmp.Y(Y() * _toMultiply);
	tmp.Z(Z() * _toMultiply);
	return tmp; 
}

Vec3f Vec3f::operator/ (float _scalar) const
{
	float rezi = 1.0f / _scalar;
	return (*this) * rezi;
}
			
Vec3f& Vec3f::operator += (const Vec3f &_toAdd) 
{ 
	X(X() + _toAdd.X());
	Y(Y() + _toAdd.Y());
	Z(Z() + _toAdd.Z());
	return *this; 
}

Vec3f& Vec3f::operator += (const Vec4f &_toAdd) 
{ 
	X(X() + _toAdd.X());
	Y(Y() + _toAdd.Y());
	Z(Z() + _toAdd.Z());
	return *this; 
}

Vec3f& Vec3f::operator -= (const Vec3f &_toSubtract) 
{ 
	X(X() - _toSubtract.X());
	Y(Y() - _toSubtract.Y());
	Z(Z() - _toSubtract.Z());
	return *this; 
}

Vec3f& Vec3f::operator -= (const Vec4f &_toSubtract) 
{ 
	X(X() - _toSubtract.X());
	Y(Y() - _toSubtract.Y());
	Z(Z() - _toSubtract.Z());
	return *this; 
}

Vec3f& Vec3f::operator *= (float _toMultiply)
{ 
	X(X() * _toMultiply);
	Y(Y() * _toMultiply);
	Z(Z() * _toMultiply);
	return *this; 
}

Vec3f& Vec3f::operator /= (float _scalar)
{
	(*this) *= (1.0f / _scalar);
	return *this;
}

bool Vec3f::operator == (const Vec3f &_toCompare) const
{ 
	if ((X() == _toCompare.X()) && (Y() == _toCompare.Y()) && (Z() == _toCompare.Z()))
		return true;
	else
		return false; 
}

bool Vec3f::operator != (const Vec3f &_toCompare) const
{
	if ((X() == _toCompare.X()) && (Y() == _toCompare.Y()) && (Z() == _toCompare.Z()))
		return false;
	else
		return true; 
}

Vec3f& Vec3f::operator = (const Vec3f& _copy)
{
	if (this != &_copy)
	{
		values[0] = _copy.X();
		values[1] = _copy.Y();
		values[2] = _copy.Z();
	}

	return *this;
}

std::ostream& operator <<(std::ostream &os, const Vec3f &obj)
{
	os << "[" << obj.X() << ", " << obj.Y() << ", " << obj.Z() << "]";
	return os;
}