#define DLL_EXPORT

#include "Vec3f.hpp"

#include <math.h>
#include <algorithm>
#include "../../Core/Hardware/SSEInstructions.hpp"
#include "../../Core/Concurrency/ThreadManager.hpp"
#include "../MathMacros.hpp"

#include "Vec4f.hpp"

using namespace Engine::Math::SSE;
using namespace Engine::Core::Concurrency;

Vec3f::Vec3f(void)
{
	values[0] = 0.0f;
	values[1] = 0.0f;
	values[2] = 0.0f;
}

Vec3f::Vec3f(float _x, float _y, float _z)
{
	values[0] = _x;
	values[1] = _y;
	values[2] = _z;
}

Vec3f::Vec3f(const Vec3f &_refToVec) 
{
	memcpy(&this->values, &_refToVec.values, 3 * sizeof(float));
}

Vec3f::Vec3f(float* _values)
{
	memcpy(&this->values, _values, 3 * sizeof(float));
}

Vec3f::Vec3f(const Vec4f &_vec4)
{
	memcpy(&this->values, _vec4.GetGlVector(), 3 * sizeof(float));
}

Vec3f::~Vec3f(void)
{ }

inline float Vec3f::X(void) const
{
	return this->values[0];
}

inline Vec3f& Vec3f::X(float _x)
{
	this->values[0] = _x;
	return *this;
}

inline float Vec3f::Y(void) const
{
	return this->values[1];
}

inline Vec3f& Vec3f::Y(float _y)
{
	this->values[1] = _y;
	return *this;
}

inline float Vec3f::Z(void) const
{
	return this->values[2];
}

inline Vec3f& Vec3f::Z(float _z)
{
	this->values[2] = _z;
	return *this;
}

inline float* Vec3f::GetGlVector() const
{ 
	return (float*)(&values[0]); 
}

inline Vec4f Vec3f::GetVec4f() const
{
	return Vec4f(X(), Y(), Z());
}

inline __m128 Vec3f::SSELoad(const float* _values)
{
	__m128i low = _mm_loadl_epi64((const __m128i*)_values);
	__m128 high = _mm_load_ss(&_values[2]);
	high = _mm_movelh_ps(_mm_castsi128_ps(low), high);
	return high;
}

inline void Vec3f::SSEStore(__m128& _register, float* _dst)
{
	_mm_store_ps(ThreadManager::tmpSSEMemory<float, 0>(), _register);
	memcpy(_dst, ThreadManager::tmpSSEMemory<float, 0>(), 3 * sizeof(float));
}

Vec3f& Vec3f::Normalize() 
{ 
	__m128 high = Vec3f::SSELoad(this->values);

	__m128 length = _mm_dp_ps(high, high, B8(01111111));
	length = _mm_sqrt_ps(length);

	high = _mm_div_ps(high, length);

	Vec3f::SSEStore(high, (float*)this->values);

	return *this; 
}

Vec3f Vec3f::Normalize(const Vec3f &_vector)
{ 
	__m128 high = Vec3f::SSELoad((float*)_vector.values);

	__m128 length = _mm_dp_ps(high, high, B8(01111111));
	length = _mm_sqrt_ps(length);

	high = _mm_div_ps(high, length);

	_mm_store_ps(ThreadManager::tmpSSEMemory<float, 0>(), high);

	return Vec3f(ThreadManager::tmpSSEMemory<float, 0>());
}

float Vec3f::Length() const
{ 
	__m128 high = Vec3f::SSELoad((float*)this->values);

	__m128 length = _mm_dp_ps(high, high, B8(01111111));
	length = _mm_sqrt_ps(length);

	_mm_storel_pi(ThreadManager::tmpSSEMemory<__m64, 0>(), length);
	return *ThreadManager::tmpSSEMemory<float, 0>();
}

float Vec3f::Length(Vec3f &_vector) 
{ 
	__m128 high = Vec3f::SSELoad((float*)_vector.values);

	__m128 length = _mm_dp_ps(high, high, B8(01111111));
	length = _mm_sqrt_ps(length);

	_mm_storel_pi(ThreadManager::tmpSSEMemory<__m64, 0>(), length);
	return *ThreadManager::tmpSSEMemory<float, 0>();
}

float Vec3f::Dot(const Vec3f &_left, const Vec3f &_right)
{ 
	__m128 v1 = Vec3f::SSELoad((float*)_left.values);
	__m128 v2 = Vec3f::SSELoad((float*)_right.values);

	__m128 dotvec = _mm_dp_ps(v1, v2, B8(01110001));
	
	_mm_store_ss(ThreadManager::tmpSSEMemory<float, 0>(), dotvec);

	return *ThreadManager::tmpSSEMemory<float, 0>();
}

Vec3f Vec3f::Cross(const Vec3f &_left, const Vec3f &_right)
{ 
	__m128 v1 = Vec3f::SSELoad((float*)_left.values);
	__m128 v2 = Vec3f::SSELoad((float*)_right.values);

	__m128 v1shift1 = _mm_shuffle_ps(v1, v1, _MM_SHUFFLE(0, 0, 2, 1));
	__m128 v2shift1 = _mm_shuffle_ps(v2, v2, _MM_SHUFFLE(0, 0, 2, 1));

	__m128 v1shift2 = _mm_shuffle_ps(v1, v1, _MM_SHUFFLE(0, 1, 0, 2));
	__m128 v2shift2 = _mm_shuffle_ps(v2, v2, _MM_SHUFFLE(0, 1, 0, 2));

	v1shift1 = _mm_mul_ps(v1shift1, v2shift2);
	v2shift1 = _mm_mul_ps(v2shift1, v1shift2);

	v1shift1 = _mm_sub_ps(v1shift1, v2shift1);

	_mm_store_ps(ThreadManager::tmpSSEMemory<float, 0>(), v1shift1);
	return Vec3f(ThreadManager::tmpSSEMemory<float, 0>());
}

Vec3f Vec3f::ComponentMax(const Vec3f& _v1, const Vec3f& _v2)
{
	__m128 v1 = Vec3f::SSELoad((float*)_v1.values);
	__m128 v2 = Vec3f::SSELoad((float*)_v2.values);

	__m128 maxVec = _mm_max_ps(v1, v2);

	_mm_store_ps(ThreadManager::tmpSSEMemory<float, 0>(), maxVec);
	return Vec3f(ThreadManager::tmpSSEMemory<float, 0>());
}

Vec3f Vec3f::ComponentMin(const Vec3f& _v1, const Vec3f& _v2)
{
	__m128 v1 = Vec3f::SSELoad((float*)_v1.values);
	__m128 v2 = Vec3f::SSELoad((float*)_v2.values);

	__m128 maxVec = _mm_min_ps(v1, v2);

	_mm_store_ps(ThreadManager::tmpSSEMemory<float, 0>(), maxVec);
	return Vec3f(ThreadManager::tmpSSEMemory<float, 0>());
}

Vec3f Vec3f::Interpolate(const Vec3f& _v1, const Vec3f& _v2, float _a1, float _a2)
{
	__m128 a1register = _mm_load1_ps(&_a1);
	__m128 a2register = _mm_load1_ps(&_a2);

	a2register = _mm_add_ps(a1register, a2register); // _a1 + _a2
	a1register = _mm_div_ps(a1register, a2register); // _a1 / (_a1 + _a2)

	__m128 v1register = Vec3f::SSELoad((float*)_v1.values);
	__m128 v2register = Vec3f::SSELoad((float*)_v2.values);

	v2register = _mm_sub_ps(v2register, v1register); // _v2 - _v1
	v2register = _mm_mul_ps(v2register, a1register); // (_v2 - _v1) * fac

	v1register = _mm_add_ps(v1register, v2register); // _v1 + facvec

	_mm_store_ps(ThreadManager::tmpSSEMemory<float, 0>(), v1register);
	return Vec3f(ThreadManager::tmpSSEMemory<float, 0>());
}

Vec3f Vec3f::operator + (const Vec3f &_toAdd) const
{ 
	return Vec3f(values[0] + _toAdd.values[0], values[1] + _toAdd.values[1], values[2] + _toAdd.values[2]);

	/*__m128i low = _mm_loadl_epi64((const __m128i*)&this->values);
	__m128 v1register = _mm_load_ss(&this->values[2]);
	v1register = _mm_movelh_ps(_mm_castsi128_ps(low), v1register);

	low = _mm_loadl_epi64((const __m128i*)&_toAdd.values);
	__m128 v2register = _mm_load_ss(&_toAdd.values[2]);
	v2register = _mm_movelh_ps(_mm_castsi128_ps(low), v2register);

	v1register = _mm_add_ps(v1register, v2register);

	_mm_store_ps(SystemConfiguration::tmpSSEMemory<float, 0>(), v1register);
	return Vec3f(SystemConfiguration::tmpSSEMemory<float, 0>());*/
}

Vec3f Vec3f::operator + (const Vec4f &_toAdd) const
{ 
	return Vec3f(values[0] + _toAdd.GetGlVector()[0], values[1] + _toAdd.GetGlVector()[1], values[2] + _toAdd.GetGlVector()[2]);

	/*__m128i low = _mm_loadl_epi64((const __m128i*)&this->values);
	__m128 v1register = _mm_load_ss(&this->values[2]);
	v1register = _mm_movelh_ps(_mm_castsi128_ps(low), v1register);

	memcpy(SystemConfiguration::tmpSSEMemory<float, 0>(), &_toAdd, 4*sizeof(float));
	__m128 v2register = _mm_load_ps(SystemConfiguration::tmpSSEMemory<float, 0>());

	v1register = _mm_add_ps(v1register, v2register);

	_mm_store_ps(SystemConfiguration::tmpSSEMemory<float, 0>(), v1register);
	return Vec3f(SystemConfiguration::tmpSSEMemory<float, 0>());*/
}

Vec3f Vec3f::operator - (const Vec3f &_toSubtract) const
{ 
	return Vec3f(values[0] - _toSubtract.values[0], values[1] - _toSubtract.values[1], values[2] - _toSubtract.values[2]);
	/*__m128i low = _mm_loadl_epi64((const __m128i*)&this->values);
	__m128 v1register = _mm_load_ss(&this->values[2]);
	v1register = _mm_movelh_ps(_mm_castsi128_ps(low), v1register);

	low = _mm_loadl_epi64((const __m128i*)&_toSubtract.values);
	__m128 v2register = _mm_load_ss(&_toSubtract.values[2]);
	v2register = _mm_movelh_ps(_mm_castsi128_ps(low), v2register);

	v1register = _mm_sub_ps(v1register, v2register);

	_mm_store_ps(SystemConfiguration::tmpSSEMemory<float, 0>(), v1register);
	return Vec3f(SystemConfiguration::tmpSSEMemory<float, 0>());*/
}

Vec3f Vec3f::operator - (const Vec4f &_toSubtract) const
{ 
	return Vec3f(values[0] - _toSubtract.GetGlVector()[0], values[1] - _toSubtract.GetGlVector()[1], values[2] - _toSubtract.GetGlVector()[2]);
	/*__m128i low = _mm_loadl_epi64((const __m128i*)&this->values);
	__m128 v1register = _mm_load_ss(&this->values[2]);
	v1register = _mm_movelh_ps(_mm_castsi128_ps(low), v1register);

	memcpy(SystemConfiguration::tmpSSEMemory<float, 0>(), &_toSubtract, 4*sizeof(float));
	__m128 v2register = _mm_load_ps(SystemConfiguration::tmpSSEMemory<float, 0>());

	v1register = _mm_sub_ps(v1register, v2register);

	_mm_store_ps(SystemConfiguration::tmpSSEMemory<float, 0>(), v1register);
	return Vec3f(SystemConfiguration::tmpSSEMemory<float, 0>());*/
}

Vec3f Vec3f::operator * (float _scalar) const
{ 
	return Vec3f(values[0] * _scalar, values[1] * _scalar, values[2] * _scalar);

	/*__m128i low = _mm_loadl_epi64((const __m128i*)&this->values);
	__m128 v1register = _mm_load_ss(&this->values[2]);
	v1register = _mm_movelh_ps(_mm_castsi128_ps(low), v1register);

	__m128 facregister = _mm_load1_ps(&_scalar);

	v1register = _mm_mul_ps(v1register, facregister);

	_mm_store_ps(SystemConfiguration::tmpSSEMemory<float, 0>(), v1register);
	return Vec3f(SystemConfiguration::tmpSSEMemory<float, 0>());*/
}

Vec3f Vec3f::operator/ (float _scalar) const
{
	__m128 v1register = Vec3f::SSELoad((float*)this->values);

	__m128 facregister = _mm_load1_ps(&_scalar);

	v1register = _mm_div_ps(v1register, facregister);

	_mm_store_ps(ThreadManager::tmpSSEMemory<float, 0>(), v1register);
	return Vec3f(ThreadManager::tmpSSEMemory<float, 0>());
}
			
Vec3f& Vec3f::operator += (const Vec3f &_toAdd) 
{ 
	values[0] += _toAdd.values[0];
	values[1] += _toAdd.values[1];
	values[2] += _toAdd.values[2];
	return *this;

	/*__m128i low = _mm_loadl_epi64((const __m128i*)&this->values);
	__m128 v1register = _mm_load_ss(&this->values[2]);
	v1register = _mm_movelh_ps(_mm_castsi128_ps(low), v1register);

	low = _mm_loadl_epi64((const __m128i*)&_toAdd.values);
	__m128 v2register = _mm_load_ss(&_toAdd.values[2]);
	v2register = _mm_movelh_ps(_mm_castsi128_ps(low), v2register);

	v1register = _mm_add_ps(v1register, v2register);

	_mm_store_ps(SystemConfiguration::tmpSSEMemory<float, 0>(), v1register);
	memcpy(&values, SystemConfiguration::tmpSSEMemory<float, 0>(), 3 * sizeof(float));
	return *this;*/
}

Vec3f& Vec3f::operator += (const Vec4f &_toAdd) 
{ 
	values[0] += _toAdd.GetGlVector()[0];
	values[1] += _toAdd.GetGlVector()[1];
	values[2] += _toAdd.GetGlVector()[2];
	return *this;

	/*__m128i low = _mm_loadl_epi64((const __m128i*)&this->values);
	__m128 v1register = _mm_load_ss(&this->values[2]);
	v1register = _mm_movelh_ps(_mm_castsi128_ps(low), v1register);

	memcpy(SystemConfiguration::tmpSSEMemory<float, 0>(), &_toAdd, 4*sizeof(float));
	__m128 v2register = _mm_load_ps(SystemConfiguration::tmpSSEMemory<float, 0>());

	v1register = _mm_add_ps(v1register, v2register);

	_mm_store_ps(SystemConfiguration::tmpSSEMemory<float, 0>(), v1register);
	memcpy(&values, SystemConfiguration::tmpSSEMemory<float, 0>(), 3 * sizeof(float));
	return *this;*/
}

Vec3f& Vec3f::operator -= (const Vec3f &_toSubtract) 
{ 
	values[0] -= _toSubtract.values[0];
	values[1] -= _toSubtract.values[1];
	values[2] -= _toSubtract.values[2];
	return *this;

	/*__m128i low = _mm_loadl_epi64((const __m128i*)&this->values);
	__m128 v1register = _mm_load_ss(&this->values[2]);
	v1register = _mm_movelh_ps(_mm_castsi128_ps(low), v1register);

	low = _mm_loadl_epi64((const __m128i*)&_toSubtract.values);
	__m128 v2register = _mm_load_ss(&_toSubtract.values[2]);
	v2register = _mm_movelh_ps(_mm_castsi128_ps(low), v2register);

	v1register = _mm_sub_ps(v1register, v2register);

	_mm_store_ps(SystemConfiguration::tmpSSEMemory<float, 0>(), v1register);
	memcpy(&values, SystemConfiguration::tmpSSEMemory<float, 0>(), 3 * sizeof(float));
	return *this;*/
}

Vec3f& Vec3f::operator -= (const Vec4f &_toSubtract) 
{ 
	values[0] -= _toSubtract.GetGlVector()[0];
	values[1] -= _toSubtract.GetGlVector()[1];
	values[2] -= _toSubtract.GetGlVector()[2];
	return *this;

	/*__m128i low = _mm_loadl_epi64((const __m128i*)&this->values);
	__m128 v1register = _mm_load_ss(&this->values[2]);
	v1register = _mm_movelh_ps(_mm_castsi128_ps(low), v1register);

	memcpy(SystemConfiguration::tmpSSEMemory<float, 0>(), &_toSubtract, 4*sizeof(float));
	__m128 v2register = _mm_load_ps(SystemConfiguration::tmpSSEMemory<float, 0>());

	v1register = _mm_sub_ps(v1register, v2register);

	_mm_store_ps(SystemConfiguration::tmpSSEMemory<float, 0>(), v1register);
	memcpy(&values, SystemConfiguration::tmpSSEMemory<float, 0>(), 3 * sizeof(float));
	return *this;*/
}

Vec3f& Vec3f::operator *= (float _scalar)
{ 
	values[0] *= _scalar;
	values[1] *= _scalar;
	values[2] *= _scalar;
	return *this;

	/*__m128i low = _mm_loadl_epi64((const __m128i*)&this->values);
	__m128 v1register = _mm_load_ss(&this->values[2]);
	v1register = _mm_movelh_ps(_mm_castsi128_ps(low), v1register);

	__m128 facregister = _mm_load1_ps(&_scalar);

	v1register = _mm_mul_ps(v1register, facregister);

	_mm_store_ps(SystemConfiguration::tmpSSEMemory<float, 0>(), v1register);
	memcpy(&values, SystemConfiguration::tmpSSEMemory<float, 0>(), 3 * sizeof(float));
	return *this; */
}

Vec3f& Vec3f::operator /= (float _scalar)
{
	__m128 v1register = Vec3f::SSELoad((float*)this->values);

	__m128 facregister = _mm_load1_ps(&_scalar);

	v1register = _mm_div_ps(v1register, facregister);

	_mm_store_ps(ThreadManager::tmpSSEMemory<float, 0>(), v1register);
	memcpy(&values, ThreadManager::tmpSSEMemory<float, 0>(), 3 * sizeof(float));
	return *this; 
}

bool Vec3f::operator == (const Vec3f &_toCompare) const
{ 
	if ((X() == _toCompare.X()) && (Y() == _toCompare.Y()) && (Z() == _toCompare.Z()))
		return true;
	else
		return false; 
}

bool Vec3f::operator != (const Vec3f &_toCompare) const
{
	if ((X() == _toCompare.X()) && (Y() == _toCompare.Y()) && (Z() == _toCompare.Z()))
		return false;
	else
		return true; 
}

Vec3f& Vec3f::operator = (const Vec3f& _copy)
{
	memcpy(&this->values, &_copy.values, 3 * sizeof(float));
	return *this;
}

std::ostream& operator <<(std::ostream &os, const Vec3f &obj)
{
	os << "[" << obj.X() << ", " << obj.Y() << ", " << obj.Z() << "]";
	return os;
}