#define DLL_EXPORT

#include "Curve.hpp"
#include "CurveData.hpp"
#include "CurvePoint.hpp"
#include "../../Util/Logger.hpp"

#include <math.h>

using namespace Engine;
using namespace Engine::Scene;
using namespace Engine::Scene::Animations;
using namespace Engine::Util;

#define CLAMP(X, A, B) X < A ? A : (X > B ? B : X)
#define SMALL -1.0e-10

Curve::Curve(Util::SharedPointer<CurveData> _data)
	: data(_data), targetValue(nullptr), lastPointIdx(0), length(0.0f)
{
	for (unsigned int i = 0; i < data->Count(); i++)
		length = max(length, data->Points()[i].Point().X());

	if (data->Points()[0].Point().X() >= 0.001f)
		Logger::Warning("First Point of animation is not at 0.0f");
}

Curve::~Curve() { }

void Curve::Interpolate(float _t)
{
	if (targetValue != nullptr)
	{
		CurvePoint* prevPoint = &(data->Points()[lastPointIdx]);
		CurvePoint* nextPoint = &(data->Points()[(lastPointIdx + 1) % data->Count()]);
	
		while (prevPoint->Point().X() >= _t && lastPointIdx != 0)
		{
			lastPointIdx--;
			prevPoint = &(data->Points()[lastPointIdx]);
			nextPoint = &(data->Points()[(lastPointIdx + 1) % data->Count()]);
		}
		while (nextPoint->Point().X() < _t && lastPointIdx != data->Count() - 1)
		{
			lastPointIdx++;
			prevPoint = &(data->Points()[lastPointIdx]);
			nextPoint = &(data->Points()[(lastPointIdx + 1) % data->Count()]);
		}

		float x = CLAMP(_t, prevPoint->Point().X(), nextPoint->Point().X());

		//Calculate interpolation
		if (prevPoint->Interpolation() == InterpolationMode::BezierInterpolationMode)
		{
			float o[32];
			int b = FindZero(x, prevPoint->Point().X(), prevPoint->NextHandle().X(), nextPoint->PrevHandle().X(), nextPoint->Point().X(), o);
			if (b)
			{
				Berekeny(prevPoint->Point().Y(), prevPoint->NextHandle().Y(), nextPoint->PrevHandle().Y(), nextPoint->Point().Y(), o, 1);
				*targetValue = o[0];
			}
		}
		else if (prevPoint->Interpolation() == InterpolationMode::LinearInterpolationMode)
		{
			float scale = (_t - prevPoint->Point().X()) / (nextPoint->Point().X() - prevPoint->Point().X());
			*targetValue = prevPoint->Point().Y() + (nextPoint->Point().Y() - prevPoint->Point().Y()) * scale;
		}
		else
		{
			Logger::Error("Unknown interpolation type");
		}
	}
	else
		Logger::Warning("TargetValue for CurveInterpolation is nullptr");
}

void Curve::TargetValue(float* _value)
{
	targetValue = _value;
}

const Util::SharedPointer<CurveData>& Curve::Data() const
{
	return data;
}

inline float Curve::Length() const
{
	return length;
}

/* The next three functions are from the blender source code
 *
 * Sqrt3d
 * FindZero
 * Berekeny
 */

double Curve::Sqrt3d(double d)
{
	if (d == 0.0)
		return 0.0;
	else if (d <  0.0)
		return -exp(log(-d) / 3.0);
	else
		return  exp(log( d) / 3.0);
}

int Curve::FindZero(float _x, float _x0, float _x1, float _x2, float _x3, float *_o)
{
	double c0, c1, c2, c3, a, b, c, p, q, d, t, phi;
	int nr = 0;

	c0 = _x0 - _x;
	c1 = 3.0f * (_x1 - _x0);
	c2 = 3.0f * (_x0 - 2.0f * _x1 + _x2);
	c3 = _x3 - _x0 + 3.0f * (_x1 - _x2);
	
	if (c3 != 0.0) {
		a = c2 / c3;
		b = c1 / c3;
		c = c0 / c3;
		a = a / 3;

		p = b / 3 - a * a;
		q = (2 * a * a * a - a * b + c) / 2;
		d = q * q + p * p * p;
		
		if (d > 0.0) {
			t = sqrt(d);
			_o[0] = (float)(Sqrt3d(-q + t) + Sqrt3d(-q - t) - a);
			
			if ((_o[0] >= (float)SMALL) && (_o[0] <= 1.000001f)) return 1;
			else return 0;
		}
		else if (d == 0.0) {
			t = Sqrt3d(-q);
			_o[0] = (float)(2 * t - a);
			
			if ((_o[0] >= (float)SMALL) && (_o[0] <= 1.000001f)) nr++;
			_o[nr] = (float)(-t - a);
			
			if ((_o[nr] >= (float)SMALL) && (_o[nr] <= 1.000001f)) return nr + 1;
			else return nr;
		}
		else {
			phi = acos(-q / sqrt(-(p * p * p)));
			t = sqrt(-p);
			p = cos(phi / 3);
			q = sqrt(3 - 3 * p * p);
			_o[0] = (float)(2 * t * p - a);
			
			if ((_o[0] >= (float)SMALL) && (_o[0] <= 1.000001f)) nr++;
			_o[nr] = (float)(-t * (p + q) - a);
			
			if ((_o[nr] >= (float)SMALL) && (_o[nr] <= 1.000001f)) nr++;
			_o[nr] = (float)(-t * (p - q) - a);
			
			if ((_o[nr] >= (float)SMALL) && (_o[nr] <= 1.000001f)) return nr + 1;
			else return nr;
		}
	}
	else {
		a = c2;
		b = c1;
		c = c0;
		
		if (a != 0.0) {
			/* discriminant */
			p = b * b - 4 * a * c;
			
			if (p > 0) {
				p = sqrt(p);
				_o[0] = (float)((-b - p) / (2 * a));
				
				if ((_o[0] >= (float)SMALL) && (_o[0] <= 1.000001f)) nr++;
				_o[nr] = (float)((-b + p) / (2 * a));
				
				if ((_o[nr] >= (float)SMALL) && (_o[nr] <= 1.000001f)) return nr + 1;
				else return nr;
			}
			else if (p == 0) {
				_o[0] = (float)(-b / (2 * a));
				if ((_o[0] >= (float)SMALL) && (_o[0] <= 1.000001f)) return 1;
				else return 0;
			}
		}
		else if (b != 0.0) {
			_o[0] = (float)(-c / b);
			
			if ((_o[0] >= (float)SMALL) && (_o[0] <= 1.000001f)) return 1;
			else return 0;
		}
		else if (c == 0.0) {
			_o[0] = 0.0;
			return 1;
		}
		
		return 0;
	}
}

void Curve::Berekeny(float _y1, float _y2, float _y3, float _y4, float* _o, int _b)
{
	float t, c0, c1, c2, c3;
	int a;

	c0 = _y1;
	c1 = 3.0f * (_y2 - _y1);
	c2 = 3.0f * (_y1 - 2.0f * _y2 + _y3);
	c3 = _y4 - _y1 + 3.0f * (_y2 - _y3);

	for (a = 0; a < _b; a++) {
		t = _o[a];
		_o[a] = c0 + t * c1 + t * t * c2 + t * t * t * c3;
	}
}