#define DLL_EXPORT

#include "Cone.hpp"
#include "../Cameras/Camera.hpp"
#include "Frustum.hpp"

#define _USE_MATH_DEFINES
#include <math.h>

using namespace Engine;
using namespace Engine::Math;
using namespace Engine::Scene;
using namespace Engine::Scene::Boundings;

Cone::Cone(float _angle, float _height) : angle(_angle), height(_height), direction(0.0f, 0.0f, -1.0f), position(0.0f, 0.0f, 0.0f, 1.0f)
{ }
Cone::~Cone()
{ }

BoundingObject* Cone::Transform(const Math::Mat4f& _matrix, BoundingObject* _target)
{
	if (_target == nullptr)
		_target = new Cone(angle, height);

	((Cone*)_target)->direction = Mat4f::GetRotation(_matrix) * direction;
	((Cone*)_target)->position = _matrix * position;

	return _target;
}

bool Cone::IsVisible(const Cameras::Camera* const _camera) const
{
	return _camera->Frustum().Intersect(this);
}

SplittingResult Cone::IntersectPlane(const Math::Vec4f& _plane) const
{
	//beta = angle between plane normal and direction

	float d = Vec4f::Dot(_plane, position);
	float beta = acosf(Vec3f::Dot(_plane, direction));

	if (d > 0.0f)
	{
		//Front
		//beta + angle > 90
		if (beta + angle < (float)M_PI_2)
			return SplittingResult::Front;
		
		float gamma = fabs((float)M_PI - beta - angle);
		float d_angle = (float)M_PI_2 - gamma;

		float dist_to_plane = d / sinf(d_angle);

		if (dist_to_plane < height)
			return SplittingResult::Split;

		return SplittingResult::Front;
	}
	else
	{
		//Back
		if (beta - angle > M_PI_2)
			return SplittingResult::Back;

		float gamma = fabs(beta - angle);
		float d_angle = (float)M_PI_2 - gamma;

		float dist_to_plane = d / sinf(d_angle);

		if (dist_to_plane < height)
			return SplittingResult::Split;

		return SplittingResult::Back;
	}
}

const Math::Vec3f& Cone::Direction() const
{
	return direction;
}
const Math::Vec4f& Cone::Position() const
{
	return position;
}
float Cone::Angle() const
{
	return angle;
}
void Cone::Angle(float _angle)
{
	angle = _angle;
}
float Cone::Height() const
{
	return height;
}

void Cone::Height(float _height)
{
	height = _height;
}

Math::Vec3f Cone::GetCenterPoint() const
{
	/* assert direction is normalized */
	return position + (direction*height*2)/3;
}

Bih::MinMaxBoxf Cone::GetAabb() const
{
	/* We use a modified cone where its base has the shape of a sphere */
	/* This Sphere is determined by the incircle */
	/* http://de.wikipedia.org/wiki/Inkreis#Inkreis_eines_Dreiecks */

	/* assert angle < PI/2 */
	float hbase = tan(angle) * height;

	float ab, c, bzab, bzc, s, circ, r;
	c = 2*hbase;
	ab = sqrt(height*height + hbase*hbase);
	circ = 2*ab + c;

	bzab = ab / circ; /* barycentric weights of incircle */
	bzc = c / circ;



	s = (ab + ab + c)/2.0f;
	r = sqrt(((s-ab)*(s-ab)*(s-c))/s); /* radius */

	/* It is a isosceles triangle, weights for a and b the same */
	Vec3f incircle = (Vec3f)position*bzc + ((Vec3f)position + (direction*height*2)) * 2 * bzab;

	Bih::MinMaxBoxf retval;
	
	retval.x_min = incircle.X() - r;
	retval.x_max = incircle.X() + r;
	retval.y_min = incircle.Y() - r;
	retval.y_max = incircle.Y() + r;
	retval.z_min = incircle.Z() - r;
	retval.z_max = incircle.Z() + r;
	
	retval.x_min = (position.X() < retval.x_min) ? position.X() : retval.x_min;
	retval.x_max = (position.X() > retval.x_max) ? position.X() : retval.x_max;
	retval.y_min = (position.Y() < retval.y_min) ? position.Y() : retval.y_min;
	retval.y_max = (position.Y() > retval.y_max) ? position.Y() : retval.y_max;
	retval.z_min = (position.Z() < retval.z_min) ? position.Z() : retval.z_min;
	retval.z_max = (position.Z() > retval.z_max) ? position.Z() : retval.z_max;

	return retval;

}