#define DLL_EXPORT

#include "Frustum.hpp"
#include "../../Core/Concurrency/ThreadManager.hpp"
#include "../Cameras/Camera.hpp"

#include <smmintrin.h>

using namespace Engine;
using namespace Engine::Math;
using namespace Engine::Scene::Boundings;
using namespace Engine::Core::Concurrency;

Frustum::Frustum(float _fov, float _ratio, float _nearPlane, float _farPlane) : nearPlane(_nearPlane), farPlane(_farPlane), projectionType(FrustumProjectionType::Perspective)
{
	Resize(_fov, _ratio, _nearPlane, _farPlane);
}

Frustum::Frustum(float _left, float _right, float _bottom, float _top, float _nearPlane, float _farPlane) : left(_left), right(_right), bottom(_bottom),
	top(_top), nearPlane(_nearPlane), farPlane(_farPlane), projectionType(FrustumProjectionType::Orthographic)
{ }

void Frustum::PlaneNormal(int _index, const Math::Vec3f& _normal)
{
	memcpy(plane[_index].GetGlVector(), _normal.GetGlVector(), 3*sizeof(float));
}

void Frustum::PlaneD(int _index, float _d)
{
	plane[_index].W(_d);
}

void Frustum::FromCameraParameter(const Math::Vec4f& _position, const Math::Vec4f& _right, const Math::Vec4f& _up, const Math::Vec4f& _view)
{
	if (projectionType == FrustumProjectionType::Perspective)
	{
		Vec4f fc = _position + _view * farPlane;
		Vec4f nc = _position + _view * nearPlane;

		//NearPlane
		plane[0] = Vec4f(_view.X(), _view.Y(), _view.Z(), -Math::Vec4f::Dot3(_view, nc));
		//FarPlane
		plane[1] = Vec4f(-_view.X(), -_view.Y(), -_view.Z(), -Math::Vec4f::Dot3(_view * -1.0f, fc));
		//Right
		Vec4f n = Math::Vec4f::Cross3(_up, Math::Vec4f::Normalize3((nc + _right * wNear) - _position));
		plane[2] = Vec4f(n.X(), n.Y(), n.Z(), -Math::Vec4f::Dot3(_position, n));
		//Left
		n = Math::Vec4f::Cross3(Math::Vec4f::Normalize3((nc - _right * wNear) - _position), _up).Normalize3();
		plane[3] = Vec4f(n.X(), n.Y(), n.Z(), -Math::Vec4f::Dot3(_position, n));
		//Top
		n = Math::Vec4f::Cross3(Math::Vec4f::Normalize3((nc + _up * hNear) - _position), _right).Normalize3();
		plane[4] = Vec4f(n.X(), n.Y(), n.Z(), -Math::Vec4f::Dot3(_position, n));
		//Bottom
		n = Math::Vec4f::Cross3(_right, Math::Vec4f::Normalize3((nc - _up * hNear) - _position)).Normalize3();
		plane[5] = Vec4f(n.X(), n.Y(), n.Z(), -Math::Vec4f::Dot3(_position, n));
	}
	else if (projectionType == FrustumProjectionType::Orthographic)
	{
		//NearPlane
		plane[0] = Vec4f(_view.X(), _view.Y(), _view.Z(), -Math::Vec4f::Dot3(_view, _position + _view * nearPlane));
		//FarPlane
		plane[1] = Vec4f(-_view.X(), -_view.Y(), -_view.Z(), -Math::Vec4f::Dot3(_view * -1.0f, _position + _view * farPlane));
		//Right
		plane[2] = Vec4f(_right.X(), _right.Y(), _right.Z(), -Math::Vec4f::Dot3(_right, _position + _right * right));
		//Left
		plane[3] = Vec4f(-_right.X(), -_right.Y(), -_right.Z(), -Math::Vec4f::Dot3(_right * -1.0f, _position + _right * right));
		//Top
		plane[4] = Vec4f(_up.X(), _up.Y(), _up.Z(),  -Math::Vec4f::Dot3(_up, _position + _up * top));
		//Bottom
		plane[5] = Vec4f(-_up.X(), -_up.Y(), -_up.Z(),  -Math::Vec4f::Dot3(_up, _position + _up * bottom));
	}
}

bool Frustum::Intersect(const BoundingObject* _bobject) const
{
	for (int i = 0; i < 6; i++)
	{
		if (_bobject->IntersectPlane(plane[i]) == SplittingResult::Back)
			return false;
	}

	return true;
}

void Frustum::Resize(float _fov, float _ratio, float _nearPlane, float _farPlane)
{
	hNear =tan(_fov / 2.0f) * _nearPlane;
	wNear = hNear * _ratio;

	hFar = tan(_fov / 2.0f) * _farPlane;
	wFar = hFar * _ratio;
}