#define DLL_EXPORT

#include "TransformedBox.hpp"
#include <cassert>
#include "../Cameras/Camera.hpp"
#include "Frustum.hpp"

using namespace Engine;
using namespace Engine::Scene::Boundings;
using namespace Engine::Scene;
using namespace Engine::Math;

TransformedBox::TransformedBox()
{ }

TransformedBox::TransformedBox(const Math::Vec4f& _location, const Math::Vec4f& _axis1, const Math::Vec4f& _axis2, const Math::Vec4f& _axis3)
	: location(_location), axis1(_axis1), axis2(_axis2), axis3(_axis3), normalizedAxis1(Vec4f::Normalize3(_axis1)), normalizedAxis2(Vec4f::Normalize3(_axis2)),
	normalizedAxis3(Vec4f::Normalize3(_axis3))
{
	Vec4f maxPoint = _location + _axis1 + _axis2 + _axis3;
	planeDistances[0] = -Vec4f::Dot3(normalizedAxis1, _location);
	planeDistances[1] = -Vec4f::Dot3(normalizedAxis1 * -1.0f, maxPoint);
	planeDistances[2] = -Vec4f::Dot3(normalizedAxis2, _location);
	planeDistances[3] = -Vec4f::Dot3(normalizedAxis2 * -1.0f, maxPoint);
	planeDistances[4] = -Vec4f::Dot3(normalizedAxis3, _location);
	planeDistances[5] = -Vec4f::Dot3(normalizedAxis3 * -1.0f, maxPoint);
}

TransformedBox::TransformedBox(const TransformedBox& _box) : location(_box.location), axis1(_box.axis1), axis2(_box.axis2), axis3(_box.axis3), normalizedAxis1(_box.normalizedAxis1)
	, normalizedAxis2(_box.normalizedAxis2), normalizedAxis3(_box.normalizedAxis3)
{ }

const Vec4f& TransformedBox::Location() const
{
	return location;
}

const Vec4f& TransformedBox::Axis1() const
{
	return axis1;
}

const Vec4f& TransformedBox::Axis2() const
{
	return axis2;
}

const Vec4f& TransformedBox::Axis3() const
{
	return axis3;
}

const Vec4f& TransformedBox::NormalizedAxis1() const
{
	return normalizedAxis1;
}

const Vec4f& TransformedBox::NormalizedAxis2() const
{
	return normalizedAxis2;
}

const Vec4f& TransformedBox::NormalizedAxis3() const
{
	return normalizedAxis3;
}

const float* TransformedBox::PlaneDistances() const
{
	return planeDistances;
}

SplittingResult TransformedBox::IntersectPlane(const Math::Vec4f& _plane) const
{
	//Transform normal to box space
	Vec3f nboxspace = Vec3f(Vec4f::Dot3(normalizedAxis1, _plane), Vec4f::Dot3(normalizedAxis2, _plane), Vec4f::Dot3(normalizedAxis3, _plane));

	//Calculate p and n vertex
	Vec4f p = location;
	Vec4f n = location;

	if (nboxspace.X() >= 0.0f)
		p += axis1;
	else
		n += axis1;
	if (nboxspace.Y() >= 0.0f)
		p += axis2;
	else
		n += axis2;
	if (nboxspace.Z() >= 0.0f)
		p += axis3;
	else
		n += axis3;

	if (Vec4f::Dot(p, _plane) < 0.01) //If p outside -> n outside
		return SplittingResult::Back;
	else if (Vec4f::Dot(n, _plane) < -0.01) //p inside and n outside
		return SplittingResult::Split;

	return SplittingResult::Front;
}

const TransformedBox& TransformedBox::operator = (const TransformedBox& _box)
{
	memcpy(planeDistances, _box.planeDistances, 6 * sizeof(float));
	location = _box.location;
	axis1 = _box.axis1;
	axis2 = _box.axis2;
	axis3 = _box.axis3;
	normalizedAxis1 = _box.normalizedAxis1;
	normalizedAxis2 = _box.normalizedAxis2;
	normalizedAxis3 = _box.normalizedAxis3;
	
	return *this;
}

BoundingObject* TransformedBox::Transform(const Mat4f& _matrix, BoundingObject* _target)
{
	assert("Operation not supported");
	return nullptr;
}

bool TransformedBox::IsVisible(const Cameras::Camera* const _camera) const
{
	return _camera->Frustum().Intersect(this);
}

Math::Vec3f TransformedBox::GetCenterPoint() const
{
	return location + axis1/2 + axis2/2 + axis3/2;
}

void enlargeMinMaxBox(Bih::MinMaxBoxf& tomod, const Vec3f& outliner)
{
	/* Makes sure that outliner fits in box */
	tomod.x_min = outliner.X() < tomod.x_min ? outliner.X() : tomod.x_min;
	tomod.x_max = outliner.X() > tomod.x_max ? outliner.X() : tomod.x_max;
	tomod.y_min = outliner.Y() < tomod.y_min ? outliner.Y() : tomod.y_min;
	tomod.y_max = outliner.Y() > tomod.y_max ? outliner.Y() : tomod.y_max;
	tomod.z_min = outliner.Z() < tomod.z_min ? outliner.Z() : tomod.z_min;
	tomod.z_max = outliner.Z() > tomod.z_max ? outliner.Z() : tomod.z_max;
}

Bih::MinMaxBoxf TransformedBox::GetAabb() const
{
	/* Corner Points */
	Vec3f a1 = (Vec3f)location;
	Vec3f a2 = a1 + axis1;
	Vec3f a3 = a1 + axis2;
	Vec3f a4 = a1 + axis3;
	Vec3f a5 = a2 + axis2;
	Vec3f a6 = a3 + axis3;
	Vec3f a7 = a2 + axis3;
	Vec3f a8 = a5 + axis3;

	Bih::MinMaxBoxf retval = {a1.X(), a1.X(), a1.Y(), a1.Y(), a1.Z(), a1.Z()};
	enlargeMinMaxBox(retval, a2);
	enlargeMinMaxBox(retval, a3);
	enlargeMinMaxBox(retval, a4);
	enlargeMinMaxBox(retval, a5);
	enlargeMinMaxBox(retval, a6);
	enlargeMinMaxBox(retval, a7);
	enlargeMinMaxBox(retval, a8);

	return retval;
}
