#define DLL_EXPORT

#include "GICamera.hpp"
#include "../SceneObject.hpp"
#include "../World.hpp"
#include "../Boundings/BoxBSPNode.hpp"
#include "../../EngineManager.hpp"
#include "../../Core/Control/RenderControl.hpp"
#include "../Components/LightComponents/LightComponent.hpp"
#include "../Components/BoundingComponent.hpp"
#include "../Components/VoxelOctreeComponent.hpp"
#include "../../Graphic/GI/Rendering/IRenderStage.hpp"
#include "../../Graphic/BasicBodies/Plane.hpp"
#include "../../Core/IO/KeyboardInput.hpp"

#define _USE_MATH_DEFINES
#include <math.h>

#include "../Components/MeshRenderComponent.hpp"
#include "../../Graphic/GI/Illumination/IIlluminationStage.hpp"

using namespace Engine;
using namespace Engine::Util;
using namespace Engine::Math;
using namespace Engine::Scene;
using namespace Engine::Scene::Boundings;
using namespace Engine::Scene::Components;
using namespace Engine::Scene::Cameras;
using namespace Engine::Scene::Lights;
using namespace Engine::Graphic::Shaders;
using namespace Engine::Graphic::GI;
using namespace Engine::Graphic::GI::Generation;
using namespace Engine::Graphic::GI::Preparation;
using namespace Engine::Core;
using namespace Engine::Core::IO;

GICamera::GICamera(float _fov, unsigned int _width, unsigned int _height, float _nearPlane, float _farPlane, Scene::SceneObject* _sceneObject,
						   Graphic::GraphicContext* _context, EngineManager* _manager, VoxelOctreeComponent* _octree)
	: Camera(_sceneObject, _manager),
	nearPlane(_nearPlane), farPlane(_farPlane), fov(_fov), glStopWatch(),
	frustum(_fov, (float)_width / (float)_height, _nearPlane, _farPlane),
	projectionMatrix(Mat4f::Persp(_fov, (float)_width / (float)_height, _nearPlane, _farPlane)),
	context(_context), octree(_octree), cppStopWatch("cppStopWatch"),
	directionalLightEffect(*this, _context),
	pointLightEffect(*this, _context)
	, deferredBuffer(_context, _width, _height), depthBufferTexture(GL_DEPTH_COMPONENT24, GL_DEPTH_COMPONENT, false),
	normalTexture(GL_RGBA16F, GL_RGBA, false), tangentTexture(GL_RGB16F, GL_RGB, false), depthTexture(GL_R32F, GL_RED, false),
	diffuseIllumination(GL_RGB16F, GL_RGB, false), specularIllumination(GL_RGB16F, GL_RGB, false), illuminationBuffer(_context, _width, _height)
	, manager(_manager)
{
	deferredBuffer.AttachTexture(&normalTexture, GL_COLOR_ATTACHMENT0);
	deferredBuffer.AttachTexture(&tangentTexture, GL_COLOR_ATTACHMENT1);
	deferredBuffer.AttachTexture(&depthTexture, GL_COLOR_ATTACHMENT2);
	deferredBuffer.AttachTexture(&depthBufferTexture, GL_DEPTH_ATTACHMENT);
	deferredBuffer.Init();

	illuminationBuffer.AttachTexture(&diffuseIllumination, GL_COLOR_ATTACHMENT0);
	illuminationBuffer.AttachTexture(&specularIllumination, GL_COLOR_ATTACHMENT1);
	illuminationBuffer.AttachTexture(&depthBufferTexture, GL_DEPTH_ATTACHMENT);

	Resize(_width, _height);
}

GICamera::~GICamera()
{ }

void GICamera::Update(const Math::Mat4f& _matrix)
{
	cameraMatrix = projectionMatrix * Mat4f::Invert(_matrix);
	inverseCameraMatrix = Mat4f::Invert(cameraMatrix);
	position = Vec4f(_matrix.M14(), _matrix.M24(), _matrix.M34(), 1.0f);

	Vec4f view = Math::Vec4f::Normalize3(Vec4f(-_matrix.M13(), -_matrix.M23(), -_matrix.M33(), 0.0f));
	Vec4f up = Math::Vec4f::Normalize3(Vec4f(_matrix.M12(), _matrix.M22(), _matrix.M32(), 0.0f));
	Vec4f right = Vec4f::Cross3(view, up);

	frustum.FromCameraParameter(position, right, up, view);
}

void GICamera::Render()
{
	cppStopWatch.Start();
	glStopWatch.Start();

	octree->Render();

	visibleObjects = 0;
	World* world = SceneObject()->World();

	
	context->Blend().Disable();
	context->DepthTest().TestAndWrite();
	context->DepthTest().DepthFunc(GL_LESS);
	context->CullFace().EnableBackFaceCulling();
	context->StencilTest().Disable();

	//Render Normals, Tangents and Bitangents
	deferredBuffer.Bind();
	deferredBuffer.Clear();
	
	for (auto it = world->RenderObjects().begin(); it != world->RenderObjects().end(); ++it)
	{
		if (this->Visibility()[(*it)->Bounding()->VisibilityIndex()] == CTRUE)
		{
			if ((*it)->GetComponentType()  < ComponentType::EndRenderComponents)
			{
				(*it)->RenderNormals(this->CameraMatrix());			
				visibleObjects++;
			}
		}
	}

	context->BackBuffer().Bind();
	context->BackBuffer().Clear();

	//Ambient Occlusion Pass
	normalTexture.Bind(10);
	tangentTexture.Bind(11);
	depthTexture.Bind(12);

	illuminationBuffer.Bind();
	illuminationBuffer.Clear(GL_COLOR_BUFFER_BIT);

	octree->IlluminationStage()->Process(inverseCameraMatrix, position, illuminationBuffer);

	context->Blend().Enable(GL_ONE, GL_ONE);
	context->DepthTest().TestOnly();
	context->DepthTest().DepthFunc(GL_GREATER);

	normalTexture.Bind(0);
	depthTexture.Bind(1);

	for (auto it = world->Lights().begin(); it != world->Lights().end(); ++it)
	{
		if ((*it)->IsEnabled())
		{
			if ((*it)->Bounding() == nullptr || Visibility()[(*it)->Bounding()->VisibilityIndex()] == CTRUE)
			{
				switch ((*it)->GetComponentType())
				{
				case ComponentType::DirectionalLightComponentType:
					directionalLightEffect.Render((Scene::Lights::DirectionalLight*)(*it)->Light());
					break;
				case ComponentType::PointLightComponentType:
					pointLightEffect.Render((Scene::Lights::PointLight*)(*it)->Light());
					break;
				}
			}
		}
	}
	

	context->CullFace().EnableBackFaceCulling();
	context->Blend().Disable();
	context->DepthTest().TestAndWrite();
	context->DepthTest().DepthFunc(GL_LEQUAL);

	context->BackBuffer().Bind();
	context->BackBuffer().Clear();

	diffuseIllumination.Bind(0);
	specularIllumination.Bind(1);

	context->PolygonOffset().Enable(0.0f, -1.0f);

	for (auto it = world->RenderObjects().begin(); it != world->RenderObjects().end(); ++it)
	{
		if (Visibility()[(*it)->Bounding()->VisibilityIndex()] == CTRUE && (*it)->GetComponentType() < ComponentType::EndRenderComponents)
		{
			if (!((MeshRenderComponent*)(*it))->Material()->IsTransparent())
			{
				(*it)->Render(this);
			}
		}
	}

	context->PolygonOffset().Disable();


	//GI DebugOutput
	if (octree != nullptr && octree->DebugRenderStage() != nullptr)
		octree->DebugRenderStage()->Process(cameraMatrix);	

	cppStopWatch.Stop();
	glStopWatch.Stop();
}

inline const Math::Mat4f& GICamera::CameraMatrix() const
{
	return cameraMatrix;
}

inline const Math::Mat4f& GICamera::InverseCameraMatrix() const
{
	return inverseCameraMatrix;
}

inline const Math::Vec4f& GICamera::Position() const
{
	return position;
}

const Boundings::Frustum& GICamera::Frustum() const
{
	return frustum;
}

const unsigned int GICamera::VisibleObjects() const
{
	return visibleObjects;
}

CameraType GICamera::GetType() const
{
	return CameraType::GICameraType;
}

void GICamera::Resize(unsigned int _width, unsigned int _height)
{
	frustum.Resize(fov, (float)_width / (float)_height, nearPlane, farPlane);
	projectionMatrix = Mat4f::Persp(fov, (float)_width / (float)_height, nearPlane, farPlane);

	deferredBuffer.Resize(_width, _height);
	illuminationBuffer.Resize(_width, _height);
}

float* GICamera::ReadBuffer(unsigned int& _size)
{
	return nullptr;
	//context->BackBuffer().ReadBuffer(_size);
}

inline float GICamera::OpenGLTime()
{
	return glStopWatch.Time();
}
inline float GICamera::CppTime()
{
	return cppStopWatch.Time();
}
inline const Mat4f& GICamera::OctreeMatrix() const
{
	return octree->MatrixZ();
}