#define DLL_EXPORT

#include "RenderCamera.hpp"
#include "../SceneObject.hpp"
#include "../World.hpp"
#include "../Boundings/BoxBSPNode.hpp"
#include "../../EngineManager.hpp"
#include "../../Core/Control/RenderControl.hpp"
#include "../Components/LightComponents/LightComponent.hpp"
#include "../Components/BoundingComponent.hpp"
#include "GICamera.hpp"
#include "../../Graphic/GI/Rendering/IRenderStage.hpp"
#include "../../Graphic/BasicBodies/Plane.hpp"
#include "../../Graphic/Effects/DirectionalLightEffect.hpp"
#include "../../Graphic/Effects/SpotLightEffect.hpp"
#include "../../Graphic/Effects/PointLightEffect.hpp"


#define _USE_MATH_DEFINES
#include <math.h>

#include "../Components/MeshRenderComponent.hpp"
#include "../../Graphic/GI/Generation/OctreeGenerationStage.hpp"
#include "../../Graphic/GI/Preparation/OctreePreparationStage.hpp"

using namespace Engine;
using namespace Engine::Util;
using namespace Engine::Math;
using namespace Engine::Scene;
using namespace Engine::Scene::Boundings;
using namespace Engine::Scene::Components;
using namespace Engine::Scene::Cameras;
using namespace Engine::Scene::Lights;
using namespace Engine::Graphic::Shaders;
using namespace Engine::Graphic::GI;
using namespace Engine::Graphic::GI::Generation;
using namespace Engine::Graphic::GI::Preparation;

RenderCamera::RenderCamera(float _fov, unsigned int _width, unsigned int _height, float _nearPlane, float _farPlane, Scene::SceneObject* _sceneObject,
						   Graphic::GraphicContext* _context, EngineManager* _manager)
	: Camera(_sceneObject, _manager),
	nearPlane(_nearPlane), farPlane(_farPlane), fov(_fov), glStopWatch(),
	frustum(_fov, (float)_width / (float)_height, _nearPlane, _farPlane),
	projectionMatrix(Mat4f::Persp(_fov, (float)_width / (float)_height, _nearPlane, _farPlane)),
	context(_context), cppStopWatch("cppStopWatch"),

	directionalLightEffect(*this, _context),
	pointLightEffect(*this, _context),
    spotLightEffect(*this, _context),
	tonemapEffect(_context, _width, _height, hdrTexture, _context->BackBuffer()),

	normalTexture(GL_RGBA16F, GL_RGBA, false), depthTexture(GL_R32F, GL_RED, false), depthBufferTexture(GL_DEPTH_COMPONENT24, GL_DEPTH_COMPONENT, false),
	deferredBuffer(_context, _width, _height),
	diffuseTexture(GL_RGB16F, GL_RGB, false), specularTexture(GL_RGB16F, GL_RGB, false),
	lightAccumulationBuffer(_context, _width, _height),
	hdrTexture(GL_RGB16F, GL_RGB, false), hdrBuffer(_context, _width, _height)
{
	deferredBuffer.AttachTexture(&normalTexture, GL_COLOR_ATTACHMENT0);
	deferredBuffer.AttachTexture(&depthTexture, GL_COLOR_ATTACHMENT2);
	deferredBuffer.AttachTexture(&depthBufferTexture, GL_DEPTH_ATTACHMENT);
	deferredBuffer.Init();

	lightAccumulationBuffer.AttachTexture(&diffuseTexture, GL_COLOR_ATTACHMENT0);
	lightAccumulationBuffer.AttachTexture(&specularTexture, GL_COLOR_ATTACHMENT1);
	lightAccumulationBuffer.AttachTexture(&depthBufferTexture, GL_DEPTH_ATTACHMENT);
	lightAccumulationBuffer.Init();

	hdrBuffer.AttachTexture(&hdrTexture, GL_COLOR_ATTACHMENT0);
	hdrBuffer.AttachTexture(&depthBufferTexture, GL_DEPTH_ATTACHMENT);
	hdrBuffer.Init();

	hdrTexture.SetFilter(GL_LINEAR_MIPMAP_NEAREST, GL_LINEAR);

	Resize(_width, _height);
}

RenderCamera::~RenderCamera()
{ }

void RenderCamera::Update(const Math::Mat4f& _matrix)
{
	cameraMatrix = projectionMatrix * Mat4f::Invert(_matrix);
	inverseCameraMatrix = Mat4f::Invert(cameraMatrix);
	position = Vec4f(_matrix.M14(), _matrix.M24(), _matrix.M34(), 1.0f);

	Vec4f view = Math::Vec4f::Normalize3(Vec4f(-_matrix.M13(), -_matrix.M23(), -_matrix.M33(), 0.0f));
	Vec4f up = Math::Vec4f::Normalize3(Vec4f(_matrix.M12(), _matrix.M22(), _matrix.M32(), 0.0f));
	Vec4f right = Vec4f::Cross3(view, up);

	frustum.FromCameraParameter(position, right, up, view);
}

void RenderCamera::Render()
{
	cppStopWatch.Start();
	glStopWatch.Start();

	visibleObjects = 0;
	World* world = SceneObject()->World();

	context->Blend().Disable();
	context->DepthTest().TestAndWrite();
	context->DepthTest().DepthFunc(GL_LESS);
	context->CullFace().EnableBackFaceCulling();
	context->StencilTest().Disable();

	deferredBuffer.Bind();
	deferredBuffer.Clear();

	//Normal render pass
	for (auto it = world->RenderObjects().begin(); it != world->RenderObjects().end(); ++it)
	{
		if (this->Visibility()[(*it)->Bounding()->VisibilityIndex()] == CTRUE)
		{
			if ((*it)->GetComponentType()  < ComponentType::EndRenderComponents)
			{
				if (!((MeshRenderComponent*)(*it))->Material()->IsTransparent())
				{
					(*it)->RenderNormals(cameraMatrix);			
					visibleObjects++;
				}
			}
		}
	}

	lightAccumulationBuffer.Bind();
	lightAccumulationBuffer.Clear(GL_COLOR_BUFFER_BIT);
	const std::vector<Scene::Components::LightComponents::LightComponent*>& lights = SceneObject()->World()->Lights();	

	context->Blend().Enable(GL_ONE, GL_ONE);
	context->DepthTest().TestOnly();
	context->DepthTest().DepthFunc(GL_GREATER);

	normalTexture.Bind(0);
	depthTexture.Bind(1);

	for (auto it = world->Lights().begin(); it != world->Lights().end(); ++it)
	{
		if ((*it)->IsEnabled())
		{
			if ((*it)->Bounding() == nullptr || Visibility()[(*it)->Bounding()->VisibilityIndex()] == CTRUE)
			{
				switch ((*it)->GetComponentType())
				{
				case ComponentType::DirectionalLightComponentType:
					directionalLightEffect.Render((Scene::Lights::DirectionalLight*)(*it)->Light());
					break;
				case ComponentType::PointLightComponentType:
					pointLightEffect.Render((Scene::Lights::PointLight*)(*it)->Light());
					break;
				case ComponentType::SpotLightComponentType:
					spotLightEffect.Render((Scene::Lights::SpotLight*)(*it)->Light());
					break;
				}
			}
		}
	}

	context->CullFace().EnableBackFaceCulling();
	context->Blend().Disable();
	context->DepthTest().TestAndWrite();
	context->DepthTest().DepthFunc(GL_LEQUAL);

	hdrBuffer.Bind();
	hdrBuffer.Clear(GL_COLOR_BUFFER_BIT);

	diffuseTexture.Bind(0);
	specularTexture.Bind(1);

	context->PolygonOffset().Enable(0.0f, -1.0f);

	for (auto it = world->RenderObjects().begin(); it != world->RenderObjects().end(); ++it)
	{
		if (Visibility()[(*it)->Bounding()->VisibilityIndex()] == CTRUE && (*it)->GetComponentType()  < ComponentType::EndRenderComponents)
		{
			if (!((MeshRenderComponent*)(*it))->Material()->IsTransparent())
			{
				(*it)->Render(this);
			}
		}
	}

	context->PolygonOffset().Disable();

	//Transparent Objects
	if (SceneObject()->World()->TransparentObjectBsp().size() > 0)
	{
		context->DepthTest().TestOnly();
		context->Blend().Disable();
		context->CullFace().Disable();
	
		RenderTransparent(SceneObject()->World()->TransparentObjectBsp()[0], visibleObjects);
	}
	
	//Post Processing
	tonemapEffect.Render(Manager()->Clock().RelativeTime());
	

#pragma region Old render code
	/*visibleObjects = 0;
	World* world = SceneObject()->World();

	context->Blend().Disable();
	context->DepthTest().TestAndWrite();
	context->DepthTest().DepthFunc(GL_LESS);
	context->CullFace().EnableBackFaceCulling();
	context->StencilTest().Disable();

	deferredBuffer.Bind();
	deferredBuffer.Clear();

	//Normal render pass
	for (auto it = world->RenderObjects().begin(); it != world->RenderObjects().end(); ++it)
	{
		if (this->Visibility()[(*it)->Bounding()->VisibilityIndex()] == CTRUE)
		{
			if ((*it)->GetComponentType()  < ComponentType::EndRenderComponents)
			{
				if (!((MeshRenderComponent*)(*it))->Material()->IsTransparent())
				{
					(*it)->RenderNormals(this);			
					visibleObjects++;
				}
			}
		}
	}

	//Light Accumulation pass
	lightAccumulationBuffer.Bind();
	lightAccumulationBuffer.Clear();
	const std::vector<Scene::Components::LightComponents::LightComponent*>& lights = SceneObject()->World()->Lights();	

	context->Blend().Enable(GL_ONE, GL_ONE);
	context->DepthTest().TestOnly();
	context->DepthTest().DepthFunc(GL_GREATER);

	deferredBuffer.NormalTexture().Bind(0);
	deferredBuffer.DepthTexture().Bind(1);

	for (auto it = world->Lights().begin(); it != world->Lights().end(); ++it)
	{
		if ((*it)->IsEnabled())
		{
			if ((*it)->Bounding() == nullptr || Visibility()[(*it)->Bounding()->VisibilityIndex()] == CTRUE)
			{
				switch ((*it)->GetComponentType())
				{
				case ComponentType::DirectionalLightComponentType:
					directionalLightEffect.Render((Scene::Lights::DirectionalLight*)(*it)->Light());
					break;
				case ComponentType::PointLightComponentType:
					pointLightEffect.Render((Scene::Lights::PointLight*)(*it)->Light());
					break;
				case ComponentType::SpotLightComponentType:
					spotLightEffect.Render((Scene::Lights::SpotLight*)(*it)->Light());
					break;
				}
			}
		}
	}

	//Material shading pass

	context->CullFace().EnableBackFaceCulling();
	context->Blend().Disable();
	hdrBuffer.Bind();
	hdrBuffer.Clear();
	context->DepthTest().TestAndWrite();
	context->DepthTest().DepthFunc(GL_LEQUAL);

	lightAccumulationBuffer.DiffuseTexture().Bind(0);
	lightAccumulationBuffer.SpecularTexture().Bind(1);

	context->PolygonOffset().Enable(0.0f, -1.0f);

	for (auto it = world->RenderObjects().begin(); it != world->RenderObjects().end(); ++it)
	{
		if (Visibility()[(*it)->Bounding()->VisibilityIndex()] == CTRUE && (*it)->GetComponentType()  < ComponentType::EndRenderComponents)
		{
			if (!((MeshRenderComponent*)(*it))->Material()->IsTransparent())
			{
				(*it)->Render(this);
			}
		}
	}

	context->PolygonOffset().Disable();

	//GI DebugOutput
	if (giCamera != nullptr && giCamera->RenderStage() != nullptr)
		giCamera->RenderStage()->Process(cameraMatrix);	

	//Render DebugInformation
	//Lights
	if (Manager()->RenderControl()->ShowLightVolumes())
		for (auto it = world->Lights().begin(); it != world->Lights().end(); ++it)
			if ((*it)->IsEnabled() && ((*it)->Bounding() == nullptr || Visibility()[(*it)->Bounding()->VisibilityIndex()]) == CTRUE)
				(*it)->RenderDebug(this->CameraMatrix());

	//BoundingBoxes
	if (Manager()->RenderControl()->ShowBoundings())
		for (auto it = world->RenderObjects().begin(); it != world->RenderObjects().end(); ++it)
			if (Visibility()[(*it)->Bounding()->VisibilityIndex()] == CTRUE)
				(*it)->RenderDebug(this);

	//Transparent Objects
	if (SceneObject()->World()->TransparentObjectBsp().size() > 0)
	{
		context->DepthTest().TestOnly();
		context->Blend().Disable();
		context->CullFace().Disable();
	
		RenderTransparent(SceneObject()->World()->TransparentObjectBsp()[0], visibleObjects);
	}

	//Tonemapping
	tonemapEffect.renderTonemapperInput(&hdrBuffer, Manager()->Clock().RelativeTime());
	finalScreenBuffer.Bind();
	tonemapEffect.Render(&hdrBuffer, Manager()->RenderControl()->ShowTonemapper(), Manager()->RenderControl()->ShowBloom());

	if (Manager()->RenderControl()->ShowFXAA())
	{
		context->BackBuffer().Bind();
		fxaaEffect.Render(finalScreenBuffer.ColorTexture());
	}
	else
	{
		finalScreenBuffer.CopyToBackbuffer();
		context->BackBuffer().Bind();
	}

	context->DepthTest().TestAndWrite();*/

#pragma endregion

	cppStopWatch.Stop();
	glStopWatch.Stop();
}

void RenderCamera::RenderTransparent(const BoxBspNode& _node, unsigned int& _count)
{
	if (_node.Index() != -1)
	{
		RenderComponent* rc = SceneObject()->World()->RenderObjects()[_node.Index()];

		if (Visibility()[rc->Bounding()->VisibilityIndex()])
		{
			rc->Render(this);
			_count ++;
		}
	}
	else
	{
		float d = Vec4f::Dot(_node.Plane(), this->Position());

		if (d >= 0.0)
		{
			if (_node.Back() != -1)
				RenderTransparent(SceneObject()->World()->TransparentObjectBsp()[_node.Back()], _count);

			if (_node.Front() != -1)
				RenderTransparent(SceneObject()->World()->TransparentObjectBsp()[_node.Front()], _count);
		}
		else
		{
			if (_node.Front() != -1)
				RenderTransparent(SceneObject()->World()->TransparentObjectBsp()[_node.Front()], _count);

			if (_node.Back() != -1)
				RenderTransparent(SceneObject()->World()->TransparentObjectBsp()[_node.Back()], _count);
		}
	}
}

inline const Math::Mat4f& RenderCamera::CameraMatrix() const
{
	return cameraMatrix;
}

inline const Math::Mat4f& RenderCamera::InverseCameraMatrix() const
{
	return inverseCameraMatrix;
}

inline const Math::Vec4f& RenderCamera::Position() const
{
	return position;
}

const Boundings::Frustum& RenderCamera::Frustum() const
{
	return frustum;
}

const unsigned int RenderCamera::VisibleObjects() const
{
	return visibleObjects;
}

CameraType RenderCamera::GetType() const
{
	return CameraType::RenderCameraType;
}

void RenderCamera::Resize(unsigned int _width, unsigned int _height)
{
	frustum.Resize(fov, (float)_width / (float)_height, nearPlane, farPlane);
	projectionMatrix = Mat4f::Persp(fov, (float)_width / (float)_height, nearPlane, farPlane);

	deferredBuffer.Resize(_width, _height);
	lightAccumulationBuffer.Resize(_width, _height);
	hdrBuffer.Resize(_width, _height);

	tonemapEffect.Resize(_width, _height);
}

float* RenderCamera::ReadBuffer(unsigned int& _size)
{
	//context->BackBuffer().ReadBuffer(_size);
	return nullptr;
}

inline float RenderCamera::OpenGLTime()
{
	return glStopWatch.Time();
}
inline float RenderCamera::CppTime()
{
	return cppStopWatch.Time();
}
/*inline const Mat4f& RenderCamera::OctreeMatrix() const
{
	return Mat4f();
}*/