#define DLL_EXPORT

#include "MeshRenderComponent.hpp"

#include "../SceneObject.hpp"
#include "../../Core/EventHandling/EventHandler.hpp"
#include "../../Core/System/SystemConfiguration.hpp"
#include "../Cameras/Camera.hpp"
#include "ComponentChangedEventArgs.hpp"
#include "PositionComponent.hpp"
#include "BoundingComponent.hpp"
#include "../../Graphic/Shaders/Shader.hpp"
#include "../../Graphic/BasicBodies/LineCube.hpp"

using namespace Engine;
using namespace Engine::Graphic;
using namespace Engine::Graphic::Meshes;
using namespace Engine::Graphic::Materials;
using namespace Engine::Util;
using namespace Engine::Scene;
using namespace Engine::Scene::Components;
using namespace Engine::Scene::Cameras;
using namespace Engine::Core::EventHandling;
using namespace Engine::Scene::Boundings;
using namespace Engine::Core::System;


MeshRenderComponent::MeshRenderComponent(Scene::SceneObject* _sceneObject, SharedPointer<Mesh> _mesh, SharedPointer<Materials::Material> _material, GraphicContext* _context) :
	RenderComponent(_sceneObject, false), mesh(_mesh), material(_material), positionComponent(_sceneObject->GetComponent<PositionComponent>()), context(_context)
{
	SceneObject()->ComponentChanged += new EventHandler<MeshRenderComponent, ComponentChangedEventArgs>(this, &MeshRenderComponent::ComponentChanged);
	bounding = SceneObject()->GetComponent<BoundingComponent>();

	if (bounding == nullptr)
	{
		bounding = new BoundingComponent(_sceneObject, (BoundingObject*)&(mesh->BoundingBox()));
		_sceneObject->AddComponent(bounding);
	}
}

MeshRenderComponent::~MeshRenderComponent()
{
	SceneObject()->ComponentChanged -= new EventHandler<MeshRenderComponent, ComponentChangedEventArgs>(this, &MeshRenderComponent::ComponentChanged);
}

inline ComponentType MeshRenderComponent::GetComponentType()
{
	return MeshRenderComponentType;
}

inline ComponentType MeshRenderComponent::ComponentTypeOf()
{
	return MeshRenderComponentType;
}

void MeshRenderComponent::UpdateComponent()
{ }

void MeshRenderComponent::Render(Camera* _camera)
{
	if (IsEnabled())
	{
		if (positionComponent != nullptr)
		{
			material->StartActivate(ShadeMode::Deferred);
			material->Shader()->UniformVariable<Math::Mat4f>("modelMatrix") = positionComponent->WorldMatrix();
			material->Shader()->UniformArrays<Math::Mat4f>("cameraMatrixArray")[0] = _camera->CameraMatrix();
			material->FinishActivate(ShadeMode::Deferred);

			mesh->Draw(material->Shader()->PrimitiveType());

			material->Deactivate();
		}
		else
			Logger::Error("MeshRenderComponent without PositionComponent detected");
	}
}

void MeshRenderComponent::RenderOctree(const Math::Mat4f& _cameraMatrixX, const Math::Mat4f& _cameraMatrixY, const Math::Mat4f& _cameraMatrixZ)
{
	if (IsEnabled() && !material->IsShadeless())
	{
		if (positionComponent != nullptr)
		{
			//Ocree
			material->StartActivate(ShadeMode::Octree);
			material->OctreeShader()->UniformVariable<Math::Mat4f>("modelMatrix") = positionComponent->WorldMatrix();
			material->OctreeShader()->UniformArrays<Math::Mat4f>("cameraMatrixArray")[0] = _cameraMatrixX;
			material->OctreeShader()->UniformArrays<Math::Mat4f>("cameraMatrixArray")[1] = _cameraMatrixY;
			material->OctreeShader()->UniformArrays<Math::Mat4f>("cameraMatrixArray")[2] = _cameraMatrixZ;
			material->OctreeShader()->UniformVariable<GLint>("linkBuffer") = 1;
			material->OctreeShader()->UniformVariable<GLint>("lockBuffer") = 2;
			material->OctreeShader()->UniformVariable<GLint>("deferredBuffer") = 3;
			
			material->FinishActivate(ShadeMode::Octree);

			mesh->Draw(material->OctreeShader()->PrimitiveType());

			material->Deactivate();
		}
		else
			Logger::Error("MeshRenderComponent without PositionComponent detected");
	}
} 

void MeshRenderComponent::RenderNormals(const Math::Mat4f& _cameraMatrix)
{
	if (IsEnabled() && positionComponent != nullptr)
	{
		this->context->Blend().Disable();
		material->StartActivate(ShadeMode::Normals);
		material->NormalShader()->UniformVariable<Math::Mat4f>("modelMatrix") = positionComponent->WorldMatrix();
		material->NormalShader()->UniformArrays<Math::Mat4f>("cameraMatrixArray")[0] = _cameraMatrix;

		material->FinishActivate(ShadeMode::Normals);
		mesh->Draw(material->NormalShader()->PrimitiveType());
	}
}

void MeshRenderComponent::RenderDebug(const Math::Mat4f& _cameraMatrix)
{
	if (IsEnabled())
	{
		context->DepthTest().TestAndWrite();
		context->LineWidth().Set(2.0f);

		context->DebugLineShader()->Activate();
		context->DebugLineShader()->UniformVariable<Math::Mat4f>("modelMatrix") = positionComponent->WorldMatrix() * Math::Mat4f::Scale(Math::Mat4f::Translate(mesh->BoundingBox().Location().GetVec3f()), mesh->BoundingBox().Size());
		context->DebugLineShader()->UniformArrays<Math::Mat4f>("cameraMatrixArray")[0] = _cameraMatrix;
		context->DebugLineShader()->UniformVariable<Math::Vec3f>("color") = SystemConfiguration::Get<Math::Vec3f>("LineColorBounding");
		context->DebugLineShader()->ChangeValues();

		context->LineCube()->Draw();
	}
}

void MeshRenderComponent::ComponentChanged(void* _sender, ComponentChangedEventArgs& _args)
{
	if (_args.Component()->GetComponentType() == ComponentType::PositionComponentType)
	{
		positionComponent = SceneObject()->GetComponent<PositionComponent>();
	}
}

Engine::Util::SharedPointer<Engine::Graphic::Materials::Material> MeshRenderComponent::Material() const
{
	return material;
}

void MeshRenderComponent::Material(Engine::Util::SharedPointer<Engine::Graphic::Materials::Material> _value)
{
	material = _value;
}

inline BoundingComponent* MeshRenderComponent::Bounding() const
{
	return bounding;
}