#define DLL_EXPORT

#include "PositionComponent.hpp"
#include "../SceneObject.hpp"
#include "../../Core/EventHandling/EventHandler.hpp"
#include "ComponentChangedEventArgs.hpp"
#include "../ParentChangedEventArgs.hpp"

using namespace Engine::Scene;
using namespace Engine::Scene::Components;
using namespace Engine::Math;
using namespace Engine::Core::EventHandling;

PositionComponent::PositionComponent(Engine::Scene::SceneObject* _sceneObject) : IComponent(_sceneObject, false), matrix(), parentPositionComponent(nullptr), worldPosition(), needUpdate(true)
{
	SceneObject()->ParentChanged += new EventHandler<PositionComponent, ParentChangedEventArgs>(this, &PositionComponent::ParentChanged);

	if (SceneObject()->Parent() != nullptr)
	{
		parentPositionComponent = SceneObject()->Parent()->GetComponent<PositionComponent>();
		SceneObject()->Parent()->ComponentChanged += new EventHandler<PositionComponent, ComponentChangedEventArgs>(this, &PositionComponent::ComponentsChanged);
	}
}

PositionComponent::~PositionComponent()
{
	SceneObject()->ParentChanged -= new EventHandler<PositionComponent, ParentChangedEventArgs>(this, &PositionComponent::ParentChanged);

	if (SceneObject()->Parent() != nullptr)
		SceneObject()->Parent()->ComponentChanged -= new EventHandler<PositionComponent, ComponentChangedEventArgs>(this, &PositionComponent::ComponentsChanged);
}

inline ComponentType PositionComponent::GetComponentType()
{
	return PositionComponent::ComponentTypeOf();
}

inline ComponentType PositionComponent::ComponentTypeOf()
{
	return PositionComponentType;
}

void PositionComponent::UpdateComponent()
{
	if (parentPositionComponent != nullptr)
	{
		worldMatrix = parentPositionComponent->WorldMatrix() * matrix;
	}
	else
	{
		worldMatrix = matrix;
	}

	worldPosition = Math::Vec3f(worldMatrix.M41(), worldMatrix.M42(), worldMatrix.M43());
	needUpdate = false;
}

inline const Mat4f& PositionComponent::Matrix() const
{
	return matrix;
}

inline PositionComponent& PositionComponent::Matrix(Mat4f& _matrix)
{
	matrix = _matrix;
	NeedUpdate(true);
	return *this;
}

inline const Mat4f& PositionComponent::WorldMatrix() const
{
	return worldMatrix;
}

inline const Vec3f& PositionComponent::WorldPosition() const
{
	return worldPosition;
}

void PositionComponent::UpdateRecursive()
{
	for (auto it = SceneObject()->Children().begin(); it != SceneObject()->Children().end(); ++it)
	{
		PositionComponent* pc = (*it)->GetComponent<PositionComponent>();
		pc->Update();
		pc->UpdateRecursive();
	}
}

void PositionComponent::ComponentsChanged(void* _sender, ComponentChangedEventArgs& _args)
{
	if (_args.Component()->GetComponentType() == ComponentType::PositionComponentType)
	{
		if (_args.ActionType() == ComponentActionType::ADD)
			parentPositionComponent = (PositionComponent*)_args.Component();
		else
			parentPositionComponent = nullptr;
	}
}

void PositionComponent::ParentChanged(void* _sender, ParentChangedEventArgs& _args)
{
	if (_args.OldValue() != nullptr)
		_args.OldValue()->ComponentChanged -= new EventHandler<PositionComponent, ComponentChangedEventArgs>(this, &PositionComponent::ComponentsChanged);

	parentPositionComponent = _args.NewValue()->GetComponent<PositionComponent>();

	if (_args.NewValue() != nullptr)
		_args.NewValue()->ComponentChanged += new EventHandler<PositionComponent, ComponentChangedEventArgs>(this, &PositionComponent::ComponentsChanged);
}

inline const unsigned int PositionComponent::StorageIndex() const
{
	return storageIndex;
}

void PositionComponent::StorageIndex(unsigned int _index)
{
	storageIndex = _index;
}

const PositionComponent* PositionComponent::ParentPositionComponent() const
{
	return parentPositionComponent;
}

const bool PositionComponent::NeedUpdate() const
{
	return needUpdate;
}

void PositionComponent::NeedUpdate(bool _value)
{
	if (_value && !needUpdate)
	{
		auto childList = SceneObject()->Children();

		for (auto it = childList.begin(); it != childList.end(); ++it)
			(*it)->GetComponent<PositionComponent>()->NeedUpdate(true);
	}

	needUpdate = _value;
}