#define DLL_EXPORT

#include "TransparentMeshRenderComponent.hpp"

#include "../SceneObject.hpp"
#include "../../Core/EventHandling/EventHandler.hpp"
#include "../../Core/System/SystemConfiguration.hpp"
#include "LightComponents\LightComponent.hpp"
#include "LightComponents\DirectionalLightComponent.hpp"
#include "LightComponents\PointLightComponent.hpp"
#include "LightComponents\SpotLightComponent.hpp"
#include "BoundingComponent.hpp"
#include "../World.hpp"
#include "../Cameras/Camera.hpp"
#include "../../Util/WeightedContainer.hpp"
#include "CameraComponent.hpp"
#include "PositionComponent.hpp"
#include "ComponentChangedEventArgs.hpp"
#include "../../Graphic/Shaders/Shader.hpp"
#include "../../Graphic/BasicBodies/LineCube.hpp"
#include "../Lights/SpotLight.hpp"
#include "../Lights/PointLight.hpp"
#include "../Lights/DirectionalLight.hpp"

#include <algorithm>

using namespace Engine;
using namespace Engine::Math;
using namespace Engine::Graphic;
using namespace Engine::Graphic::Meshes;
using namespace Engine::Graphic::Shaders;
using namespace Engine::Graphic::Materials;
using namespace Engine::Util;
using namespace Engine::Scene;
using namespace Engine::Scene::Cameras;
using namespace Engine::Scene::Lights;
using namespace Engine::Scene::Components;
using namespace Engine::Scene::Components::LightComponents;
using namespace Engine::Core::EventHandling;
using namespace Engine::Scene::Boundings;
using namespace Engine::Core::System;


TransparentMeshRenderComponent::TransparentMeshRenderComponent(Scene::SceneObject* _sceneObject, SharedPointer<TransparentMesh> _mesh, SharedPointer<Materials::Material> _material, GraphicContext* _context) :
	RenderComponent(_sceneObject), mesh(_mesh), material(_material), positionComponent(_sceneObject->GetComponent<PositionComponent>()), context(_context), isVisible(false)
{
	SceneObject()->ComponentChanged += new EventHandler<TransparentMeshRenderComponent, ComponentChangedEventArgs>(this, &TransparentMeshRenderComponent::ComponentChanged);
	bounding = SceneObject()->GetComponent<BoundingComponent>();

	if (bounding == nullptr)
	{
		bounding = new BoundingComponent(_sceneObject, (BoundingObject*)&(mesh->BoundingBox()));
		_sceneObject->AddComponent(bounding);
	}

	indexArray = new unsigned int[mesh->Count()];
}

TransparentMeshRenderComponent::~TransparentMeshRenderComponent()
{
	SceneObject()->ComponentChanged -= new EventHandler<TransparentMeshRenderComponent, ComponentChangedEventArgs>(this, &TransparentMeshRenderComponent::ComponentChanged);
	delete indexArray;
}

inline ComponentType TransparentMeshRenderComponent::GetComponentType()
{
	return TransparentMeshRenderComponentType;
}

inline ComponentType TransparentMeshRenderComponent::ComponentTypeOf()
{
	return TransparentMeshRenderComponentType;
}

void TransparentMeshRenderComponent::UpdateComponent()
{
	if (IsEnabled())
	{
		Vec4f camPos = SceneObject()->World()->ScreenCamera()->Camera()->Position();
		//Calculate Position in ModelSpace
		Vec4f modelCam = Mat4f::Invert(positionComponent->WorldMatrix()) * camPos;
		unsigned int* indexPointer = indexArray;

		mesh->BuildIndices(&indexPointer, modelCam);
	}
}

void TransparentMeshRenderComponent::Render(Camera* _camera)
{
	if (IsEnabled())
	{
		if (positionComponent != nullptr)
		{
			World* world = SceneObject()->World();

			material->StartActivate(ShadeMode::Forward);
			Shader* sh = material->ForwardShader().Get();
			sh->UniformVariable<Math::Mat4f>("modelMatrix") = positionComponent->WorldMatrix();
			sh->UniformArrays<Math::Mat4f>("cameraMatrixArray")[0] = _camera->CameraMatrix();

			//Find best Lights
			float bestDirectionalLightInfluence = 0.0f;
			DirectionalLightComponent* bestDirectionalLight = nullptr;
			std::vector<WeightedContainer<PointLight*>> bestPointLights;
			std::vector<WeightedContainer<SpotLight*>> bestSpotLights;

			for (auto it = world->Lights().begin(); it != world->Lights().end(); ++it)
			{
				LightComponent* lc = (*it);

				if (lc->IsEnabled())
				{
					if (lc->Bounding() == nullptr/* || _camera->Visibility()[lc->Bounding()->VisibilityIndex()]*/)
					{
						switch (lc->GetComponentType())
						{
						case ComponentType::DirectionalLightComponentType:
							{
								float infl = lc->Light()->GetInfluence(positionComponent->WorldPosition());
								if (infl > bestDirectionalLightInfluence)
									bestDirectionalLight = (DirectionalLightComponent*)lc;
							}
							break;
						case ComponentType::PointLightComponentType:
							{
								PointLight* pl = (PointLight*)lc->Light();
								bestPointLights.push_back(WeightedContainer<PointLight*>(pl, pl->GetInfluence(positionComponent->WorldPosition())));
							}
							break;
						case ComponentType::SpotLightComponentType:
							{
								SpotLight* sl = (SpotLight*)lc->Light();
								bestSpotLights.push_back(WeightedContainer<SpotLight*>(sl, sl->GetInfluence(positionComponent->WorldPosition())));
							}
							break;
						}
					}
				}
			}

			sh->UniformVariable<Math::Vec3f>("cameraPosition") = _camera->Position();

			if (bestDirectionalLight != nullptr)
			{
				sh->UniformVariable<GLboolean>("useDirectionalLight") = true;
				sh->UniformVariable<Vec3f>("directionalLightDirection") = ((DirectionalLight*)bestDirectionalLight->Light())->WorldDirection();
				sh->UniformVariable<Vec3f>("directionalLightIntensity") = ((DirectionalLight*)bestDirectionalLight->Light())->Intensity();
			}
			else
				sh->UniformVariable<GLboolean>("useDirectionalLight") = false;

			//Best n point lights
			size_t maxPLight = min(sh->UniformArrays<Vec3f>("pointlightPosition").Count(), bestPointLights.size());
			std::nth_element(bestPointLights.begin(), bestPointLights.begin() + maxPLight, bestPointLights.end(), WeightedContainer<PointLight*>::OrderDescending);

			for (size_t i = 0; i < maxPLight; i++)
			{
				sh->UniformArrays<Vec3f>("pointlightPosition")[i] = bestPointLights[i]->WorldPosition();
				sh->UniformArrays<Vec3f>("pointlightIntensity")[i] = bestPointLights[i]->Intensity();
				sh->UniformArrays<Vec3f>("pointlightAttenuation")[i] = bestPointLights[i]->Attenuation();
			}

			sh->UniformVariable<GLint>("pointLightCount") = (GLint)maxPLight;

			//Best n spot lights
			size_t maxSLight = min(sh->UniformArrays<Vec3f>("spotLightPosition").Count(), bestSpotLights.size());
			std::nth_element(bestSpotLights.begin(), bestSpotLights.begin() + maxSLight, bestSpotLights.end(), WeightedContainer<SpotLight*>::OrderDescending);

			for (size_t i = 0; i < maxSLight; i++)
			{
				sh->UniformArrays<Vec3f>("spotLightPosition")[i] = bestSpotLights[i]->WorldPosition();
				sh->UniformArrays<Vec3f>("spotLightDirection")[i] = bestSpotLights[i]->WorldDirection();
				sh->UniformArrays<Vec3f>("spotLightIntensity")[i] = bestSpotLights[i]->Intensity();
				sh->UniformArrays<Vec3f>("spotLightAttenuation")[i] = bestSpotLights[i]->Attenuation();
				sh->UniformArrays<GLfloat>("spotLightAngle")[i] = bestSpotLights[i]->Angle();
				sh->UniformArrays<GLfloat>("spotLightAngularAttenuation")[i] = bestSpotLights[i]->AngularAttenuation();
			}

			sh->UniformVariable<GLint>("spotLightCount") = (GLint)maxSLight;

			material->FinishActivate(ShadeMode::Forward);

			this->context->Blend().Enable(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

			mesh->Draw(material->Shader()->PrimitiveType(), indexArray);

			material->Deactivate();
		}
		else
			Logger::Error("TransparentMeshRenderComponent without PositionComponent detected");
	}
}

void TransparentMeshRenderComponent::RenderNormals(const Math::Mat4f& _cameraMatrix)
{ /* Currently not used */ }

void TransparentMeshRenderComponent::RenderOctree(const Math::Mat4f& _cameraMatrixX, const Math::Mat4f& _cameraMatrixY, const Math::Mat4f& _cameraMatrixZ)
{ /* Currently not used */ }

void TransparentMeshRenderComponent::RenderDebug(const Math::Mat4f& _cameraMatrix)
{
	if (IsEnabled())
	{
		context->DepthTest().TestAndWrite();
		context->LineWidth().Set(2.0f);

		context->DebugLineShader()->Activate();
		context->DebugLineShader()->UniformVariable<Math::Mat4f>("modelMatrix") = positionComponent->WorldMatrix() * Math::Mat4f::Scale(Math::Mat4f::Translate(mesh->BoundingBox().Location().GetVec3f()), mesh->BoundingBox().Size());
		context->DebugLineShader()->UniformVariable<Math::Mat4f>("cameraMatrix") = _cameraMatrix;
		context->DebugLineShader()->UniformVariable<Math::Vec3f>("color") = SystemConfiguration::Get<Vec3f>("LineColorBounding");
		context->DebugLineShader()->ChangeValues();

		context->LineCube()->Draw();
	}
}

void TransparentMeshRenderComponent::ComponentChanged(void* _sender, ComponentChangedEventArgs& _args)
{
	if (_args.Component()->GetComponentType() == ComponentType::PositionComponentType)
	{
		positionComponent = SceneObject()->GetComponent<PositionComponent>();
	}
}

Engine::Util::SharedPointer<Engine::Graphic::Materials::Material> TransparentMeshRenderComponent::Material() const
{
	return material;
}

void TransparentMeshRenderComponent::Material(Engine::Util::SharedPointer<Engine::Graphic::Materials::Material> _value)
{
	material = _value;
}

inline BoundingComponent* TransparentMeshRenderComponent::Bounding() const
{
	return bounding;
}