#define DLL_EXPORT

#include "VoxelOctreeComponent.hpp"
#include "../../Graphic/GI/Generation/IGenerationStage.hpp"
#include "../../Graphic/GI/Illumination/IIlluminationStage.hpp"
#include "../../Graphic/GI/Preparation/IPreparationStage.hpp"
#include "../../Graphic/GI/Rendering/IRenderStage.hpp"
#include "../SceneObject.hpp"
#include "PositionComponent.hpp"

#define _USE_MATH_DEFINES
#include <math.h>

using namespace Engine;
using namespace Engine::Math;
using namespace Engine::Scene;
using namespace Engine::Scene::Components;
using namespace Engine::Graphic;
using namespace Engine::Graphic::GI;
using namespace Engine::Graphic::GI::Generation;
using namespace Engine::Graphic::GI::Preparation;
using namespace Engine::Graphic::GI::Illumination;
using namespace Engine::Graphic::GI::Rendering;

VoxelOctreeComponent::VoxelOctreeComponent(Scene::SceneObject* _sceneObject, float _extend) :
	IComponent(_sceneObject, true), extend(_extend), cppStopWatch("VoxelOctreeComponent.CppTime"),
	generationStage(nullptr), preparationStage(nullptr), illuminationStage(nullptr), debugRenderStage(nullptr), 
	projectionMatrix(Mat4f::Ortho(-_extend, _extend, -_extend, _extend, _extend, -_extend)), firstFrame(true)
{ }

VoxelOctreeComponent::~VoxelOctreeComponent()
{
	if (generationStage != nullptr)
		delete generationStage;
	if (preparationStage != nullptr)
		delete preparationStage;
	if (illuminationStage != nullptr)
		delete illuminationStage;
	if (debugRenderStage != nullptr)
		delete debugRenderStage;
}

Components::ComponentType VoxelOctreeComponent::GetComponentType()
{
	return ComponentType::VoxelOctreeComponentType;
}

Components::ComponentType VoxelOctreeComponent::ComponentTypeOf()
{
	return ComponentType::VoxelOctreeComponentType;
}

void VoxelOctreeComponent::Resize(unsigned int _entryCount)
{ }


void VoxelOctreeComponent::UpdateComponent()
{
	Mat4f matrix = SceneObject()->GetComponent<PositionComponent>()->WorldMatrix();

	position = Vec4f(matrix.M14(), matrix.M24(), matrix.M34(), 1.0f);
    matrixX = projectionMatrix * Mat4f::RotateY((float)-M_PI_2) * Mat4f::Translate(position * -1.0f);
    matrixY = projectionMatrix * Mat4f::RotateX((float)M_PI_2) * Mat4f::Translate(position * -1.0f);
    matrixZ = projectionMatrix * Mat4f::Translate(position * -1.0f);
	inverseMatrixZ = Mat4f::Invert(matrixZ);

    Vec4f view = Math::Vec4f::Normalize3(Vec4f(-matrix.M13(), -matrix.M23(), -matrix.M33(), 0.0f));
    Vec4f up = Math::Vec4f::Normalize3(Vec4f(matrix.M12(), matrix.M22(), matrix.M32(), 0.0f));
    Vec4f right = Vec4f::Cross3(view, up);
}

void VoxelOctreeComponent::Render()
{
	if (firstFrame)
	{
		glStopWatch.Start();
		cppStopWatch.Start();

		generationStage->Process();
		preparationStage->Process();

		glStopWatch.Stop();
		cppStopWatch.Stop();

		firstFrame = false;
	}
}


inline const Math::Mat4f& VoxelOctreeComponent::MatrixX() const
{
	return matrixX;
}

inline const Math::Mat4f& VoxelOctreeComponent::MatrixY() const
{
	return matrixY;
}

inline const Math::Mat4f& VoxelOctreeComponent::MatrixZ() const
{
	return matrixZ;
}

inline const Math::Mat4f& VoxelOctreeComponent::InverseMatrixZ() const
{
	return inverseMatrixZ;
}

inline IGenerationStage* VoxelOctreeComponent::GenerationStage() const
{
	return generationStage;
}

inline void VoxelOctreeComponent::GenerationStage(IGenerationStage* _stage) 
{
	generationStage = _stage;
}

inline IPreparationStage* VoxelOctreeComponent::PreparationStage() const
{
	return preparationStage;
}

inline void VoxelOctreeComponent::PreparationStage(IPreparationStage* _stage) 
{
	preparationStage = _stage;
}

inline IIlluminationStage* VoxelOctreeComponent::IlluminationStage() const
{
	return illuminationStage;
}

inline void VoxelOctreeComponent::IlluminationStage(IIlluminationStage* _stage) 
{
	illuminationStage = _stage;
}

inline IRenderStage* VoxelOctreeComponent::DebugRenderStage() const
{
	return debugRenderStage;
}

inline void VoxelOctreeComponent::DebugRenderStage(IRenderStage* _stage) 
{
	debugRenderStage = _stage;
}

inline float VoxelOctreeComponent::CppTime()
{
	return cppStopWatch.Time();
}

inline float VoxelOctreeComponent::OpenGLTime()
{
	return glStopWatch.Time();
}

inline Graphic::GI::GIData* VoxelOctreeComponent::Data() const
{
	return data;
}

inline void VoxelOctreeComponent::Data(Graphic::GI::GIData* _data)
{
	data = _data;
}