#define DLL_EXPORT

#include "SceneObject.hpp"
#include "Components/IComponent.hpp"
#include <assert.h>
#include <algorithm>
#include "World.hpp"
#include "Components\PositionComponent.hpp"
#include "Components\ComponentChangedEventArgs.hpp"
#include "ParentChangedEventArgs.hpp"

using namespace Engine;
using namespace Engine::Scene;
using namespace Engine::Scene::Components;
using namespace Engine::Core::EventHandling;

SceneObject::SceneObject(Scene::World* _world, std::string _name) : parent(nullptr), children(), components(), subTreeCount(0), world(_world), name(_name)
{
	AddComponent(new PositionComponent(this));
	world->RegisterSceneObject(this);
}

SceneObject::~SceneObject()
{
	//Alle Children lschen
	for (auto it = children.begin(); it != children.end(); ++it)
	{
		delete (*it);
	}

	children.clear();

	//Componenten lschen
	for (auto it = components.begin(); it != components.end(); ++it)
	{
		IComponent* x = it->second;
		ComponentChanged(this, ComponentChangedEventArgs(x, ComponentActionType::REMOVE));
		delete it->second;
	}

	components.clear();

	world->RemoveSceneObject(this);
}

inline const SceneObject::ChildList& SceneObject::Children() const
{
	return this->children;
}

inline Scene::World* SceneObject::World() const
{
	return world;
}

void SceneObject::AddChild(SceneObject* _child)
{
	children.push_back(_child);
	_child->Parent(this);

	//Liste neu sortieren
	std::sort(children.begin(), children.end(), SceneObject::SortSubtreeCount);

	//Subtree erhhen
	SubTreeCount(subTreeCount + _child->subTreeCount + 1);
}

void SceneObject::RemoveChild(SceneObject* _child)
{
	assert(_child != nullptr);
	assert(_child->Parent() == this);

	//Subtree reduzieren
	SubTreeCount(subTreeCount - _child->SubTreeCount());

	for (ChildList::iterator it = children.begin(); it != children.end(); ++it)
	{
		if ((*it) == _child)
		{
			children.erase(it);
			break;
		}
	}

	_child->Parent(nullptr);
}

void SceneObject::ClearChildren()
{
	SubTreeCount(0);

	for (ChildList::iterator it = children.begin(); it != children.end(); ++it)
	{
		(*it)->Parent(nullptr);
	}

	children.clear();
}

inline const SceneObject::ComponentList& SceneObject::Components() const
{
	return this->components;
}

void SceneObject::AddComponent(IComponent* _component)
{
	components.insert(ComponentListEntry(_component->GetComponentType(), _component));

	ComponentChanged(this, ComponentChangedEventArgs(_component, ComponentActionType::ADD));
}

void SceneObject::RemoveComponent(IComponent* _component)
{
	std::pair<ComponentListIterator, ComponentListIterator> range = components.equal_range(_component->GetComponentType());

	for (auto it = range.first; it != range.second; ++it)
	{
		if (it->second == _component)
		{
			components.erase(it);
			ComponentChanged(this, ComponentChangedEventArgs(_component, ComponentActionType::REMOVE));
			break;
		}
	}
}

inline SceneObject* SceneObject::Parent() const
{
	return parent;
}

void SceneObject::Parent(SceneObject* _parent)
{
	if (parent != _parent)
	{
		SceneObject* lastParent = parent;
		parent = _parent;

		ParentChanged(this, ParentChangedEventArgs(lastParent, _parent));
	}
}

unsigned short SceneObject::SubTreeCount() const
{
	return subTreeCount;
}

void SceneObject::SubTreeCount(unsigned short _count)
{
	if (Parent() != nullptr)
		Parent()->SubTreeCount(Parent()->SubTreeCount() - subTreeCount + _count);

	subTreeCount = _count;
}

bool SceneObject::SortSubtreeCount(SceneObject* _first, SceneObject* _second)
{
	if (_first->subTreeCount < _second->subTreeCount)
		return true;
	return false;
}

inline const std::string& SceneObject::Name() const
{
	return name;
}