#define DLL_EXPORT

#include "World.hpp"

#include "SceneObject.hpp"
#include "../Core/System/SystemConfiguration.hpp"
#include "../Core/EventHandling/EventHandler.hpp"
#include "ParentChangedEventArgs.hpp"
#include "Components\ComponentChangedEventArgs.hpp"
#include "Components\RenderComponent.hpp"
#include "Components\LightComponents\LightComponent.hpp"
#include "Components\CameraComponent.hpp"
#include "Components\TransparentMeshRenderComponent.hpp"
#include "Components\PositionComponent.hpp"
#include "Components\BoundingComponent.hpp"
#include "Cameras\Camera.hpp"
#include <algorithm>
#include <queue>

using namespace Engine;
using namespace Engine::Math;
using namespace Engine::Scene;
using namespace Engine::Scene::Boundings;
using namespace Engine::Scene::Components;
using namespace Engine::Scene::Components::LightComponents;
using namespace Engine::Scene::Cameras;
using namespace Engine::Core::EventHandling;
using namespace Engine::Core::System;
using namespace Engine::Core::Collections;

World::World() : sceneObjectCount(0) { }

World::~World()
{
	//SceneObjects will be removed by SceneObject::dtor() calling RemoveSceneObject()
	while (sceneObjects.size() != 0)
		delete sceneObjects[0];
}

inline const std::vector<SceneObject*>& World::SceneObjects() const
{
	return sceneObjects;
}

inline const std::vector<RenderComponent*>& World::RenderObjects() const
{
	return renderObjects;
}

inline Components::CameraComponent* World::ScreenCamera() const
{
	return screenCamera;
}

inline unsigned int World::SceneObjectCount() const
{
	return sceneObjectCount;
}

void World::ScreenCamera(Components::CameraComponent* _screenCamera)
{
	screenCamera = _screenCamera;
}

const std::vector<Components::CameraComponent*>& World::Cameras() const
{
	return cameras;
}

void World::RegisterCamera(Components::CameraComponent* _camera)
{
	cameras.push_back(_camera);
}

void World::RegisterSceneObject(SceneObject* _sceneObject)
{
	_sceneObject->ParentChanged += new EventHandler<World, ParentChangedEventArgs>(this, &World::SceneObject_ParentChanged);
	_sceneObject->ComponentChanged += new EventHandler<World, ComponentChangedEventArgs>(this, &World::SceneObject_ComponentsChanged);

	if (_sceneObject->Parent() == nullptr)
	{
		sceneObjects.push_back(_sceneObject);
		std::sort(sceneObjects.begin(), sceneObjects.end(), SceneObject::SortSubtreeCount);
	}

	sceneObjectCount++;

	std::pair<HashString, SceneObject*> insertPair(HashString(_sceneObject->Name().c_str()), _sceneObject);
	nameLookupTable.insert(insertPair);

	OrderSceneTree();
}

void World::RemoveSceneObject(SceneObject* _sceneObject)
{
	_sceneObject->ParentChanged -= new EventHandler<World, ParentChangedEventArgs>(this, &World::SceneObject_ParentChanged);
	_sceneObject->ComponentChanged -= new EventHandler<World, ComponentChangedEventArgs>(this, &World::SceneObject_ComponentsChanged);

	if (_sceneObject->Parent() == nullptr)
	{
		for (auto it = sceneObjects.begin(); it != sceneObjects.end(); ++it)
		{
			if ((*it) == _sceneObject)
			{
				sceneObjects.erase(it);
				break;
			}
		}
	}

	sceneObjectCount--;
}

inline const std::vector<Components::LightComponents::LightComponent*>& World::Lights() const
{
	return lights;
}

inline const std::vector<IComponent*>& World::UpdateComponents() const
{
	return updateComponents;
}

inline const std::vector<Components::BoundingComponent*>& World::BoundingObjects() const
{
	return boundingObjects;
}

inline const std::vector<Components::PositionComponent*>& World::PositionComponents() const
{
	return positionComponents;
}

void World::RegisterUpdateComponent(IComponent* _component)
{
	updateComponents.push_back(_component);
}

void World::RemoveUpdateComponent(IComponent* _component)
{
	for (auto it = updateComponents.begin(); it != updateComponents.end(); ++it)
	{
		if (*it == _component)
		{
			updateComponents.erase(it);
			break;
		}
	}
}

void World::SceneObject_ParentChanged(void* _sender, ParentChangedEventArgs& _args)
{
	SceneObject* sceneObject = (SceneObject*)_sender;

	if (_args.OldValue() == nullptr)
	{
		//Remove from top elements
		for (auto it = sceneObjects.begin(); it != sceneObjects.end(); ++it)
		{
			if (*it == sceneObject)
			{
				sceneObjects.erase(it);
				break;
			}
		}
	}
	else if (_args.NewValue() == nullptr)
	{
		//Add to top elements
		sceneObjects.push_back(sceneObject);
	}

	OrderSceneTree();
}

void World::SceneObject_ComponentsChanged(void* _sender, ComponentChangedEventArgs& _args)
{
	ComponentType ctype = _args.Component()->GetComponentType();

	if (ctype > ComponentType::BeginRenderComponents)
	{
		if (_args.ActionType() == ComponentActionType::ADD)
		{
			renderObjects.push_back((RenderComponent*)_args.Component());
		}
		else if (_args.ActionType() == ComponentActionType::REMOVE)
		{
			for (auto it = renderObjects.begin(); it != renderObjects.end(); ++it)
			{
				if (*it == _args.Component())
				{
					renderObjects.erase(it);
					break;
				}
			}
		}
	}

	if (ctype > ComponentType::BeginLightComponents && ctype < ComponentType::EndLightComponents)
	{
		if (_args.ActionType() == ComponentActionType::ADD)
		{
			lights.push_back((LightComponent*)_args.Component());
		}
		else if (_args.ActionType() == ComponentActionType::REMOVE)
		{
			for (auto it = lights.begin(); it != lights.end(); ++it)
			{
				if (*it == _args.Component())
				{
					lights.erase(it);
					break;
				}
			}
		}
	}

	if (ctype == ComponentType::BoundingComponentType)
	{
		if (_args.ActionType() == ComponentActionType::ADD)
			boundingObjects.push_back((BoundingComponent*)_args.Component());
		else if (_args.ActionType() == ComponentActionType::REMOVE)
		{
			for (auto it = boundingObjects.begin(); it != boundingObjects.end(); ++it)
			{
				if (*it == _args.Component())
				{
					boundingObjects.erase(it);
					break;
				}
			}
		}
	}
}

void World::Resize(unsigned int _width, unsigned int _height)
{
	this->ScreenCamera()->Camera()->Resize(_width, _height);
}

void World::OrderTransparentObjects()
{
	transparentCache1.clear();
	transparentCache2.clear();

	for (int i = 0; i < renderObjects.size(); i++)
	{
		if (renderObjects[i]->GetComponentType() == ComponentType::TransparentMeshRenderComponentType)
		{
			if (((TransparentMeshRenderComponent*)renderObjects[i])->Bounding()->IsVisible())
			{
				transparentCache1.push_back(i);
				transparentCache2.push_back(i);
			}
		}
	}

	transparentObjectBsp.clear();

	OrderTransparentObjects(transparentCache1, transparentCache2, 0, transparentCache1.size());
}

int World::OrderTransparentObjects(std::vector<int>& _readVector, std::vector<int>& _writeVector, size_t _start, size_t _end)
{
	//std::cout << "Bsp from " << _start << " to " << _end << std::endl;

	if (_end - _start == 0)
		return -1;
	else if (_end - _start == 1)
	{
		transparentObjectBsp.push_back(BoxBspNode(_readVector[_start], Vec4f()));
		return (int)transparentObjectBsp.size() - 1;
	}
	else
	{
		bool collision = false;
		bool splitsSpaces = false;

		//Find seperating plane
		for (auto planeBoxIt = _readVector.begin() + _start; planeBoxIt != _readVector.begin() + _end; ++planeBoxIt)
		{
			const TransformedBox* planeBox = (TransformedBox*)((TransparentMeshRenderComponent*)renderObjects[*planeBoxIt])->Bounding()->WorldObject();
			Vec4f planes[6];
			planes[0] = planeBox->NormalizedAxis1();
			planes[1] = planeBox->NormalizedAxis1() * -1.0f;
			planes[2] = planeBox->NormalizedAxis2();
			planes[3] = planeBox->NormalizedAxis2() * -1.0f;
			planes[4] = planeBox->NormalizedAxis3();
			planes[5] = planeBox->NormalizedAxis3() * -1.0f;

			for (int i = 0; i < 6; i++)
			{
				planes[i].W(planeBox->PlaneDistances()[i]);
				collision = false;
				splitsSpaces = false;
				size_t insertStart = _start;
				size_t insertEnd = _end - 1;

				SplittingResult planeBoxResult = planeBox->IntersectPlane(planes[i]);

				for (auto testBoxIt = _readVector.begin() + _start; testBoxIt != _readVector.begin() + _end; ++testBoxIt)
				{
					if (*testBoxIt != *planeBoxIt)
					{
						const TransformedBox* testBox = (TransformedBox*)((TransparentMeshRenderComponent*)renderObjects[*testBoxIt])->Bounding()->WorldObject();
						SplittingResult sresult = testBox->IntersectPlane(planes[i]);

						//Test for collision
						if (sresult == SplittingResult::Split)
						{
							collision = true;
							break;
						}
						else if (sresult == SplittingResult::Front)
						{
							_writeVector[insertStart] = *testBoxIt;
							insertStart++;
						}
						else if (sresult == SplittingResult::Back)
						{
							_writeVector[insertEnd] = *testBoxIt;
							insertEnd--;
						}

						if (sresult != planeBoxResult)
							splitsSpaces = true;
					}
				}

				if (!collision && splitsSpaces)
				{
					_writeVector[insertStart] = *planeBoxIt;
					BoxBspNode node(-1, planes[i]);
					transparentObjectBsp.push_back(node);
					int idx = (int)transparentObjectBsp.size() - 1;

					node.Front(OrderTransparentObjects(_writeVector, _readVector, _start, insertStart+1));
					node.Back(OrderTransparentObjects(_writeVector, _readVector, insertEnd+1, _end));
					transparentObjectBsp[idx] = node;

					return idx;
				}
			}
		}
	}

	//If there is no plane -> Approximate one
	const TransformedBox* b1 = (TransformedBox*)((TransparentMeshRenderComponent*)renderObjects[_readVector[_start]])->Bounding()->WorldObject();
	const TransformedBox* b2 = (TransformedBox*)((TransparentMeshRenderComponent*)renderObjects[_readVector[_start + 1]])->Bounding()->WorldObject();
	Vec4f centerB1 = b1->Location() + b1->Axis1() * 0.5f + b1->Axis2() * 0.5f + b1->Axis3() * 0.5f;
	Vec4f centerB2 = b2->Location() + b2->Axis1() * 0.5f + b2->Axis2() * 0.5f + b2->Axis3() * 0.5f;

	Vec4f between = centerB2 - centerB1;
	Vec4f plane = Vec4f::Normalize3(between);
	plane.W(-Vec4f::Dot3(plane, centerB1 + between * 0.5f));
	
	size_t insertStart = _start;
	size_t insertEnd = _end - 1;

	for (auto testBoxIt = _readVector.begin() + _start; testBoxIt != _readVector.begin() + _end; ++testBoxIt)
	{
		const TransformedBox* testBox = (TransformedBox*)((TransparentMeshRenderComponent*)renderObjects[*testBoxIt])->Bounding()->WorldObject();
		
		Vec4f centerTestBox = testBox->Location() + testBox->Axis1() * 0.5f + testBox->Axis2() * 0.5f + testBox->Axis3() * 0.5f;
		float d = Vec4f::Dot(centerTestBox, plane);

		if (d >= 0.0f)
		{
			_writeVector[insertStart] = *testBoxIt;
			insertStart++;
		}
		else
		{
			_writeVector[insertEnd] = *testBoxIt;
			insertEnd--;
		}
	}
	BoxBspNode node(-1, plane);
	transparentObjectBsp.push_back(node);
	int idx = (int)transparentObjectBsp.size() - 1;

	node.Front(OrderTransparentObjects(_writeVector, _readVector, _start, insertStart));
	node.Back(OrderTransparentObjects(_writeVector, _readVector, insertEnd+1, _end));
	transparentObjectBsp[idx] = node;

	return idx;
}

inline const std::vector<BoxBspNode>& World::TransparentObjectBsp() const
{
	return transparentObjectBsp;
}

void World::OrderSceneTree()
{
	positionComponents.clear();
	
	std::queue<SceneObject*> q;
	
	for (auto it = sceneObjects.begin(); it != sceneObjects.end(); ++it)
		q.push(*it);

	while (!q.empty())
	{
		SceneObject* so = q.front();
		positionComponents.push_back(so->GetComponent<PositionComponent>());
		so->GetComponent<PositionComponent>()->StorageIndex((unsigned int)positionComponents.size() - 1);

		for (auto it = so->Children().begin(); it != so->Children().end(); ++it)
		{
			q.push(*it);
		}

		q.pop();
	}
}

SceneObject* World::FindSceneObject(HashString _name) const
{
	auto it = nameLookupTable.find(_name);

	if (it == nameLookupTable.end())
		return nullptr;
	else
		return it->second;
}