#define DLL_EXPORT

#include "InternalDebugger.hpp"

#include "../Logger.hpp"
#include "../Convert.hpp"

#include <sstream>
#include <iomanip>
#include <tchar.h>

using namespace Engine;
using namespace Engine::Util;
using namespace Engine::Util::Debugging;

#ifdef _ENGINE_USE_INTERNALDEBUGGING
HANDLE InternalDebugger::process = 0;
DWORD InternalDebugger::processID = 0;
#endif

void InternalDebugger::Init()
{
	#ifdef _ENGINE_USE_INTERNALDEBUGGING
	process = GetCurrentProcess();
	processID = GetCurrentProcessId();
	
	DWORD test = SymSetOptions(SYMOPT_LOAD_LINES);

	if ((test & SYMOPT_LOAD_LINES) == 0)
		Logger::Fatal("SymSetOptions failed");

	if (SymInitialize(process, NULL, TRUE) == FALSE)
	{
		Logger::Fatal("SymInitialize failed");
	}
	#endif
}

std::vector<StackFrame> InternalDebugger::GetStackTrace()
{
	#ifdef _ENGINE_USE_INTERNALDEBUGGING
	Init();
	HANDLE thread = GetCurrentThread();

	CONTEXT threadContext;
	memset(&threadContext, 0, sizeof(CONTEXT));
	threadContext.ContextFlags = CONTEXT_FULL;
	
	if (GetThreadContext(thread, &threadContext) == 0)
		Logger::Fatal("GetThreadContext failed");

	STACKFRAME64 stackFrame;
	memset(&stackFrame, 0, sizeof(STACKFRAME64));
	stackFrame.AddrPC.Offset = threadContext.Rip;
	stackFrame.AddrPC.Mode = AddrModeFlat;
	stackFrame.AddrFrame.Offset = threadContext.Rbp;
	stackFrame.AddrFrame.Mode = AddrModeFlat;
	stackFrame.AddrStack.Offset = threadContext.Rsp;
	stackFrame.AddrStack.Mode = AddrModeFlat;
	stackFrame.AddrReturn.Offset = 1;

	SYMBOL_INFO* symInfo = (SYMBOL_INFO*)malloc(sizeof(SYMBOL_INFO) + 150 * sizeof(TCHAR));
	symInfo->MaxNameLen = 150;
	symInfo->SizeOfStruct = sizeof(SYMBOL_INFO);
	IMAGEHLP_LINE64 lineInfo;
	memset(&lineInfo, 0, sizeof(IMAGEHLP_LINE64));
	lineInfo.SizeOfStruct = sizeof(IMAGEHLP_LINE64);
	IMAGEHLP_MODULE64  moduleInfo;
	memset(&moduleInfo, 0, sizeof(IMAGEHLP_MODULE64));
	moduleInfo.SizeOfStruct = sizeof(IMAGEHLP_MODULE64);

	std::vector<StackFrame> stacktrace;

	while (stackFrame.AddrPC.Offset != 0)
	{
		if (StackWalk64(
			IMAGE_FILE_MACHINE_AMD64,
			process,
			thread,
			&stackFrame,
			&threadContext,
			NULL,
			SymFunctionTableAccess64,
			SymGetModuleBase64,
			NULL
			) == 0)
		{
			break;
		}
		else
		{
			std::string symbol, assembly, file;
			unsigned int line;
			assembly = "";
			line = 0;
			file = "";
			

			//Resolve Symbol name
			if (SymFromAddr(process, stackFrame.AddrPC.Offset, 0, symInfo) == FALSE)
			{
					symbol = "[Unknown Address]";
			}
			else
			{
				symbol = std::string(symInfo->Name);
			}

			if (SymGetModuleInfo64(process, stackFrame.AddrPC.Offset, &moduleInfo) == FALSE)
			{
				assembly = "";
			}
			else
				assembly = moduleInfo.ModuleName;

			//Resolve line number
			DWORD  dwDisplacement;
			if (SymGetLineFromAddr64(process, stackFrame.AddrPC.Offset, &dwDisplacement, &lineInfo) == FALSE)
			{
				line = 0;
				file = "";
			}
			else
			{
				line = lineInfo.LineNumber;
				file = Convert::ToString(lineInfo.FileName);
			}

			stacktrace.push_back(StackFrame(symbol, assembly, file, line));

			if (symbol == "WinMain")
				break;
		}
	}

	free(symInfo);

	ShutDown();

	return stacktrace;
	#else
		return "";
	#endif
}

void InternalDebugger::PrintStackTrace()
{
	std::vector<StackFrame> frames = GetStackTrace();

	//Count length
	size_t lSymbol, lAssembly, lFile, lLine;
	lSymbol = 6;
	lAssembly = 8;
	lFile = 4;
	lLine = 8;

	auto startIt = frames.begin();
	while (startIt->SymbolName() != "Engine::Util::Debugging::InternalDebugger::PrintStackTrace")
		startIt++;
	startIt++;

	for (auto it = startIt; it != frames.end(); ++it)
	{
		lSymbol = max(lSymbol, it->SymbolName().length());
		lAssembly = max(lAssembly, it->Assembly().length());
		lFile = max(lFile, it->Filename().length());

		unsigned int l = it->Line();
		unsigned int thisLength = 0;
		while (l > 0)
		{
			l /= 10;
			thisLength++;
		}

		lLine = max(lLine, thisLength);
	}

	std::stringstream headline;
	headline << "Symbol" << std::setw(lSymbol + 2 - 6) << " ";
	headline << "Assembly" << std::setw(lAssembly + 2 - 8) << " ";
	headline << std::setw(lLine + 2) << "Line";
	Logger::InfoMessage(headline.str());

	int unknownCounter = 0;
	StackFrame unknownFrame;

	for (auto it = startIt; it != frames.end(); ++it)
	{
		if (unknownCounter > 0 && (it->Line() != 0 || it->Assembly() != unknownFrame.Assembly()))
		{
			std::stringstream printline;
			printline << unknownFrame.SymbolName() << std::setw(lSymbol + 2 - unknownFrame.SymbolName().length()) << " ";
			printline << unknownFrame.Assembly() << std::setw(lAssembly + 2 - unknownFrame.Assembly().length()) << " ";
			printline << std::setw(lLine + 2 - 7) << unknownCounter << " Frames";
			Logger::InfoMessage(printline.str());
			unknownCounter = 0;
		}

		if (it->Line() == 0)
			unknownCounter++;

		if (it->Line() != 0)
		{
			std::stringstream printline;
			printline << it->SymbolName() << std::setw(lSymbol + 2 - it->SymbolName().length()) << " ";
			printline << it->Assembly() << std::setw(lAssembly + 2 - it->Assembly().length()) << " ";
			printline << std::setw(lLine + 2) << it->Line();

			Logger::InfoMessage(printline.str());
		}

		unknownFrame = *it;
	}
}

void InternalDebugger::ShutDown()
{
	#ifdef _ENGINE_USE_INTERNALDEBUGGING
	SymCleanup(process);
	#endif
}