#ifndef _ENGINE_UTIL_SHAREDPOINTER
#define _ENGINE_UTIL_SHAREDPOINTER

#include "../dllconfig.hpp"

#include "../Core/Concurrency/Atomic.hpp"
#include "Logger.hpp"

#include <stdlib.h>
#include <sstream>

namespace Engine
{
	namespace Util
	{
		/// <summary>
		/// Defines a reference counting pointer. The pointer itself manages storagehandling.
		/// </summary>
		template <typename T>
		class SharedPointer
		{
		public:
			/// <summary>
			/// Initializes a new instanze of the SharedPointer class
			/// </summary>
			SharedPointer() : data(), counter(new Core::Concurrency::Atomic<int>(0)) { }
			/// <summary>
			/// Creates a SharedPointer from another SharedPointer of a subtype of the target pointer
			/// </summary>
			/// <param name="_value">SharedPointer to copy from</param>
			SharedPointer(const SharedPointer<T> &_value) : counter(_value.GetCounter()), data(_value.Get())
			{
				if (counter != nullptr) 
				{
					++(*counter);
				}
			}

			template <typename U>
			SharedPointer(const SharedPointer<U> &_value) : counter(_value.GetCounter()), data(_value.Get())
			{
				if (counter != nullptr) 
				{
					++(*counter);
				}
			}
			/// <summary>
			/// Creates a SharedPointer from a pointer to the data. The pointer has to be a subtype of the classtemplate type
			/// </summary>
			/// <param name="_pointer">Pointer to a subtype of the classtype</param>
			template <typename U>
			SharedPointer(U* _pointer) : data(_pointer)
			{
				counter = new Core::Concurrency::Atomic<int>(1);
			}

			/// <summary>
			/// Destructor for the SharedPointer
			/// </summary>
			~SharedPointer()
			{
				if (counter != nullptr)
				{
					--(*counter);
					
					//Wenn 0 dann Pointer auf alles Lschen, da keine Instanzen mehr da sein drfen
					if (*counter == 0)
					{
						delete data;
						delete counter;
					}
				}
			}

			/// <summary>
			/// Returns a pointer to the SharedPointer's data
			/// </summary>
			/// <returns>A pointer to the data</returns>
			T* Get() const
			{
				return data;
			}

			/// <summary>
			/// Returns the ThreadSafeCounter of the SharedPointer
			/// </summary>
			Engine::Core::Concurrency::Atomic<int>* GetCounter() const
			{
				return counter;
			}

			/// <summary>
			/// Copies the data from a SharedPointer of the same type
			/// </summary>
			/// <param name="_sourcePointer">Source to copy from</param>
			/// <returns>A reference to the copied pointer</returns>
			SharedPointer& operator = (const SharedPointer<T>& _sourcePointer)
			{
				if (_sourcePointer.GetCounter() != nullptr)
					++(*_sourcePointer.GetCounter());

				if (counter != nullptr)
				{
					--(*counter);

					if (*counter == 0)
					{
						delete data;
						delete counter;
					}
				}

				counter = _sourcePointer.GetCounter();
				data = _sourcePointer.Get();

				return *this;
			}

			/// <summary>
			/// Copies the data from a SharedPointer of a subtype of the pointers type
			/// <param name="_sourcePointer">Source to copy from</param>
			/// <returns>A reference to the copied pointer</returns>
			template <typename U>
			SharedPointer& operator = (const SharedPointer<U>& _sourcePointer)
			{
				if (_sourcePointer.GetCounter() != nullptr)
					++(*_sourcePointer.GetCounter());

				--(*counter);
				if (*counter == 0)
				{
					delete counter;
					delete data;
				}

				counter = _sourcePointer->GetCounter();
				data = _sourcePointer->Get();

				return *this;
			}
			/// <summary>
			/// Dereference operator for the SharedPointer
			/// </summary>
			/// <returns>A reference to the data object</returns>
			T& operator * () const
			{
				return *data;
			}
			/// <summary>
			/// Pointer operator for the SharedPointer
			/// <returns>A pointer to the data object</returns>
			T* operator -> () const
			{
				return data;
			}

			/// <summary>
			/// Frees the data of the SharedPointer
			/// </summary>
			inline void Delete()
			{
				delete counter;
				delete data;

				data = nullptr;
				counter = nullptr;
			}

			/// <summary>
			/// Defines the equal operator for two SharedPointers
			/// </summary>
			/// <param name="_toCompare">SharedPointer to compare with. Must be of the same type or subtype of the SharedPointer</param>
			/// <returns>True if the two objects are equal, otherwise false</returns>
			template <typename U>
			bool operator == (const SharedPointer<U>& _toCompare)
			{
				if (this->Get() == _toCompare.Get())
					return true;
				return false;
			}
			/// <summary>
			/// Defines the equal operator for a SharedPointer and a Pointer
			/// </summary>
			/// <param name="_toCompare">A pointer to the type or a subtype of the SharedPointer
			/// <returns>True if the two objects are equal, otherwise false</returns>
			template <typename U>
			bool operator == (const U* _toCompare)
			{
				if (this->Get() == _toCompare)
					return true;
				return false;
			}
			
			template <typename U, typename V>
			friend bool operator == (const U* _pointer, const SharedPointer<V>& _sharedPointer);
			/// <summary>
			/// Defines the unequal operator for two SharedPointer
			/// </summary>
			/// <param name="_toCompare">SharedPointer to compare</param>
			/// <returns>False if the two SharedPointer are unequal, true otherwise</returns>
			template <typename U>
			bool operator != (const SharedPointer<U>& _toCompare)
			{
				if (this->Get() != _toCompare.Get())
					return true;
				return false;
			}
			/// <summary>
			/// Defines the unequal operator for a SharedPointer and a pointer
			/// </summary>
			/// <param name="_toCompare">Pointer to compare</param>
			/// <returns>False if the two objects are unequal, true otherwise</returns>
			template <typename U>
			bool operator != (const U* _toCompare)
			{
				if (this->Get() != _toCompare)
					return true;
				return false;
			}

			template <typename U, typename V>
			friend bool operator != (const U* _pointer, const SharedPointer<V>& _sharedPointer);

		private:
			T* data;
			Engine::Core::Concurrency::Atomic<int> *counter;
		};

		/// <summary>
		/// Defines the equal operator for a SharedPointer and a Pointer
		/// </summary>
		/// <param name="_toCompare">A pointer to the type or a subtype of the SharedPointer
		/// <returns>True if the two objects are equal, otherwise false</returns>
		template <typename U, typename V>
		bool operator == (const U* _pointer, const SharedPointer<V>& _sharedPointer)
		{
			if (_pointer == _sharedPointer.Get())
				return true;
			return false;
		}
		/// <summary>
		/// Defines the unequal operator for a SharedPointer and a pointer
		/// </summary>
		/// <param name="_toCompare">Pointer to compare</param>
		/// <returns>False if the two objects are unequal, true otherwise</returns>	
		template <typename U, typename V>
		bool operator != (const U* _pointer, const SharedPointer<V>& _sharedPointer)
		{
			if (_pointer != _sharedPointer.Get())
				return true;
			return false;
		}
	}
}

#endif