#include <Engine\Engine.hpp>

#include <gl\glew.h>
#include <tchar.h>

#define _USE_MATH_DEFINES
#include <math.h>

using namespace Engine;

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, 
				   LPSTR lpCmdLine, int iCmdShow)
{
	bool showConsole = false;
	char* scene = "gitest";
	int cameraLocation = 2;

	//Check input parameter
	for (int currentArgument = 0; currentArgument < __argc; currentArgument++)
	{
		if (strcmp(__argv[currentArgument], "-console") == 0)
			showConsole = true;
		else if (strcmp(__argv[currentArgument], "-scene") == 0)
		{
			scene = __argv[currentArgument + 1];
			currentArgument ++;
		}
		else if (strcmp(__argv[currentArgument], "-cameraLocation") == 0)
		{
			cameraLocation = atoi(__argv[currentArgument + 1]);
			currentArgument ++;
		}
	}

	if (showConsole)
	{
		Core::Control::SystemConsole::OpenConsole();
		Core::Control::SystemConsole::PrintInfo();
		Core::Hardware::CPUInformation::Print();
	}

	Core::Control::Window* window = new Core::Control::Window(hInstance, _T("BlubbEngine2 - Game"), 1280, 768);
	window->Show();
	window->ShowBoundings(false);
	window->ShowTonemapper(false);
	window->ShowBloom(false);
	window->ShowLightVolumes(false);
	//window->ShowTimes(false);

	Scene::World* world = new Scene::World();
	EngineManager* em = new EngineManager(world, window);

	Engine::Util::StopWatch loadingStopWatch("Loading Time");
	loadingStopWatch.Start();

	Core::IO::AssetLoader loader(window->Context(), em);

	Scene::SceneObject* testso = nullptr;

	if (scene != nullptr)
		testso = loader.LoadAsset(std::string(scene), world);
	else
		testso = loader.LoadAsset(0, world);

	//CameraObject
	//Global Illumination
	Scene::SceneObject* giObject = new Scene::SceneObject(world, "GIObject");
	Scene::Components::VoxelOctreeComponent* octree = new Scene::Components::VoxelOctreeComponent(giObject, 100.0f);
	octree->GenerationStage(new Graphic::GI::Generation::OctreeGenerationStage(octree, em, window->Context(), 512*512*10));
	octree->PreparationStage(new Graphic::GI::Preparation::OctreePreparationStage(octree, em, window->Context()));
	//octree->IlluminationStage(new Graphic::GI::Illumination::AmbientOcclusionIlluminationStage(octree, window->Context()));
	//octree->IlluminationStage(new Graphic::GI::Illumination::DirectIlluminationStage(octree, window->Context()));
	octree->IlluminationStage(new Graphic::GI::Illumination::IndirectIlluminationStage(octree, window->Context()));
	octree->DebugRenderStage(new Graphic::GI::Rendering::OctreeDebugRenderStage(octree,	window->Context(), em));

	giObject->AddComponent(octree);

	Scene::SceneObject* cameraObject = new Scene::SceneObject(world, "CameraObject");
	cameraObject->GetComponent<Scene::Components::PositionComponent>()->Matrix(Math::Mat4f::Translate(Math::Vec3f(0.0f, 0.0f, 5.5f)));

	Scene::Cameras::GICamera* camera = new Scene::Cameras::GICamera((float)M_PI / 3.0f, 1280, 768, 0.1f, 100.0f, cameraObject, window->Context(), em, octree);
	//Scene::Cameras::RenderCamera* camera = new Scene::Cameras::RenderCamera((float)M_PI / 3.0f, 1280, 768, 0.1f, 100.0f, cameraObject, window->Context(), em);
	cameraObject->AddComponent(new Scene::Components::CameraComponent(cameraObject, camera));
	cameraObject->AddComponent(new Scene::Components::ControllerComponents::OrbitControllerComponent(cameraObject, em));
	//cameraObject->AddComponent(new Scene::Components::ControllerComponents::FreeFlyControllerComponent(cameraObject, em));

	


	//Lights
	Scene::SceneObject* pointLightObject3 = new Scene::SceneObject(world, "User_PointLightObject");
	pointLightObject3->GetComponent<Scene::Components::PositionComponent>()->Matrix( Math::Mat4f::Translate(Math::Vec3f(6.0f, 1.0f, 0.0f)));

	world->ScreenCamera(cameraObject->GetComponent<Scene::Components::CameraComponent>());

	bool running = true;

	loadingStopWatch.Stop();
	loadingStopWatch.Display(true);


	em->Clock().Start();
	em->Clock().Tick();


	Scene::SceneObject* movingCylinder = world->FindSceneObject("Zylinder1");
	if (movingCylinder != nullptr)
	{
		movingCylinder->GetComponent<Scene::Components::AnimationComponent>()->AnimationList().at("CylinderAction")->Speed(5.0f);
		movingCylinder->GetComponent<Scene::Components::AnimationComponent>()->AnimationList().at("CylinderAction")->Start();
	}

	while (running)
	{
		if (!Core::Control::Window::CheckMessages(running))
		{
			em->RenderFrame();
			em->WaitForFrame();
		
			Util::Logger::GLErrors();

			window->SwapBuffer();
		}
	}
	
	delete world;
	delete window;
	delete em;
}